/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <bin/sstmac.h>

#include <iostream>
#include <iterator>
#include <string>
#include <map>
#include <getopt.h>
#include <stdint.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <boost/lexical_cast.hpp>

#include <sstmac/backends/native/nodeid.h>

using namespace std;
using namespace sstmac;
using namespace sstmac::sw;
using namespace sstmac::hw;
using namespace sstmac::native;

using boost::lexical_cast;

std::ostream&
operator<<(std::ostream &os, const opts &oo)
{
  os << "  --help=" << oo.help << "\\\n" << "  --debug=" << oo.debug << "\\\n"
      << "  --configfile=" << oo.configfile << "\\\n" << "--mode=" << oo.mode
      << "\\\n";
  return os;
}

int
main(int argc, char **argv)
{
  opts oo;
  if (!parse_opts(argc, argv, oo))
    {
      if (oo.debug != "")
        logger::set_user_param(oo.debug);

      if(oo.color)
        logger::set_debug_coloring(true);

      if (oo.configfile == "")
        {
          std::cout << "ERROR: need a configuration file" << std::endl;
          return 1;
        }

      file_parameters::ptr params = file_parameters::construct(oo.configfile);

      if (oo.params)
        { // there were command-line overrides
          oo.params->combine_into(params);
        }

      // make a dummy to make sure the converter gets built
      //  (see nodeid.h)
      nodeaddress::ptr dummy = nodeid::construct(-1);

      if (oo.mode == "normal" || oo.mode == "notime")
        {
          run(params, oo.stoptime, (oo.mode != "notime"));
        }
      else if (oo.mode == "debug")
        {
          run_debugger(params, oo);
        }
    }
  else
    {
      return 1;
    }
}

int
parse_opts(int argc, char **argv, opts &oo)
{
  option gopt[] =
      {
          { "help", no_argument, NULL, 'h' },
          { "debug", required_argument, NULL, 'd' },
          { "stoptime", required_argument, NULL, 't' },
          { "configfile", required_argument, NULL, 'f' },
          { "param", required_argument, NULL, 'p' },
          { "mode", required_argument, NULL, 'm' },
          { "dull", no_argument, NULL, 'c' },
          { NULL, 0, NULL, '\0' } };
  int ch;
  bool errorflag = false;
  while ((ch = getopt_long(argc, argv, "hcd:f:t:p:m:", gopt, NULL)) != -1)
    {
      switch (ch)
      {
      case 'h':
        oo.help = 1;
        break;
      case 'd':
        oo.debug = optarg;
        break;
      case 'f':
        oo.configfile = optarg;
        break;
      case 't':
        oo.stoptime = timestamp(atof(optarg));
        break;
      case 'p':
        {
          if (!oo.params)
            {
              oo.params = command_parameters::construct();
            }
          pair<string, string> keyval = sim_parameters::parse_param(optarg);
          oo.params->add_param(keyval.first, keyval.second);
          break;
        }
      case 'm':
        oo.mode = optarg;
        break;

      case 'c':
        oo.color = false;
        break;

      default:
        std::cerr << "Unhandled input flag.\n";
        errorflag = true;
        std::cout << "optarg: " << optarg << std::endl;
        break;
      }
      if (oo.help)
        {
          print_help(argc, argv);
          return 1;
        }
      if (errorflag)
        {
          std::cerr << "Error return\n";
          return 2;
        }
    }
  std::cerr << oo;

  return 0;
}

void
run_debugger(sim_parameters::ptr params, opts oo)
{
  std::cout << "------------------------------------------ \n";
  std::cout << "--- Welcome to the SST/macro debugger ---- \n";
  std::cout << "------------------------------------------- \n";
  std::cout << "Type \"help\" for more information. \n";

  timestamp stop = oo.stoptime;
  std::string dbg_str = "timestamps | " + oo.debug;
  logger::set_user_param(dbg_str);
  bool exit = false;
  bool hasrunyet = false;
  eventmanager::ptr eventman;



  while (!exit)
    {
      std::cout << logger::red << "sstmac-debug >> "  << logger::black;
      std::string command;
      std::cin >> command;

      if (command == "help" || command == "h")
        {
          std::cout << "Commands are: \n";
          std::cout << "  help (h): print this help list \n";
          std::cout << "  start (st): start the simulation \n";
          std::cout << "  stopat (sa) <value>: set a breakpoint at the given time \n";
          std::cout << "  exit: exit the debugger \n";
          std::cout << "  silly: print something silly \n";
          std::cout << "  debug (d) <value>: add <value> to the debug list \n";
          std::cout << "  debug-list (dl): see the list of things being debugged \n";
          std::cout << "  debug-clear (cl): clear the debug list \n";
          std::cout << "  now: print the current time of the simulation \n";
          std::cout << "  step|next (n): execute one event \n";
          std::cout << "  reset: reset the simulation so it can be started again\n";
        }
      else if (command == "start" || command == "st")
        {
          if (!hasrunyet)
            eventman = run(params, stop, false);
          else
            eventman->run();
          hasrunyet = true;
        }
      else if (command == "stopat" || command == "sa")
        {
          double time;
          std::cin >> time;
          stop = timestamp(time);

          if(hasrunyet){
              eventman->schedule(stop, manager::stophandler::construct(),
                  sst_message::ptr());
          }
        }
      else if (command == "exit")
        {
          exit = true;
        }
      else if (command == "silly")
        {
          std::cout << "firetruck lightbulb purple-monkey-dishwasher \n";
        }
      else if (command == "debug" || command == "d")
        {
          std::string val;
          std::cin >> val;
          if (dbg_str == "")
            {
              dbg_str = "<debug> " + val;
            }
          else
            {
              dbg_str = dbg_str + " | <debug> " + val;
            }

          logger::set_user_param(dbg_str);
        }
      else if (command == "debug-list" || command == "dl")
        {
          std::cout << "debug list: " << dbg_str << "\n";
        }
      else if (command == "debug-clear" || command == "cl")
        {
          dbg_str = "timestamps";
          logger::set_user_param(dbg_str);
        }
      else if (command == "next" || command == "step" || command == "n")
        {
          if (!hasrunyet)
            {
              std::cout << "haven't started yet, have to start first \n";
            }
          else
            {
              eventman->step();
            }
        }
      else if (command == "now")
        {
          if (!hasrunyet)
            std::cout << "0 \n";
          else
            std::cout << eventman->now() << "\n";
        }else if(command == "reset"){

            stop = oo.stoptime;
            dbg_str =  "timestamps | " + oo.debug;
            hasrunyet = false;
            logger::set_user_param(dbg_str);

        }
      else
        {

          std::cout << "unknown command: " << command << "\n";
        }
    }

  std::cout << "---------------------------------------------------------- \n";
  std::cout << "--- Exiting the SST/macro debugger normally, so long! ---- \n";
  std::cout << "---------------------------------------------------------- \n";

}

void
print_help(int, char **argv)
{
  char *name = argv[0];
  char *tail = name + strlen(name);
  while (tail != name && *tail != '/')
    --tail;
  if (tail != name)
    ++tail;
  std::cerr
  << tail
  << " \\\n"
  << "    [(--help|h)]                           \\\n"
  << "    [(--dull|c)]                           \\\n"
  << "    [(--debug|-d)             <value> ]    \\\n"
  << "    [(--configfile|-f)        <value> ]    \\\n"
  << "    [(--param|-p)       <key>=<value> ]    \\\n"
  << "    [(--stoptime|-t)          <value> ]    \\\n"
  << "    [(--mode|-m)              <value> ]    \\\n"

  << "Configuration file is not optional. See parameters.ini for \n"
  << "    a list of parameters. \n"
  << "Turn output coloring off with --dull or -c, for make check and \n"
  << "   if you're not using bash or something else that can interpret \n"
  << "   ANSI escape sequences \n"
  << "--param (-p) is used to override any parameters in the \n "
  << "       configuration file, for example: \n"
  << "            -p network_name=buffer \n"
  << "--stoptime (-t) stops the simulation at the given time, \n"
  << "   which is useful for debugging \n"
  << "--mode (-m) can be used to open the debugger by saying -m debug, \n"
  << "    and can turn off printing the simulation at the end with -m notime \n"
  << "Valid arguments to --debug are strings of the form \n"
  << "       \"<(debug|stats)> (name1) | (name2) | ... \" \n"
  << "       - examples: \n"
  << "            - \"<debug> mpi\" turns on all mpi output \n"
  << "            - \"<debug> app | mpicollective\" turns on all \n"
  << "                   skeleton and mpicollective output \n"
  << "            - \"<stats> all\" turns on all stat collection \n"
  << "       - some current useful debug tags are: \n"
  << "            all sim app mpi mpiapi mpicollective mpiqueue os processor \n"
  << "            node network manager eventmanager allocation indexing \n"
  << "            launch lib_compute lib_sleep flower \n"
  << "       - current stat tags are: \n"
  << "            all processor network \n";

}

