/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODE_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODE_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/software/process/taskid.h>
#include <sstmac/backends/native/nodeid.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/eventmodule.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/common/debug.h>

#include <sstmac/software/services/launch/launcher.h>

#include <list>
#include <string>

namespace sstmac { namespace native
  {

  // Forward declarations of relevant types.

  class networkinterface;

  /**
   * An interface for "physical" nodes -- may eventually be decorated
   * with additional functionality (OS/cpu modeling or whatever).
   */
  class node: public eventmodule, public sstmac::sw::AMI {

    friend class message_handler;

  public:
    typedef boost::intrusive_ptr<node> ptr;

  protected:
    class message_handler: public eventhandler {
    protected:
      node::ptr parent_;

      message_handler(const node::ptr &p) :
        parent_(p) {

      }

    public:
      virtual void handle(const sst_message::ptr &msg) {
        if(parent_ == 0){
            throw ssterror("node::message_handler - my parent is null!");
        }
        if(parent_->os_ == 0){
            throw ssterror("node::message_handler - my parent's os is null!");
        }
        parent_->os_->handle_message(msg);
      }

      static boost::intrusive_ptr<message_handler> construct(
          const node::ptr &p) {
        boost::intrusive_ptr<message_handler> ret(new message_handler(p));
        return ret;
      }

      virtual std::string to_string() const {
        return "node::message_handler";
      }
    };

    boost::intrusive_ptr<message_handler> msg_handler_;

    sstmac::sw::environment::ptr env_;
    sstmac::sw::operatingsystem::ptr os_;

    /// The id for this node.
    nodeid::ptr nodeid_;

    sstmac::sw::launcher::ptr launcher_;

    static debug dbg_;

    node();



    public:
    void set_environment(const sstmac::sw::environment::ptr &env,
        sstmac::sw::appid aid);

    void start_application(sstmac::sw::app::ptr &a,
        const sim_parameters::ptr &params);

    /// Get a new node of the same type as the current one.
    virtual node::ptr
    clone() const = 0;

    /// Assign a node index and event call interface for this node.
    /// \throw nullerror if ev is null.
    virtual void
    initialize(nodeid::ptr my_id, const sstmac::sw::service::ptr &the_launcher);

    virtual void
    set_nic(const boost::intrusive_ptr<networkinterface> &n) = 0;

    virtual boost::intrusive_ptr<networkinterface>
    get_nic() const = 0;

    virtual void handle(const sst_message::ptr &msg);

    virtual std::string to_string() const {
      std::stringstream ss;
      ss << "node(" << nodeid_->to_string() << ")";
      return ss.str();
    }


  };

  } } // end of namespace sstmac

#endif
