/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/backends/native/components/node/simplenode.h>

#include <sstmac/backends/native/components/nic/networkinterface.h>

#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <iostream>
#include <sstream>

namespace sstmac
{
  namespace native
  {

    simplenode::simplenode(const timestamp &memory_latency,
        double memory_bandwidth, double frequency, int n_core,
        timestamp preempt) :
        memory_latency_(memory_latency), memory_bandwidth_(memory_bandwidth), freq_(
            frequency), cores_(n_core), preempt_(preempt)
    {
      os_->setAMI(sstmac::sw::AMI::AMI_COMM, this);
      os_->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
      os_->setAMI(sstmac::sw::AMI::AMI_QUERY, this);

      SSTMAC_DEBUG << "constructing a simplenode with parameters mem_lat = "
          << memory_latency_ << ", memory_bandwidth = " << memory_bandwidth_
          << ", freq = " << freq_ << ", cores = " << cores_ << "\n";

    }

    void
    simplenode::initialize(nodeid::ptr my_id,
        const sstmac::sw::service::ptr &the_launcher)
    {

      node::initialize(my_id, the_launcher);

      double tcycle = 1.0 / (freq_);

      timestamp tflop(tcycle * 10); // 10 clock cycles for a floating point
      //timestamp tmemop(memory_latency_.sec() + 1.0 / (memory_bandwidth_));
      timestamp tmemop = tflop;

      sstmac::hw::simpleprocessor_config::ptr cfg =
          sstmac::hw::simpleprocessor_config::construct(cores_, freq_, 10, 30);

      compute_sched_ = sstmac::hw::computescheduler::construct(cores_, preempt_);

      proc_model_ = sstmac::hw::simpleprocessor::construct(cfg, msg_handler_,
          compute_sched_);

      sstmac::hw::instructionprocessor_config::ptr pcfg =
          sstmac::hw::instructionprocessor_config::construct(cores_,
              INST_FLOP_LABEL, tflop, INST_MEMOP_LABEL, tmemop,
              INST_MEMOP_LOCAL_LABEL, tmemop, INST_MEMOP_RAND_LABEL, tmemop);

      proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg, msg_handler_,
          compute_sched_);

      mem_model_ = sstmac::hw::simple_memorymodel::construct(memory_bandwidth_, memory_latency_, msg_handler_);
    }

    node::ptr
    simplenode::clone() const
    {
      node::ptr ret = construct(memory_latency_, memory_bandwidth_, freq_, cores_,
          preempt_);

      return ret;
    }

    void
    simplenode::set_eventmanager(const eventmanager::ptr &m)
    {
      eventmodule::set_eventmanager(m);

      compute_sched_->set_eventmanager(m);
      mem_model_->set_eventmanager(m);

      os_->start_apps();
    }

    void
    simplenode::set_nic(const networkinterface::ptr &n)
    {
      nic_ = n;
    }

    networkinterface::ptr
    simplenode::get_nic() const
    {
      return nic_;
    }

    // --- sstmac::sw::AMI pure virtuals ---- //
    void
    simplenode::execute_kernel(int func, sst_message::ptr data)
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_COMM_SEND:
        {

          SSTMAC_DEBUG << "[t=" << now() << "] -- Node(" << nodeid_->to_string()
                << "): sending to " << data->toaddr_->to_string() << "\n";

          nodeid::ptr nid = boost::dynamic_pointer_cast < nodeid > (data->toaddr_);
          nic_->send(nid, data, msg_handler_);
          break;
        }

      case sstmac::sw::AMI::AMI_COMP_TIME:
        {

          proc_model_->compute(data);

          break;
        }

      case sstmac::sw::AMI::AMI_COMP_INSTR:
        {

          proc_inst_->compute(data);

          break;
        }

      case sstmac::sw::AMI::AMI_COMP_MEMACCESS:
        mem_model_->do_access(data);
        break;

      case sstmac::sw::AMI::AMI_COMP_SLEEP:
        {
          timestamp delay = now() + data->get_time();
          SSTMAC_DEBUG << "node (" << nodeid_->to_string() << "): sleeping until "
              << delay << "\n";
          schedule(delay, msg_handler_, data);
          break;
        }

      default:
        std::cout << "Kernel number: " << func << std::endl;
        throw ssterror("simplenode: cannot process kernel");

      }

    }

    ptr_type::ptr
    simplenode::execute_kernel(int func)
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_QUERY_TOPOLOGY:
        {
          sstmac::ptr_type::ptr top = nic_->get_topology();
          return top;
          break;
        }
      case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
        {
          return nodeid_;
          break;
        }

      case sstmac::sw::AMI::AMI_QUERY_MACHINE:
        {
          return sstmac::sw::machine_descriptor::construct(nic_->get_topology()->n(),
              nic_->get_topology()->name());
          break;
        }

        std::cout << "Kernel number: " << func << std::endl;
        throw ssterror("simplenode: cannot process kernel");
      }
    }

    timestamp
    simplenode::now()
    {
      return eventmodule::now();
    }

    bool
    simplenode::kernel_supported(int func)
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_COMM_SEND:
        return true;

      case sstmac::sw::AMI::AMI_COMM_NIC_SEND_ACK:
        return true;

      case sstmac::sw::AMI::AMI_COMP_TIME:
        return true;

      case sstmac::sw::AMI::AMI_COMP_SLEEP:
        return true;

      case sstmac::sw::AMI::AMI_COMP_INSTR:
        return true;

      case sstmac::sw::AMI::AMI_COMP_MEMACCESS:
        return true;

      case sstmac::sw::AMI::AMI_QUERY_TOPOLOGY:
        return true;

      case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
        return true;

      default:
        return false;

      }
    }

  }
} // end of namespace sstmac.
