/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "abstractnodeprocessor.h"

#include <sstmac/software/skeletons/skeletonfactory.h>

#include <sstmac/software/ami/hardware_descriptor.h>

#include <sstmac/common/logger.h>

#include <sstmac/software/libraries/compute/lib_compute_inst.h>
#include <sstmac/software/libraries/util/lib_display.h>

#include <sstmac/software/services/launch/instantlaunch.h>

#include <sstmac/common/driver_util.h>
#include <sstmac/backends/omnetpp/omnetpp_params.h>

namespace sstmac
{
  namespace omnetpp
  {

    Define_Module(sstmac::omnetpp::abstractnodeprocessor);

    using namespace std;

    debug abstractnodeprocessor::dbg_;

    sstmac::sw::allocationstrategy::ptr abstractnodeprocessor::allocation_;

    //map<int, Indexer*> Core::appIndexers;
    //map<sstmac::sw::appid, sstmac::sw::environment::ptr> abstractnodeprocessor::env_;
    bool abstractnodeprocessor::debug_init = false;

    abstractnodeprocessor::abstractnodeprocessor()
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> core | processor");
          logger::register_logger(&dbg_);
        }

    }

    abstractnodeprocessor::~abstractnodeprocessor()
    {

    }

    void
    abstractnodeprocessor::initialize()
    {

      int c = 0;

      int p = 0;

      int n = par("nodeid");

      int r = par("rackid");

      myid = omnetpp_address::construct(c, p, n, r);

      sim_parameters::ptr params = omnetpp_params::construct(this);




      coresPerProc = par("node_cores");
      numOfNodes = par("node_num");
      procsPerNode = par("node_procs");
      nodesPerRack = par("rack_nodes");
      numRacks = par("rack_num");

      omnetpp_address::set_system_params(coresPerProc, procsPerNode,
          nodesPerRack, numRacks);

      outgate = gate("toNetwork$o");
      ingate = gate("toNetwork$i");

      fillColor = sstmac::sw::GREY;
      borderColor = sstmac::sw::BLACK;

      string temp = par("logger_params");
      bool color = par("debug_coloring");
      logger::set_debug_coloring(color);

      if (!debug_init)
        {
          logger::set_user_param(temp);
          debug_init = true;
        }

      // ------- operating system -------
      sstmac::sw::hardware_descriptor::ptr hw =
          sstmac::sw::hardware_descriptor::construct(coresPerProc);
      opsys = sstmac::sw::operatingsystem::construct(hw);

      opsys->setAMI(sstmac::sw::AMI::AMI_COMM, this);
      opsys->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
      opsys->setAMI(sstmac::sw::AMI::AMI_DISP, this);
      opsys->setAMI(sstmac::sw::AMI::AMI_QUERY, this);

      // ---- get parameters and such
      string app_p_fname = par("app_paramfile");
      sim_parameters::ptr appparams = file_parameters::construct(app_p_fname);
      std::list<sstmac::sw::launchinfo::ptr> linfos = construct_launchers(
          appparams);

      instantiate_launcher(params, linfos);

      //--- processor models ----//
      msg_handler_ = msg_handler::construct(this);
      boost::intrusive_ptr<fakeeventmanager> fem = fakeeventmanager::construct(
          this);

      double pfreq = par("node_frequency");
      double mem_bw = par("node_mem_bandwidth");
      double mem_lat = par("node_mem_latency");
      double pre = par("node_preemption");
      // if (myid->rack() % 4 == 1) //degrade every other rack
      //   pfreq = pfreq / 2.0;
      double tcycle = 1.0 / (pfreq);

      timestamp tflop(tcycle * 10); // 10 clock cycles for a floating point
      timestamp tmemop(mem_lat + 1.0 / mem_bw);

      sstmac::hw::instructionprocessor_config::ptr pcfg =
          sstmac::hw::instructionprocessor_config::construct(coresPerProc,
              INST_FLOP_LABEL, tflop, INST_MEMOP_LABEL, tmemop,
              INST_MEMOP_LOCAL_LABEL, tmemop, INST_MEMOP_RAND_LABEL, tmemop);

      sstmac::hw::simpleprocessor_config::ptr scfg =
          sstmac::hw::simpleprocessor_config::construct(coresPerProc, pfreq, 10,
              30);

      compute_sched_ = sstmac::hw::computescheduler::construct(coresPerProc, timestamp(pre));

      proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg,
          msg_handler_, compute_sched_);
      proc_time_ = sstmac::hw::simpleprocessor::construct(scfg, msg_handler_,
          compute_sched_);

      compute_sched_->set_eventmanager(fem);

      logger::timer_ = fem;

      opsys->start_apps();
    }

    void
    abstractnodeprocessor::finish()
    {

      opsys->apps_kill();

    }

    void
    abstractnodeprocessor::instantiate_launcher(sim_parameters::ptr params,
        std::list<sstmac::sw::launchinfo::ptr> &tolaunch)
    {

      bool errorflag = false;

      indexstrategy::ptr idx = construct_indexing(
          params->get_param("launch_indexing"), params, errorflag);



      if (!allocation_)
        {
          allocation_ = construct_allocation(
              params->get_param("launch_allocation"), params,
              sstmac::hw::topology::ptr(), errorflag);

        }

      if (params->get_param("launch_name") == "instant")
        {
          launcher_ = sstmac::sw::instantlaunch::construct(tolaunch,
              allocation_, idx);
        }
      else
        {
          std::cout << "launchname: " << params->get_param("launch_name") << std::endl;
          opp_error("abstractnodeprocessor: unable to find launcher");
        }

      opsys->register_ser(launcher_);
    }

    bool
    abstractnodeprocessor::is_master_core()
    {
      return ((myid->rack() == 0 || myid->rack() < 0) && myid->node() == 0);
    }

    void
    abstractnodeprocessor::handleMessage(cMessage* cmsg)
    {

      if (cmsg->isSelfMessage())
        {

          EventWrapperMsg* emsg = check_and_cast<EventWrapperMsg*>(cmsg);
          eventhandlerpod epod = emsg->getEvhandler();
          eventhandler::ptr evh = epod.load;
          sst_message::ptr m = emsg->getMsgcontent().load;
          evh->handle(m);

          delete emsg;

        }
      else if (cmsg->getArrivalGate() == ingate)
        { //message arriving in from the NoC
          CoreMsg *m = check_and_cast<CoreMsg*>(cmsg);

          if (m->getToAddr()->rack() != myid->rack()
              || m->getToAddr()->node() != myid->node())
            {
              std::cout << "My ID: " << myid->to_string() << endl;
              std::cout << "Msg Dest: " << m->getToAddr()->to_string() << endl;
              opp_error("Application: got a message destined for NOT me");
            }

          switch (m->getMsgType())
            {

          case APP:
            {
              SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc "
                    << myid->to_string() << ": app message arrived \n";
              ApplicationMsg* appmsg = check_and_cast<ApplicationMsg*>(m);
              payloadpod ppod = appmsg->getMsgcontent();
              sst_message::ptr bsstm = ppod.load;

              opsys->handle_message(bsstm);

              break;
            }

          case NIC_SEND_ACK:
            {
              SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc "
                    << myid->to_string() << ": NIC send ack arrived ";
              ApplicationMsg* appmsg = check_and_cast<ApplicationMsg*>(
                  m->decapsulate());

              payloadpod ppod = appmsg->getMsgcontent();
              sst_message::ptr bsstm = ppod.load;

              SSTMAC_DEBUG << bsstm->to_string() << "\n";
              opsys->handle_message(bsstm);

              delete appmsg;
              //}
              break;
            }

          case OS_KILL:
            //delete app;
            // app = NULL;

            break;

          default:
            std::cout << "Proc id: " << myid->to_string() << "\n";
            opp_error("Proc: unknown OS message type received.");

            }

          delete cmsg;

        }
      else
        { //it's a scheduled message to send out

          send(cmsg, outgate);

        }
    }

    // --- sstmac::sw::AMI pure virtuals ---- //
    ptr_type::ptr
    abstractnodeprocessor::execute_kernel(int func)
    {
      switch (func)
        {

      case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
        {
          return myid;
          break;
        }

        }

      std::cout << "Kernel number: " << func << std::endl;
      throw ssterror("simplenode: cannot process kernel");
    }

    void
    abstractnodeprocessor::execute_kernel(int func, sst_message::ptr data)
    {
      switch (func)
        {
      case sstmac::sw::AMI::AMI_COMM_SEND:
        {

          payloadpod ppod(data);

          ApplicationMsg* cmsg = new ApplicationMsg();
          cmsg->setMsgType(APP);
          cmsg->setMsgcontent(ppod);
          omnetpp_address::ptr froma = boost::dynamic_pointer_cast<
              omnetpp_address>(data->fromaddr_);
          omnetpp_address::ptr toa =
              boost::dynamic_pointer_cast<omnetpp_address>(data->toaddr_);
          cmsg->setFromAddr(froma);
          cmsg->setToAddr(toa);
          cmsg->setByteLength(data->get_byte_length());
          cmsg->setKind(data->get_kind() + 1); //add 1, because 0 is reserved for Processor/management type stuff like launch

          SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc("
                << myid->to_string() << "): sending to " << toa->to_string()
                << "\n";

          send(cmsg, outgate);
          break;
        }

      case sstmac::sw::AMI::AMI_COMP_TIME:
        {

          proc_time_->compute(data);

          break;
        }

      case sstmac::sw::AMI::AMI_COMP_SLEEP:
        {
          EventWrapperMsg* emsg = new EventWrapperMsg();
          eventhandlerpod epod;
          epod.load = msg_handler_;

          emsg->setEvhandler(epod);
          emsg->setMsgcontent(data);

          scheduleAt(simTime() + data->get_time().sec(), emsg);

          break;
        }

      case sstmac::sw::AMI::AMI_COMP_INSTR:
        proc_inst_->compute(data);
        break;

      case sstmac::sw::AMI::AMI_DISP_FILL:
        {
          SSTMAC_DEBUG << "Proc: changing color to code " << data->get_kind()
                << "\n";
          fillColor = data->get_kind();
          cDisplayString& dispStr =
              getParentModule()->getParentModule()->getDisplayString();
          std::stringstream ss;
          ss << "b=30,30,rectangle," << get_color_string(fillColor) << ","
              << get_color_string(borderColor) << ",4";

          dispStr.parse(ss.str().c_str());
          break;
        }

      case sstmac::sw::AMI::AMI_DISP_BORDER:
        {
          SSTMAC_DEBUG << "Proc: changing color to code " << data->get_kind()
                << "\n";
          borderColor = data->get_kind();
          cDisplayString& dispStr =
              getParentModule()->getParentModule()->getDisplayString();
          std::stringstream ss;
          ss << "b=30,30,rectangle," << get_color_string(fillColor) << ","
              << get_color_string(borderColor) << ",4";

          dispStr.parse(ss.str().c_str());
          break;
        }

      default:
        std::cout << "Kernel number: " << func << endl;
        opp_error("Proc: received a kernel which I don't know how to execute");

        }

    }

    std::string
    abstractnodeprocessor::get_color_string(int c)
    {
      switch (c)
        {
      case sstmac::sw::BLACK:
        return "black";
        break;
      case sstmac::sw::RED:
        return "red";
        break;
      case sstmac::sw::GREEN:
        return "green";
        break;
      case sstmac::sw::YELLOW:
        return "yellow";
        break;
      case sstmac::sw::ORANGE:
        return "orange";
        break;
      case sstmac::sw::PURPLE:
        return "purple";
        break;
      case sstmac::sw::PINK:
        return "pink";
        break;

      case sstmac::sw::GREY:
        return "gray";
        break;

      default:
        return "white";
        }

    }

    timestamp
    abstractnodeprocessor::now()
    {
      timestamp t(simTime().dbl());
      return t;
    }

    bool
    abstractnodeprocessor::kernel_supported(int func)
    {
      switch (func)
        {
      case sstmac::sw::AMI::AMI_COMM_SEND:
        return true;

      case sstmac::sw::AMI::AMI_COMM_NIC_SEND_ACK:
        return true;

      case sstmac::sw::AMI::AMI_COMP_TIME:
        return true;

      case sstmac::sw::AMI::AMI_COMP_SLEEP:
        return true;

      case sstmac::sw::AMI::AMI_COMP_INSTR:
        return true;

      case sstmac::sw::AMI::AMI_DISP_BORDER:
        return true;

      case sstmac::sw::AMI::AMI_DISP_FILL:
        return true;

      default:
        return false;

        }
    }
  // ------------------------------------ //

  }
}
