/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_OMNETPP_NODE_PROCESSOR_ABSTRACTNODEPROCESSOR_H_INCLUDED
#define SSTMAC_BACKENDS_OMNETPP_NODE_PROCESSOR_ABSTRACTNODEPROCESSOR_H_INCLUDED

#include <omnetpp.h>

#include <sstmac/software/ami/environment.h>

#include <omnetppaddress.h>
#include <node_m.h>

#include <sstmac/common/debug.h>

#include <sstmac/common/sim_parameters.h>

#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/process/softwareid.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/hardware/processor/instructionprocessor.h>
#include <sstmac/hardware/processor/simpleprocessor.h>
#include <sstmac/hardware/processor/computescheduler.h>

#include <sstmac/common/eventmanager.h>
#include <Node/Processor/eventhandlerpod.h>
#include <queue>

#include <sstmac/software/services/launch/launcher.h>
#include <sstmac/software/services/launch/launchinfo.h>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>

namespace sstmac
{
  namespace omnetpp
  {

    //class Application;

#define APP_STACK_SIZE 16384

#define OS_MSG_HEADER_SIZE 8
#define APP_SIZE 1000

#define APP_START_DELAY 0.00001

#define SEND_DELAY 0.00001

    class abstractnodeprocessor : public cSimpleModule, public sstmac::sw::AMI
    {
      friend class msg_handler;
      friend class fakeeventmanager;

    protected:
      class msg_handler : public sstmac::eventhandler
      {
      protected:
        abstractnodeprocessor *parent_;

        msg_handler(abstractnodeprocessor *p) :
            parent_(p)
        {

        }

      public:
        virtual void
        handle(const sstmac::sst_message::ptr &msg)
        {
          if (parent_ == 0)
            {
              throw sstmac::ssterror(
                  "node::message_handler - my parent is null!");
            }
          if (parent_->opsys == 0)
            {
              throw sstmac::ssterror(
                  "node::message_handler - my parent's os is null!");
            }
          parent_->opsys->handle_message(msg);
        }

        static boost::intrusive_ptr<msg_handler>
        construct(abstractnodeprocessor *p)
        {
          boost::intrusive_ptr<msg_handler> ret(new msg_handler(p));
          return ret;
        }

        virtual std::string
        to_string() const
        {
          return "abstractnodeprocessor::message_handler";
        }
      };

      class fakeeventmanager : public sstmac::eventmanager
      {
        abstractnodeprocessor *parent_;

        fakeeventmanager(abstractnodeprocessor *p) :
            parent_(p)
        {

        }

      public:

        static boost::intrusive_ptr<fakeeventmanager>
        construct(abstractnodeprocessor *p)
        {
          boost::intrusive_ptr<fakeeventmanager> ret(new fakeeventmanager(p));
          return ret;
        }

        virtual void
        clear(const sstmac::timestamp &zero_time = sstmac::timestamp(0))
        {

        }

        virtual void
        run()
        {

        }

        virtual void
        step(){

        }

        virtual const sstmac::timestamp&
        now()
        {
          now_ = parent_->now();
          return now_;
        }

        /// Set off the given eventhandler at the given time.
        virtual void
        schedule(const sstmac::timestamp &start_time,
            const sstmac::eventhandler::ptr &handler,
            const sstmac::sst_message::ptr &msg)
        {
          payloadpod ppod(msg);
          eventhandlerpod epod(handler);

          EventWrapperMsg* emsg = new EventWrapperMsg();
          emsg->setMsgcontent(ppod);
          emsg->setEvhandler(epod);

          double d = start_time.sec();
          simtime_t time(d);

          //have to check for a conversion error
          if(time < parent_->opp_simtime())
            parent_->scheduleAt(parent_->opp_simtime(), emsg);
          else
            parent_->scheduleAt(time, emsg);
        }
      };

    public:
      abstractnodeprocessor();
      virtual
      ~abstractnodeprocessor();

      //void
      //processSignal(CoreSignal* sig);

      //virtual bool is_master_core();

      omnetpp_address::ptr
      get_addr()
      {
        return myid;
      }

      simtime_t
      opp_simtime(){
        return simTime();
      }

    protected:

      virtual void
      handleMessage(cMessage* cmg);
      virtual void
      initialize();
      virtual void
      finish();

      void
      instantiate_launcher(sim_parameters::ptr params,
          std::list<sstmac::sw::launchinfo::ptr> &tolaunch);




      bool
      is_master_core();

      omnetpp_address::ptr myid;
      int coresPerProc, procsPerNode, numOfNodes;
      int nodesPerRack;
      int numRacks;


      cGate* outgate;
      cGate* ingate;

      static sstmac::sw::allocationstrategy::ptr allocation_;

      // resourceallocator* allocator;

      //launching stuff
      //int nextApp;
      //std::map<int, std::string> appNames;
      // std::map<int, double> appStartTimes;
      // std::map<int, int> appSizes;
      // std::map<int, std::list<std::string>*> appStringParams;

      // std::map<int, indexer*> appIndexers;

      //Application* app;

      sstmac::sw::launcher::ptr launcher_;

      sstmac::sw::operatingsystem::ptr opsys;

      //map<int, API*> apis;
      //Space* appspace;

      //string app1Param1;

      //string app2Param1;

      static sstmac::debug dbg_;
      static bool debug_init;

     // static std::map<sstmac::sw::appid, sstmac::sw::environment::ptr> env_;

      // --- sstmac::sw::AMI pure virtuals ---- //
      virtual void
      execute_kernel(int func, sstmac::sst_message::ptr data);

      virtual ptr_type::ptr
      execute_kernel(int func);

      virtual bool
      kernel_supported(int func);

      virtual sstmac::timestamp
      now();

      // ------------------------------------ //

      //scheduling

      sstmac::hw::computescheduler::ptr compute_sched_;
      sstmac::hw::instructionprocessor::ptr proc_inst_;
      sstmac::hw::simpleprocessor::ptr proc_time_;
      boost::intrusive_ptr<msg_handler> msg_handler_;

      int fillColor;
      int borderColor;

      std::string
      get_color_string(int c);

    };

  }
}
#endif
