/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_OMNETPP_OMNETPPADDRESS_H_INCLUDED
#define SSTMAC_BACKENDS_OMNETPP_OMNETPPADDRESS_H_INCLUDED

#include <omnetpp.h>

#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/messages/sst_serializer.h>

#include <sstmac/common/logger.h>
#include <sstmac/common/ptr_type.h>

#include <iostream>
#include <sstream>
#include <stdint.h>
#include <vector>

namespace sstmac { namespace omnetpp {


  class omnetpp_address: public  sstmac::nodeaddress
  {

  protected:

    friend class omnetpp_address_converter;

    class omnetpp_address_converter : public nodeaddress::address_converter
    {

      omnetpp_address_converter()
                {
                }

    public:
      virtual nodeaddress::ptr
      convert(long uid)
      {
        long c = uid % omnetpp_address::numCores;
        long p = (uid / omnetpp_address::numCores) % omnetpp_address::numProcs;
        long n = (uid / (omnetpp_address::numCores * omnetpp_address::numProcs))
                      % omnetpp_address::numNodes;
        long r = (uid / (omnetpp_address::numCores * omnetpp_address::numProcs
            * omnetpp_address::numNodes)) % omnetpp_address::numRacks;
        return omnetpp_address::construct(c, p, n, r);
      }

      virtual
      ~omnetpp_address_converter()
      {
      }

      typedef boost::intrusive_ptr<omnetpp_address_converter> ptr;

      static omnetpp_address_converter::ptr
      construct()
      {
        return ptr(new omnetpp_address_converter());
      }
    };

    std::vector<int64_t> id;

    static long numCores;
    static long numProcs;
    static long numNodes;
    static long numRacks;



    omnetpp_address()
    {
      id.push_back(-1);
      id.push_back(-1);
      id.push_back(-1);
      id.push_back(-1);

      if(!converter_)
        converter_ = omnetpp_address_converter::construct();

    }

    omnetpp_address(std::vector<int64_t> &v)
    {
      std::vector<int64_t>::const_iterator it, end = v.end();

      for (it = v.begin(); it != end; it++)
        {
          id.push_back(*it);
        }

      if(!converter_)
        converter_ = omnetpp_address_converter::construct();
    }

    omnetpp_address(int c, int p, int n, int r) //: dbg_("<debug> address")
    {
      id.push_back(c);
      id.push_back(p);
      id.push_back(n);
      id.push_back(r);

      if(!converter_)
        converter_ = omnetpp_address_converter::construct();
    }

    public:

    typedef boost::intrusive_ptr<omnetpp_address> ptr;

    virtual ~omnetpp_address()
    {

    }

    static void
    set_system_params(long c, long p, long n, long r){
      numCores = c;
      numProcs = p;
      numNodes = n;
      numRacks = r;
    }

    virtual long unique_id() const
    {
      return rack() * (numCores * numProcs * numNodes) + node() * (numCores * numProcs) + proc() * numCores + core();
    }

    omnetpp_address operator=(const omnetpp_address &a)
    {
      id = a.id;
      return *this;
    }

    static omnetpp_address::ptr construct(std::vector<int64_t> &v)
    {
      omnetpp_address::ptr retval(new omnetpp_address(v));

      return retval;
    }

    static omnetpp_address::ptr construct()
    {
      omnetpp_address::ptr retval(new omnetpp_address());

      return retval;
    }

    static omnetpp_address::ptr construct(int c, int p, int n,
        int r)
    {
      omnetpp_address::ptr retval(
          new omnetpp_address(c, p, n, r));

      return retval;
    }

    static omnetpp_address::ptr construct(omnetpp_address::ptr &addr)
    {
      omnetpp_address::ptr retval(
          new omnetpp_address(addr->core(), addr->proc(), addr->node(),
              addr->rack()));

      return retval;
    }

    virtual void serialize(sstmac::sst_serializer* ser) const;

    virtual void
    deserialize(sst_serializer* ser);

    virtual std::string to_string() const
    {
      std::stringstream ss(std::stringstream::in | std::stringstream::out);
      ss << "Address(" << "rack: " << rack() << ", node: " << node()
                        << ", proc: " << proc() << ", core: " << core() << ")";
      return ss.str();
    }

    virtual bool equals(const nodeaddress::ptr &p) const
    {
      omnetpp_address::ptr addr = boost::dynamic_pointer_cast<omnetpp_address>(p);
      return (addr->rack() == rack() && addr->node() == node()
          && addr->proc() == proc() && addr->core() == core());
    }

    /*void
         copy(const Address &addr)
         {
         rackid = addr.rackid;
         nodeid = addr.nodeid;
         procid = addr.procid;
         // id = addr.id;
         }*/

    int64_t rack() const
    {
      return id[3];
    }

    int64_t node() const
    {
      return id[2];
    }

    int64_t proc() const
    {
      return id[1];
    }

    int64_t core() const
    {
      return id[0];
    }

    void rack(int64_t i)
    {
      id[3] = i;
    }

    void node(int64_t i)
    {
      id[2] = i;
    }

    void proc(int64_t i)
    {
      id[1] = i;
    }

    void core(int64_t i)
    {
      id[0] = i;
    }

  };



  inline bool operator==(const omnetpp_address::ptr &a, const omnetpp_address::ptr &b)
                    {
    return (a->equals(b));
                    }
  inline bool operator!=(const omnetpp_address::ptr &a, const omnetpp_address::ptr &b)
                    {
    return (!a->equals(b));
                    }

  inline std::ostream&
  operator<<(std::ostream &os, const omnetpp_address::ptr &n)
  {
    return (os << n->to_string());
  }

  inline sstmac::logger&
  operator<<(sstmac::logger &os, const omnetpp_address::ptr &n)
  {
    return (os << n->to_string());
  }

  /*inline bool operator<(const omnetpp_address &a, const omnetpp_address &b)
 {
 return (a.id < b.id);
 }
 inline bool operator<=(const omnetpp_address &a, const omnetpp_address &b)
 {
 return (a.id <= b.id);
 }
 inline bool operator>(const omnetpp_address &a, const omnetpp_address &b)
 {
 return (a.id > b.id);
 }
 inline bool operator>=(const omnetpp_address &a, const omnetpp_address &b)
 {
 return (a.id >= b.id);
 }*/

} }

void
 doPacking(cCommBuffer *buffer, sstmac::omnetpp::omnetpp_address::ptr& addr);
 void
 doUnpacking(cCommBuffer *buffer, sstmac::omnetpp::omnetpp_address::ptr& addr);

#endif
