/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/driver_util.h>

namespace sstmac
{

  using namespace std;



  void
  file_parameters::parse_file(std::string fname)
  {
    std::ifstream in(fname.c_str());
    if (in.is_open())
      {

        std::string line;

        while (in.good())
          {
            std::getline(in, line);

            if (line[0] == '#')
              { //a comment
                continue;
              }
            else if (line.find("=") != std::string::npos)
              { //an assignment
                pair<string, string> keyval = sim_parameters::parse_param(line);
                add_param_override(keyval.first, keyval.second);
              }
            else if (line.find("include") != string::npos)
              { //an include line
                std::string f = line.substr(7);
                std::string dir = "";
                size_t pos = fname.find_last_of('/');
                if (pos != string::npos)
                  {
                    dir = fname.substr(0, pos + 1);
                  }
                f = trim_str(f);
                parse_file(dir + f);
              }
          }

      }
    else
      {
        std::stringstream ss;
        ss << "file_parameters:  Failed to open infile " << fname;

        throw valueerror(ss.str());
      }
  }

  file_parameters::file_parameters(std::string fname)
  {
    parse_file(fname);

  }

  /**
   * Construct a circuit model
   * @param type circuittype obtained with a previous call to parse_circuit
   * @param top topology pointer
   * @param sim_params the simulation parameters, to get more (like latency, bandwidth, flower)
   */

  interconnect::ptr
  construct_circuit(sim_parameters::ptr params)
  {
    bool errorflag = false;
    circuittype cir_type = parse_circuit(
        params->get_param("network_name").c_str(), errorflag);

    interconnect::ptr ret;
    if (cir_type == FULLNETWORK)
      {
        ret = fullnetwork::construct(params);
      }
    else
      {
        topology::ptr top = construct_topology(params);
        if (cir_type == EXCLUSIVE)
          {
            ret = circuit::construct(params, top);
          }
        else if (cir_type == SHARED)
          {

            ret = sharedcircuit::construct(params, top);
          }
        else if (cir_type == BUFFER)
          {

            ret = sharedbuffer::construct(params, top);
          }
      }

    return ret;
  }

  /**
   * Construct a topology
   * @param args
   * @param netname
   * @return
   */

  topology::ptr
  construct_topology(sim_parameters::ptr params)
  {
    bool errorflag = false;
    topologytype netname = parse_topology(
        params->get_param("topology_name").c_str(), errorflag);
    std::vector<int> args = get_intvec(
        params->get_param("topology_geometry").c_str(), errorflag);

    topology::ptr net;
    if (netname == GAMMA)
      {
        if (args.size() == 2)
          {
            net = sstmac::hw::gamma::construct(args[0], args[1]);
          }
        else if (args.size() == 3)
          {
            net = sstmac::hw::gamma::construct(args[0], args[1], args[2]);
          }
        else if (args.size() == 4)
          {
            net = sstmac::hw::gamma::construct(args[0], args[1], args[2],
                args[3]);
          }
      }
    else if (netname == HCUBE)
      {
        if (args.size() == 1)
          {
            net = sstmac::hw::hypercube::construct(args[0]);
          }
      }
    else if (netname == TORUS)
      {
        if (args.size() == 2)
          {
            net = sstmac::hw::torus::construct_2D(args[0], args[1]);
          }
        else if (args.size() == 3)
          {
            net = sstmac::hw::torus::construct_2D(args[0], args[1], args[2]);
          }
        else if (args.size() == 4)
          {
            net = sstmac::hw::torus::construct_2D(args[0], args[1], args[2],
                args[3]);
          }
      }
    else if (netname == TORUS3)
      {
        if (args.size() == 3)
          {
            net = sstmac::hw::torus::construct_3D(args[0], args[1], args[2]);
          }
        else if (args.size() == 4)
          {
            net = sstmac::hw::torus::construct_3D(args[0], args[1], args[2],
                args[3]);
          }
        else if (args.size() == 5)
          {
            net = sstmac::hw::torus::construct_3D(args[0], args[1], args[2],
                args[3], args[4]);
          }
      }
    /* else if (netname == "hdtorus2D") {
     if (args.size() == 2) {
     net = sstmac::hw::hdtorus::construct_2D(args[0],args[1],1,1);
     }
     else if (args.size() == 3) {
     net = sstmac::hw::hdtorus::construct_2D(args[0],args[1],args[2],1);
     }
     else if (args.size() == 4) {
     net = sstmac::hw::hdtorus::construct_2D(args[0],args[1],args[2],args[3]);
     }
     }
     else if (netname == "hdtorus3D") {
     if (args.size() == 3) {
     net = sstmac::hw::hdtorus::construct_3D(args[0],args[1],args[2],1,1);
     }
     else if (args.size() == 4) {
     net = sstmac::hw::hdtorus::construct_3D(args[0],args[1],args[2],args[3],1);
     }
     else if (args.size() == 5) {
     net = sstmac::hw::hdtorus::construct_3D(args[0],args[1],args[2],args[3],args[4]);
     }
     }*/
    else if (netname == RING)
      {
        if (args.size() == 1)
          {
            net = sstmac::hw::ring::construct(args[0]);
          }
        else if (args.size() == 2)
          {
            net = sstmac::hw::ring::construct(args[0], args[1]);
          }
        else if (args.size() == 3)
          {
            net = sstmac::hw::ring::construct(args[0], args[1], args[2]);
          }
      }
    else if (netname == FATTREE)
      {
        if (args.size() == 2)
          {
            net = sstmac::hw::fattree::construct(args[0], args[1]);
          }
        if (args.size() == 3)
          {
            net = sstmac::hw::fattree::construct(args[0], args[1], args[2]);
          }
        if (args.size() == 4)
          {
            net = sstmac::hw::fattree::construct(args[0], args[1], args[2],
                args[3]);
          }
      }
    else if (netname == CROSSBAR)
      {
        if (args.size() == 1)
          {
            net = sstmac::hw::crossbar::construct(args[0]);
          }
      }
    else if (netname == CLOS)
      {
        if (args.size() == 1)
          {
            net = sstmac::hw::clos::construct(args[0]);
          }
      }
    else if (netname == TBIRD)
      {
        if (args.size() == 0)
          {
            net = sstmac::hw::tbird::construct();
          }
      }
    else if (netname == HDTORUS)
      {
        if (args.size() == 2)
          {
            net = sstmac::hw::hdtorus::construct_2D(args[0], args[1]);
          }
        if (args.size() == 3)
          {
            net = sstmac::hw::hdtorus::construct_3D(args[0], args[1], args[2]);
          }
        if (args.size() == 4)
          {
            net = sstmac::hw::hdtorus::construct_4D(args[0], args[1], args[2],
                args[3]);
          }
      }

    return net;
  }

  /**
   * Parse a string to find a corresponding topologytype enum value
   * @param value name of the topology
   * @param errorflag set to true if nothing matched
   * @return a standard topologytype value
   */

  netsharing
  parse_flower(const char *value, bool &errorflag)
  {
    std::string arg = value;
    for (unsigned i = 0; i < arg.size(); ++i)
      arg[i] = tolower(arg[i]);
    netsharing ns = ORDERED;
    if (arg == "overcommit")
      ns = OVERCOMMIT;
    else if (arg == "throttle_overcommits" || arg == "throttle-overcommits"
        || arg == "throttle")
      ns = THROTTLE_OVERCOMMITS;
    else if (arg == "ordered")
      ns = ORDERED;
    else
      {
        errorflag = true;
        std::cerr << "Invalid netsharing type: " << value << "\n";
      }
    return ns;
  }

  /**
   * Parse a string to find a corresponding topologytype enum value
   * @param value name of the topology
   * @param errorflag set to true if nothing matched
   * @return a standard topologytype value
   */

  topologytype
  parse_topology(const char *value, bool &errorflag)
  {
    // TORUS, CROSSBAR, GAMMA, FATTREE, TORUS3
    std::string arg = value;
    for (unsigned i = 0; i < arg.size(); ++i)
      arg[i] = tolower(arg[i]);
    topologytype nt = TORUS;
    if (arg == "torus" || arg == "torus2d" || arg == "2dtorus")
      nt = TORUS;
    else if (arg == "crossbar" || arg == "xbar" || arg == "cross")
      nt = CROSSBAR;
    else if (arg == "gamma" || arg == "gammagraph" || arg == "gamma-graph")
      nt = GAMMA;
    else if (arg == "fattree" || arg == "fat-tree" || arg == "fat")
      nt = FATTREE;
    else if (arg == "torus3" || arg == "torus3d" || arg == "3dtorus"
        || arg == "3d-torus")
      nt = TORUS3;
    else if (arg == "ring")
      nt = RING;
    else if (arg == "clos")
      nt = CLOS;
    else if (arg == "tbird")
      nt = TBIRD;
    else if (arg == "hcube" || arg == "hypercube")
      nt = HCUBE;
    else if(arg == "hdtorus") arg = HDTORUS;
    else
      {
        errorflag = true;
        std::cerr << "Invalid network type: " << value << std::endl;
      }
    return nt;
  }

  /**
   * Parses a string to find a corresponding circuittype enum value
   * @param value name of the circuit model
   * @param errorflag set to true if nothing matched
   * @return a standard circuittype value
   */

  circuittype
  parse_circuit(const char *value, bool &errorflag)
  {
    // EXCLUSIVE, SHARED
    std::string arg(value);
    for (unsigned i = 0; i < arg.size(); ++i)
      arg[i] = tolower(arg[i]);
    circuittype ct = SHARED;
    if (arg == "exclusive")
      ct = EXCLUSIVE;
    else if (arg == "shared")
      ct = SHARED;
    else if (arg == "buffer")
      ct = BUFFER;
    else if (arg == "full" || arg == "fullnetwork")
      ct = FULLNETWORK;
    else
      {
        errorflag = true;
        std::cerr << "Invalid circuit type: " << value << "\n";
      }
    return ct;
  }

  allocationstrategy::ptr
  construct_allocation(std::string all_name, sim_parameters::ptr params,
      topology::ptr top, bool &errorflag)
  {
    allocationstrategy::ptr all;
    std::string arg(all_name);
    for (unsigned i = 0; i < arg.size(); ++i)
      arg[i] = tolower(arg[i]);

    if (arg == "random")
      {
        all = randomallocation::construct(params);
      }
    else if (arg == "firstavailable")
      {
        all = firstavailableallocation::construct(params);
      }
    else if (arg == "rack")
      {
        all = rackallocator::construct(params);
      }
    else if (arg == "cart")
      {
        all = cartallocation::construct(params, top);
      }
    else if (arg == "dumpi")
      {
        all = dumpiallocation::construct(params, top);
      }
    else
      {
        errorflag = true;
        std::cerr << "Invalid allocation type: " << all_name << "\n";
      }

    return all;
  }

  indexstrategy::ptr
  construct_indexing(std::string ind_name, sim_parameters::ptr params,
      bool &errorflag)
  {
    std::string arg(ind_name);
    for (unsigned i = 0; i < arg.size(); ++i)
      arg[i] = tolower(arg[i]);
    indexstrategy::ptr idx;
    if (arg == "random")
      {
        idx = randomindexing::construct();
      }
    else if (arg == "block")
      {
        idx = blockindexing::construct();
      }
    else if (arg == "roundrobin")
      {
        idx = roundrobinindexing::construct();
      }
    else if (arg == "dumpi")
      {
        idx = dumpiindexer::construct(params);
      }
    else
      {
        errorflag = true;
        std::cerr << "Invalid indexing type: " << ind_name << "\n";
      }

    return idx;
  }

  std::list<launchinfo::ptr>
  construct_launchers(sim_parameters::ptr params)
  {
    std::list<launchinfo::ptr> ret;

    int appnum = 1;
    std::stringstream ss;
    while (true)
      {
        ss << "launch_app" << appnum;
        if (!params->has_param(ss.str()))
          {
            break;
          }

        std::string appname = params->get_param(ss.str());
        sstmac::sw::appid aid(appnum);

        ss << "_start";

        timestamp start = params->get_time_param(ss.str());

        ss.str("");

        ss << "launch_app" << appnum << "_size";

        long tasks = params->get_long_param(ss.str());

        launchinfo::ptr li = launchinfo::construct(appname, aid, params, start,
            tasks);
        ret.push_back(li);

        appnum++;
        ss.str("");
      }

    return ret;
  }

  double
  get_positive_float(const char *value, bool &errorflag)
  {
    char *endptr = NULL;
    double val = strtod(value, &endptr);
    // This is really pretty bad form
    errno = 0;
    if (*value == '\0' || *endptr != '\0')
      {
        errorflag = true;
        std::cerr << "Not a valid floating point value: " << value << "\n";
      }
    if (errno == ERANGE)
      {
        errorflag = true;
        std::cerr << "Value outside double precision range: " << value << "\n";
      }
    if (val <= 0)
      {
        errorflag = true;
        std::cerr << "Not a positive floating point value: " << value << "\n";
      }
    return val;
  }

  int
  get_int(const char *value, bool &errorflag)
  {
    char *endptr;
    errno = 0;
    long val = strtol(value, &endptr, 10);
    if (*value == '\0' || *endptr != '\0')
      {
        errorflag = true;
        std::cerr << "Invalid integer value: " << value << "\n";
      }
    if (errno == ERANGE || val > std::numeric_limits<int>::max()
        || val < std::numeric_limits<int>::min())
      {
        errorflag = true;
        std::cerr << "Value outside 32-bit integer range: " << value << "\n";
      }
    if (errno == EINVAL)
      {
        errorflag = true;
        std::cerr << "Value contains invalid characters: " << value << "\n";
      }
    return val;
  }

  int
  get_positive_int(const char *value, bool &errorflag)
  {
    int v = get_int(value, errorflag);
    if ((!errorflag) && v <= 0)
      {
        errorflag = true;
        std::cerr << "Not a positive non-zero integer value: " << value << "\n";
      }
    return v;
  }

  std::vector<int>
  get_intvec(const char *value, bool &errorflag)
  {
    std::vector<int> retval;
    char *brkt;
    char *val = strdup(value);
    const char *sep = " ,;_-";
    for (char *word = strtok_r(val, sep, &brkt); word != NULL;
        word = strtok_r(NULL, sep, &brkt))
      {
        retval.push_back(get_int(word, errorflag));
        if (errorflag)
          break;
      }
    free(val);
    return retval;
  }

}
