/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_DRIVER_UTIL_H
#define SSTMAC_DRIVER_UTIL_H

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>

//skeletons
#include <sstmac/software/skeletons/skeletonfactory.h>

//network topologies
#include <sstmac/hardware/network/topology/clos.h>
#include <sstmac/hardware/network/topology/composite.h>
#include <sstmac/hardware/network/topology/crossbar.h>
#include <sstmac/hardware/network/topology/fattree.h>
#include <sstmac/hardware/network/topology/gamma.h>
#include <sstmac/hardware/network/topology/hdtorus.h>
#include <sstmac/hardware/network/topology/hypercube.h>
#include <sstmac/hardware/network/topology/ring.h>
#include <sstmac/hardware/network/topology/tbird.h>
#include <sstmac/hardware/network/topology/torus.h>
#include <sstmac/hardware/network/topology/line.h>

//network congestion models
#include <sstmac/hardware/network/congestion/circuit.h>
#include <sstmac/hardware/network/congestion/fullnetwork.h>
#include <sstmac/hardware/network/congestion/ordered_flower.h>
#include <sstmac/hardware/network/congestion/overcommit_flower.h>
#include <sstmac/hardware/network/congestion/sharedcircuit.h>
#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#include <sstmac/hardware/network/congestion/throttle_flower.h>
#include <sstmac/hardware/network/congestion/typical_flower.h>

#include <sstmac/hardware/network/congestion/linearlatency.h>
#include <sstmac/hardware/network/congestion/constantlatency.h>
#include <sstmac/hardware/network/congestion/noisylatency.h>

//launching
#include <sstmac/software/services/launch/instantlaunch.h>

//indexing
#include <sstmac/software/services/launch/indexing/blockindexing.h>
#include <sstmac/software/services/launch/indexing/fixedindexing.h>
#include <sstmac/software/services/launch/indexing/randomindexing.h>
#include <sstmac/software/services/launch/indexing/roundrobinindexing.h>
#include <sstmac/software/services/launch/indexing/fromfileindexer.h>
#include <sstmac/software/services/launch/indexing/dumpiindexer.h>

//allocation
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/software/services/launch/allocation/cartallocation.h>
#include <sstmac/software/services/launch/allocation/rackallocator.h>
#include <sstmac/software/services/launch/allocation/dumpiallocation.h>



#include <sstmac/common/sim_parameters.h>

#include <sstmac/common/sstmac-config.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/common/units.h>
#include <ostream>
#include <boost/foreach.hpp>


namespace sstmac
{

  using namespace sstmac::hw;
  using namespace sstmac;
  using namespace sstmac::native;
  using namespace sstmac::sw;

  /**
   * The topologies that can be constructed
   */
  enum topologytype
  {
    TORUS, CROSSBAR, GAMMA, FATTREE, TORUS3, RING, TBIRD, CLOS, HCUBE, HDTORUS
  };

  /**
   * A way to convert a topologytype back to a string, for printing help
   * @param os
   * @param nt
   * @return
   */
  static std::ostream&
  operator<<(std::ostream &os, const topologytype &nt)
  {
    switch (nt)
      {
    case TORUS:
      os << "TORUS";
      break;
    case CROSSBAR:
      os << "CROSSBAR";
      break;
    case GAMMA:
      os << "GAMMA";
      break;
    case FATTREE:
      os << "FATTREE";
      break;
    case TORUS3:
      os << "TORUS3";
      break;
    case RING:
      os << "RING";
      break;
    case TBIRD:
      os << "TBIRD";
      break;
    case CLOS:
      os << "CLOS";
      break;
    case HCUBE:
      os << "HCUBE";
      break;
    case HDTORUS:
      os << "HDTORUS";
      break;
    default:
      os << "GARBAGE";
      break;

      }
    return os;
  }

  /**
   * The circuit types that can be constructed
   */
  enum circuittype
  {
    EXCLUSIVE, SHARED, BUFFER, FULLNETWORK
  };

  /**
   * A way to convert a circuittype back to a string, for printing help
   * @param os
   * @param ct
   * @return
   */
  inline std::ostream&
  operator<<(std::ostream &os, const circuittype &ct)
  {
    switch (ct)
      {
    case EXCLUSIVE:
      os << "EXCLUSIVE";
      break;
    case SHARED:
      os << "SHARED";
      break;
    case BUFFER:
      os << "BUFFER";
      break;
    case FULLNETWORK:
      os << "FULLNETWORK";
      break;
    default:
      os << "GARBAGE";
      break;
      }
    return os;
  }

  /// The network sharing models recognized by shared-bandwidth interconnects.
  enum netsharing
  {
    OVERCOMMIT, // Turn the network into a crossbar.
    THROTTLE_OVERCOMMITS,
    ORDERED
  };

  /**
   * A way to convert a netsharing back to a string, for printing help
   * @param os
   * @param ns
   * @return
   */
  inline std::ostream&
  operator<<(std::ostream &os, netsharing ns)
  {
    os << "netsharing(";
    switch (ns)
      {
    case OVERCOMMIT:
      os << "OVERCOMMIT";
      break;
    case THROTTLE_OVERCOMMITS:
      os << "THROTTLE_OVERCOMMITS";
      break;
    case ORDERED:
      os << "ORDERED";
      break;
    default:
      os << "INVALID_VALUE=" << int(ns);
      }
    os << ")";
    return os;
  }


  class command_parameters : public sstmac::sim_parameters
  {
  protected:
    command_parameters()
    {
    }

  public:
    typedef boost::intrusive_ptr<command_parameters> ptr;

    static ptr
    construct()
    {
      return ptr(new command_parameters());
    }

    virtual std::string
    get_param(std::string key) const
    {

      boost::unordered_map<std::string, std::string>::const_iterator it =
          params_.find(key);
      if (it == params_.end())
        {
          std::cout << "parameter named: " << key << std::endl;
          throw sstmac::valueerror("native_parameters: could not find parameter");
        }
      return it->second;
    }

    virtual bool
    has_param(std::string key) const
    {
      boost::unordered_map<std::string, std::string>::const_iterator it =
          params_.find(key);
      return (it != params_.end());
    }

    virtual void
    add_param(std::string key, std::string val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "\n";

      if (params_.find(key) != params_.end())
        {
          throw sstmac::valueerror(
              "command_parameters::add_param - key already in params");
        }
      else
        {
          params_[key] = val;
        }
    }

    virtual void
    add_param_override(std::string key, std::string val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "...";

      if (params_.find(key) != params_.end())
        {
          SSTMAC_DEBUG << "overriding \n";
        }
      else
        {
          SSTMAC_DEBUG << "adding new param \n";
        }

      params_[key] = val;

    }

    virtual void
    combine_into(sim_parameters::ptr sp)
    {
      BOOST_FOREACH(maptype::value_type it, params_)
            {
              sp->add_param_override(it.first, it.second);
            }
    }

    virtual std::string
    to_string() const
    {
      return "command_parameters";
    }

    virtual
    ~command_parameters()
    {
    }

  protected:
    typedef boost::unordered_map<std::string, std::string> maptype;
    maptype params_;
  };

  class file_parameters : public sstmac::sim_parameters
  {
  protected:
    file_parameters(std::string fname);

    void
    parse_file(std::string fname);
  public:
    typedef boost::intrusive_ptr<file_parameters> ptr;

    static ptr
    construct(std::string filename)
    {
      return ptr(new file_parameters(filename));
    }

    virtual std::string
    get_param(std::string key) const
    {

      boost::unordered_map<std::string, std::string>::const_iterator it =
          params_.find(key);
      if (it == params_.end())
        {
          std::cout << "parameter named: " << key << std::endl;
          throw sstmac::valueerror("file_parameters: could not find parameter");
        }
      return it->second;
    }

    virtual bool
    has_param(std::string key) const
    {
      boost::unordered_map<std::string, std::string>::const_iterator it =
          params_.find(key);
      return (it != params_.end());
    }

    virtual void
    add_param(std::string key, std::string val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "\n";

      if (params_.find(key) != params_.end())
        {
          throw sstmac::valueerror(
              "file_parameters::add_param - key already in params");
        }
      else
        {
          params_[key] = val;
        }
    }

    virtual void
    add_param_override(std::string key, std::string val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "...";

      if (params_.find(key) != params_.end())
        {
          SSTMAC_DEBUG << "overriding \n";
        }
      else
        {
          SSTMAC_DEBUG << "adding new param \n";
        }

      params_[key] = val;

    }

    virtual void
    combine_into(sim_parameters::ptr sp)
    {
      BOOST_FOREACH(maptype::value_type it, params_)
            {
              sp->add_param_override(it.first, it.second);
            }
    }

    virtual std::string
    to_string() const
    {
      return "file_parameters";
    }

    virtual
    ~file_parameters()
    {
    }

  protected:
    typedef boost::unordered_map<std::string, std::string> maptype;
    maptype params_;
  };




  /**
   * Parse a string to find a corresponding topologytype enum value
   * @param value name of the topology
   * @param errorflag set to true if nothing matched
   * @return a standard topologytype value
   */

  netsharing
  parse_flower(const char *value, bool &errorflag);

  /**
   * Parse a string to find a corresponding topologytype enum value
   * @param value name of the topology
   * @param errorflag set to true if nothing matched
   * @return a standard topologytype value
   */

  topologytype
  parse_topology(const char *value, bool &errorflag);

  /**
   * Parses a string to find a corresponding circuittype enum value
   * @param value name of the circuit model
   * @param errorflag set to true if nothing matched
   * @return a standard circuittype value
   */

  circuittype
  parse_circuit(const char *value, bool &errorflag);

  /**
   * Construct a circuit model
   * @param type circuittype obtained with a previous call to parse_circuit
   * @param top topology pointer
   * @param sim_params the simulation parameters, to get more (like latency, bandwidth, flower)
   */

  interconnect::ptr
  construct_circuit(sim_parameters::ptr sim_params);

  /**
   * Construct a topology
   * @param args
   * @param netname
   * @return
   */

  topology::ptr
  construct_topology(sim_parameters::ptr sim_params);

  allocationstrategy::ptr
  construct_allocation(std::string all_name, sim_parameters::ptr params,
      topology::ptr top, bool &errorflag);

  indexstrategy::ptr
  construct_indexing(std::string ind_name, sim_parameters::ptr params,
      bool &errorflag);

  std::list<launchinfo::ptr>
  construct_launchers(sim_parameters::ptr params);

  double
  get_positive_float(const char *value, bool &errorflag);

  int
  get_int(const char *value, bool &errorflag);

  int
  get_positive_int(const char *value, bool &errorflag);

  std::vector<int>
  get_intvec(const char *value, bool &errorflag);

}
#endif
