/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_ERRORS_H_INCLUDED
#define SSTMAC_COMMON_ERRORS_H_INCLUDED

#include <string>
#include <exception>
#include <stdlib.h> // abort()

namespace sstmac
{

  /**
   * General SST/macro errors, or base class for more specific errors.
   */
  struct ssterror : public std::exception
  {
    ssterror(const std::string &msg) :
        message(msg)
    {
    }
    virtual
    ~ssterror() throw ()
    {
    }
    virtual const char*
    what() const throw ()
    {
      return message.c_str();
    }
    const std::string message;
  };

  /**
   * Error indicating something was null and shouldn't have been
   */
  struct nullerror : public ssterror
  {
    nullerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~nullerror() throw ()
    {
    }
  };

  /**
   * Error indicating some internal value was unexpected.
   */
  struct valueerror : public ssterror
  {
    valueerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~valueerror() throw ()
    {
    }
  };

  /**
   * A general error somewhere from a child of sstmac::sw::library
   */
  struct libraryerror : public ssterror
  {
    libraryerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~libraryerror() throw ()
    {
    }
  };

  /**
   * An error indicating something to do with the advancement of simulation time
   */
  struct timeerror : public ssterror
  {
    timeerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~timeerror() throw ()
    {
    }
  };

  /**
   * File open, read, or write error
   */
  struct ioerror : public ssterror
  {
    ioerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~ioerror() throw ()
    {
    }
  };

  /**
   * An error indicating some format was not correct
   */
  struct illformederror : public ssterror
  {
    illformederror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~illformederror() throw ()
    {
    }
  };

  /**
   * An error having to do with sstmac::sw::operatingsystem
   */
  struct oserror : public ssterror
  {
    oserror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~oserror() throw ()
    {
    }
  };

  /**
   * Something happened when allocating, deallocating, or mapping
   * a memory region.
   */
  struct memoryerror : public ssterror
  {
    memoryerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~memoryerror() throw ()
    {
    }
  };

  /**
   * Something happened when iterating over a collection
   */
  struct iteratorerror : public ssterror
  {
    iteratorerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~iteratorerror() throw ()
    {
    }
  };

  /**
   * A function was intentionally unimplemented because it doesn't make sense,
   * or it is ongoing work
   */
  struct unimplementederror : public ssterror
  {
    unimplementederror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~unimplementederror() throw ()
    {
    }
  };

  /**
   * Key to a map was not in the map
   */
  struct invalidkeyerror : public ssterror
  {
    invalidkeyerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~invalidkeyerror() throw ()
    {
    }
  };

  /**
   * An index was out of range in a collection.
   */
  struct rangeerror : public ssterror
  {
    rangeerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~rangeerror() throw ()
    {
    }
  };

  /**
   * Invalid user input
   */
  struct inputerror : public ssterror
  {
    inputerror(const std::string &msg) :
        ssterror(msg)
    {
    }
    virtual
    ~inputerror() throw ()
    {
    }
  };

} // end of namespace sstmac
#endif

