/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_SST_MESSAGE_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_SST_MESSAGE_H_INCLUDED

#include <sstmac/common/nodeaddress.h>

#include <sstmac/common/messages/sst_serializer.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/common/ptr_type.h>


#include <map>


namespace sstmac {

/**
 * A class describing an event.
 */
class sst_message : public ptr_type
{

public:

  typedef boost::intrusive_ptr<sst_message> ptr;

  /**
   * Child classes should register their type here, for convenience
   */
  enum MESSAGE_TYPES
  {
    SST, MPI, ACTOR, SEND_ACK, COMPUTE, LAUNCH
  };


protected:

  /**
   * Protected constructor
   * @param t The type (child class, if any) of this message
   */
  sst_message(MESSAGE_TYPES t) :
       kind_(0), msgtype_(t)
  {
  }

  /**
   * Protected constructor with more parameters
   * @param libn The name of the library this message should be delivered to
   * @param t  A multi-use time parameter
   * @param ty Message type (child class)
   * @param k An int describing the kind, multi-use, non-critical
   */
  sst_message(std::string libn, timestamp t, MESSAGE_TYPES ty, int k);

  /**
   *
   * Protected constructor with all possible parameters
   * @param libn The name of the library this message should be delivered to
   * @param t  A multi-use time parameter
   * @param from Address of the source node
   * @param to Address of the destination node
   * @param ty Message type (child class)
   * @param k An int describing the kind, multi-use, non-critical
   */
  sst_message(std::string libn, timestamp t,
      nodeaddress::ptr from, nodeaddress::ptr to,
      MESSAGE_TYPES ty, int k);

public:

  virtual
  ~sst_message();

  /**
   * Smart pointer constructor
    * @param libn The name of the library this message should be delivered to
   * @param t  A multi-use time parameter
   * @param ty Message type (child class)
   * @param k An int describing the kind, multi-use, non-critical
   * @return A smart pointer to an sst_message
   */
  static ptr
  construct(std::string libn, timestamp t, MESSAGE_TYPES ty = SST, int k = 0);

  /**
     * Smart pointer constructor with all the params
      * @param libn The name of the library this message should be delivered to
     * @param t  A multi-use time parameter
     * @param from Address of the source node
     * @param to Address of the destination node
     * @param ty Message type (child class)
     * @param k An int describing the kind, multi-use, non-critical
     * @return A smart pointer to an sst_message
     */
    static ptr
    construct(std::string libn, timestamp t, nodeaddress::ptr from, nodeaddress::ptr to, MESSAGE_TYPES ty = SST, int k = 0);

  /**
   * Default no-parameter smart pointer constructor
   * @return
   */
  static ptr
  construct()
  {
    ptr ret(new sst_message(SST));
    return ret;
  }

  /**
   * Stringifier
   * @return String description
   */
  virtual std::string to_string() const{
          return "sst_message";
        }

  /**
   * Null pointer convenience function
   * @return A null smart pointer
   */
  static ptr
  null()
  {
    ptr ret;
    return ret;
  }
  // --------------------------------------//

public:

  /**
   * Library name getter
   * @return Name of deliver-to library
   */
  std::string
  get_lib_name() const
  {
    return lib_name_;
  }

  /**
   * Library name setter
   * @param s Name of deliver-to library
   */
  void
  set_lib_name(std::string s)
    {
      lib_name_ = s;
    }

  /**
   * Time getter
   * @return time field
   */
  timestamp
  get_time() const
  {
    return time_;
  }

  /**
   * Time setter
   * @param t time value
   */
  void
  set_time(timestamp t)
  {
    time_ = t;
  }

  /**
   * Copy smart constructor
   * @return A smart pointer to an identical but different sst_message
   */
  virtual ptr
  clone() const
  {
    return construct(lib_name_, time_, fromaddr_, toaddr_, msgtype_, kind_);
  }

  /**
   * Convenient smart constructor which changes the message type
   * @param t new message type
   * @return A smart pointer to a new sst_message with the given message type
   */
  virtual ptr
  clone(MESSAGE_TYPES t) const
  {
    return construct(lib_name_, time_, t);
  }

  /**
   * Virtual function to return size. Child classes should impement this
   * if they want any size tracked / modeled.
   * @return Zero size, meant to be implemented by children.
   */
  virtual long
  get_byte_length() const
  {
    return 0;
  }

  /**
   * Serialize this message during parallel simulation.
   * @param ser The serializer to use
   */
  virtual void
  serialize(sst_serializer* ser);

  /**
   * Deserialize during parallel simulation.
   * @param ser The serializer to use
   * @return A smart pointer to an sst_message
   */
  static ptr
  deserialize(sst_serializer* ser);

  /**
   * Actually deserialize the fields of sst_message.
   * This should be used by children class from their
   * own deserialize(...) functions.
   * @param ser The serializer to use
   */
  void
  do_deser(sst_serializer* ser);

  /**
   * Message type getter
   * @return message type (child class)
   */
  MESSAGE_TYPES
  get_type() const
  {
    return msgtype_;
  }

  /**
   * Message kind getter
   * @return message kind
   */
  int get_kind() {return kind_;}

  /**
   * Message kind setter
   * @param k message kind value
   */
  void set_kind(int k){kind_ = k;}

protected:

  std::string lib_name_;

  timestamp time_;

  //used by libraries to distinguish categories of messages
  int kind_;

  MESSAGE_TYPES msgtype_;

public:
  nodeaddress::ptr fromaddr_;
  nodeaddress::ptr toaddr_;
};
} // end of namespace sstmac
#endif
