/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_NODEADDRESS_H_INCLUDED
#define SSTMAC_COMMON_NODEADDRESS_H_INCLUDED


#include <sstmac/common/messages/sst_serializer.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/logger.h>

#include <sstmac/common/errors.h>

namespace sstmac {

  /**
   * The abstract base class for an address of a node. This allows different
   * backends to structure themselves however they want and have convenient
   * addressing.
   */
  class nodeaddress : public ptr_type
  {
  public:
    typedef boost::intrusive_ptr<nodeaddress> ptr;

  protected:
    nodeaddress(){};

    /**
     * Protected internal class to convert a long unique id
     * back into a smart nodeaddress pointer.  An implementation
     * of this should be set to nodeaddress::converter_
     */
    class address_converter : public ptr_type {
    public:
      virtual nodeaddress::ptr convert(long uid) = 0;

      virtual ~address_converter() {}

      typedef boost::intrusive_ptr<address_converter> ptr;

      virtual std::string
      to_string() const {
        return "address_converter";
      }
    };

    static address_converter::ptr converter_;

    public:
    virtual std::string to_string() const{
      return "nodeaddress";
    }



    virtual ~nodeaddress()
    {

    }

    /**
     * Comparer of nodeaddresses for sorted collections.
     */
    struct nodeaddresscomp {
      bool operator() (const nodeaddress::ptr& lhs, const nodeaddress::ptr& rhs) const
      {return lhs->unique_id()<rhs->unique_id();}
    };


    virtual bool equals(const nodeaddress::ptr &p) const = 0;

    /// This needs to happen for working with topologies and network models in SSTMM
    virtual long unique_id() const = 0;

    virtual void serialize(sst_serializer* ser) const = 0;

    virtual void
    deserialize(sst_serializer* ser){
      throw unimplementederror("nodeaddress::deserialize: not implemented");
    }

    static nodeaddress::ptr
    convert_from_uniqueid(long uid){
      if(converter_)
        return converter_->convert(uid);
      else
        return ptr();
    }




  };



  inline std::ostream&
  operator<<(std::ostream &os, const nodeaddress &n)
  {
    return (os << n.to_string());
  }
  inline logger&
  operator<<(logger &os, const nodeaddress &n)
  {
    return (os << n.to_string());
  }

} // end of namespace sstmac
#endif
