/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_TIMESTAMP_H_INCLUDED
#define SSTMAC_COMMON_TIMESTAMP_H_INCLUDED

#include <sstmac/common/errors.h>
#include <sstmac/common/sstmac-config.h>
#include <iosfwd>
#include <limits>
#include <stdint.h>
#include <boost/lexical_cast.hpp>
#include <iostream>

#ifdef SSTMAC_USE_GMPXX
#include <gmpxx.h>
#endif

namespace sstmac {

/**
 * A basic container for time (subject to future transplant).
 * Stores time as an integral number of picoseconds (tentatively).
 * With 1 psec resolution, a 64 bit int can hold roughly +/- 106 days.
 *
 * Intended to be reasonably compatible with ns3::HighPrecision time.
 */
class timestamp
{
public:
#ifndef SSTMAC_USE_GMPXX
  /// The type that holds a timestamp.
  typedef int64_t tick_t;
#else
  /// If GMP is enabled, timestamps have no limit (but run slower).
  typedef mpz_class tick_t;
#endif // ! SSTMAC_USE_GMPXX
private:
  /// Picoseconds between clock ticks.
  static tick_t *psec_tick_spacing_;
  static std::string *tick_spacing_string_;
  static void
  init_stamps();
  static timestamp
  scaled_time(int64_t value, int64_t scaling, const char *caller,
      const char *units);

  /// The current time value of this container in ticks.
  tick_t ticks_;

public:
  /// Round the given time value (in seconds) to the nearest
  /// representable internal time.
  /// \throw time_error if this time is outside the range that can
  /// can be represented with this time container.
  explicit
  timestamp(double t = 0);

  /// Convert a tick type to int64_t.
  int64_t
  ticks_int64();

  /// Return the current time in seconds.
  double
  sec() const;

  /// Return the current time in milliseconds.
  double
  msec() const;

  /// Return the current time in microseconds.
  double
  usec() const;

  /// Return the current time in nanoseconds.
  double
  nsec() const;

  /// Return the current time in picoseconds.
  double
  psec() const;

public:
 // double
 // to_dbl() const{
 //   tick_t psec = ticks() / tick_interval();
 //   return (double)psec * 1e-12;
 // }

  /// Get the number of ticks.
  inline tick_t
  ticks() const
  {
    return ticks_;
  }

  /// Get the tick interval in picoseconds.
  static tick_t
  tick_interval();

  /// Get the tick interval in string form (for example, "1ps").
  static const std::string &
  tick_interval_string();

  /// Get the number of ticks per second (1e12/tick_interval()).
  static tick_t
  frequency();

  /// Get the largest time value possible (in seconds).
  static double
  max_time();

  /// Get the smallest (most negative) time value possible (in seconds).
  static double
  min_time();

  /// Get a time value corresponding exactly to the given number of
  /// picoseconds.
  /// \throw time_error if this time cannot be exactly represented.
  static timestamp
  exact_psec(int64_t psec);

  /// Get a time value corresponding exactly to the given number of
  /// nanoseconds.
  /// \throw time_error if this time cannot be exactly represented.
  static timestamp
  exact_nsec(int64_t nsec);

  /// Get a time value corresponding exactly to the given number of
  /// microseconds.
  /// \throw time_error if this time cannot be exactly represented.
  static timestamp
  exact_usec(int64_t usec);

  /// Get a time value corresponding exactly to the given number of
  /// milliseconds.
  /// \throw time_error if this time cannot be exactly represented.
  static timestamp
  exact_msec(int64_t msec);

  /// Get a time value corresponding exactly to the given number of
  /// seconds.
  /// \throw time_error if this time cannot be exactly represented.
  static timestamp
  exact_sec(int64_t sec);

  /// Get a time value with exactly the given number of ticks.
  /// This is a template function to ensure that we do proper range checking
  /// on input values.
  template<typename T>
    static timestamp
    exact_ticks(T val)
    {
#ifndef SSTMAC_USE_GMPXX
      // This test is unnecessary, but allows the test on val to be
      // optimized out if tick_t == T.
      if (typeid(T) != typeid(tick_t))
        {
          if (std::numeric_limits<T>::is_signed
              && !std::numeric_limits<tick_t>::is_signed)
            {
              if (val < 0)
                {
                  throw valueerror(
                      "timestamp::exact_ticks: value is negative but tick_t is not signed");
                }
            }
          else if (std::numeric_limits<T>::is_signed
              && std::numeric_limits<tick_t>::is_signed)
            {
              if (val < std::numeric_limits<tick_t>::min())
                {
                  throw valueerror(
                      "timestamp::exact_ticks: value is too negative to represent");
                }
            }
          if (val > std::numeric_limits<tick_t>::max())
            {
              throw valueerror(
                  "timestamp::exact_ticks: value is too large to represent");
            }
        }
#endif // ! SSTMAC_USE_GMPXX
      timestamp ts(0);
      ts.ticks_ = tick_t(val);
      return ts;
    }

  /// Fast and exact comparison operations.
  inline bool
  operator==(const timestamp &other) const
  {
    return (ticks_ == other.ticks_);
  }
  inline bool
  operator!=(const timestamp &other) const
  {
    return (ticks_ != other.ticks_);
  }
  inline bool
  operator<(const timestamp &other) const
  {
    return (ticks_ < other.ticks_);
  }
  inline bool
  operator<=(const timestamp &other) const
  {
    return (ticks_ <= other.ticks_);
  }
  inline bool
  operator>(const timestamp &other) const
  {
    return (ticks_ > other.ticks_);
  }
  inline bool
  operator>=(const timestamp &other) const
  {
    return (ticks_ >= other.ticks_);
  }
  timestamp&
  operator+=(const timestamp &other);
  timestamp&
  operator-=(const timestamp &other);
  timestamp&
  operator*=(double scale);
  timestamp&
  operator/=(double scale);
};

timestamp
operator+(const timestamp &a, const timestamp &b);
timestamp
operator-(const timestamp &a, const timestamp &b);
timestamp
operator*(const timestamp &t, double scaling);
timestamp
operator*(double scaling, const timestamp &t);
timestamp
operator/(const timestamp &t, double scaling);

std::ostream&
operator<<(std::ostream &os, const timestamp &t);

} // end of namespace sstmac

#endif
