/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/units.h>
#include <errno.h>

namespace sstmac
{

  /// Multiply two 64 bit integer values and check for overflow.
  /// This would be two lines if we had 128-bit integers.
  int64_t
  multiply64(int64_t a, int64_t b, bool &errorflag)
  {
    // Use detection of 32-bit overflow to help us.
    //  let
    //      signed    scaling = 1 << 32
    //      unsigned  mask = scaling >> 1
    //      signed    au = a / scaling
    //      unsigned  al = a & mask
    //      signed    bu = b / scaling
    //      unsigned  bl = b & mask
    //  then
    //      a * b = (au << 32 + al) * (bu << 32 + bl)
    //            = (au * bu << 64) + ((au * bl + bu * al) << 32) + (al * bl)
    // so we have an overflow if any of the following hold:
    //      au and bu are both non-zero
    //      au*bl + bu*al is greater than 0x7fffffff (now that au or bu is zero)
    //      al * bl is greater than 0x7fffffff
    //      the sum ((au*bl + bu*al) << 32) + (al*bl) overflows
    int64_t scaling = int64_t(1) << 32;
    int64_t mask = scaling - 1;
    int64_t max32 = scaling >> 1;
    uint64_t max64 = (~uint64_t(0)) >> 1;
    int64_t au = a / scaling, bu = b / scaling;
    int64_t al = a & mask, bl = b & mask;
    int64_t mul1 = au * bl + bu * al;
    uint64_t mul2 = uint64_t(al) * uint64_t(bl);
    uint64_t uval = uint64_t(mul1) * uint64_t(scaling) + mul2;
    if ((au && bu) || (mul1 > max32) || (mul2 > max64) || (uval > max64))
      {
        errorflag = true;
        std::cerr << "64-bit integer overflow in " << a << "*" << b << "\n";
      }
    return int64_t(uval);
  }

  void
  populate_bandwidth_names(std::map<std::string, int64_t> &abbrname,
  std::map<std::string, int64_t> &fullname)
  {
    abbrname["bps"] = abbrname["b/s"] = abbrname["b/sec"] = fullname["bits/s"] =
        fullname["bits/sec"] = fullname["bits/second"] =
            fullname["bits per second"] = 1;
    // Bytes per second.
    abbrname["Bps"] = abbrname["B/s"] = abbrname["B/sec"] =
        fullname["bytes/s"] = fullname["bytes/sec"] = fullname["bytes/second"] =
            fullname["bytes per second"] = 8;
    // Kilobits per second.
    abbrname["kbps"] = abbrname["Kbps"] = abbrname["kb/s"] = abbrname["Kb/s"] =
        fullname["kbits/s"] = fullname["kbits/sec"] = fullname["kbits/second"] =
            fullname["kilobits/s"] = fullname["kilobits/sec"] =
                fullname["kilobits/second"] = fullname["kilobits per second"] =
                    1000;
    // Kibibits per second.
    abbrname["kibps"] = abbrname["Kibps"] = abbrname["kib/s"] =
        abbrname["Kib/s"] = fullname["kibits/s"] = fullname["kibits/sec"] =
            fullname["kibits/second"] = fullname["kibibits/s"] =
                fullname["kibibits/sec"] = fullname["kibibits/second"] =
                    fullname["kibibits per second"] = 1024;
    // Kilobytes per second.
    abbrname["kBps"] = abbrname["KBps"] = abbrname["kB/s"] = abbrname["KB/s"] =
        fullname["kbytes/s"] = fullname["kbytes/sec"] =
            fullname["kbytes/second"] = fullname["kilobytes/s"] =
                fullname["kilobytes/sec"] = fullname["kilobytes/second"] =
                    fullname["kilobytes per second"] = 1000 * 8;
    // Kibibytes per second.
    abbrname["kiBps"] = abbrname["KiBps"] = abbrname["kiB/s"] =
        abbrname["KiB/s"] = fullname["kibytes/s"] = fullname["kibytes/sec"] =
            fullname["kibytes/second"] = fullname["kibibytes/s"] =
                fullname["kibibytes/sec"] = fullname["kibibytes/second"] =
                    fullname["kibibytes per second"] = 1024 * 8;
    // Megabits per second.
    abbrname["mbps"] = abbrname["Mbps"] = abbrname["mb/s"] = abbrname["Mb/s"] =
        fullname["mbits/s"] = fullname["mbits/sec"] = fullname["mbits/second"] =
            fullname["megabits/s"] = fullname["megabits/sec"] =
                fullname["megabits/second"] = fullname["megabits per second"] =
                    1000 * 1000;
    // Mebibits per second.
    abbrname["mibps"] = abbrname["Mibps"] = abbrname["mib/s"] =
        abbrname["Mib/s"] = fullname["mibits/s"] = fullname["mibits/sec"] =
            fullname["mibits/second"] = fullname["mebibits/s"] =
                fullname["mebibits/sec"] = fullname["mebibits/second"] =
                    fullname["mebibits per second"] = 1024 * 1024;
    // Megabytes per second.
    abbrname["mBps"] = abbrname["MBps"] = abbrname["mB/s"] = abbrname["MB/s"] =
        fullname["mbytes/s"] = fullname["mbytes/sec"] =
            fullname["mbytes/second"] = fullname["megabytes/s"] =
                fullname["megabytes/sec"] = fullname["megabytes/second"] =
                    fullname["megabytes per second"] = 1000 * 1000 * 8;
    // Mebibytes per second.
    abbrname["miBps"] = abbrname["MiBps"] = abbrname["miB/s"] =
        abbrname["MiB/s"] = fullname["mibytes/s"] = fullname["mibytes/sec"] =
            fullname["mibytes/second"] = fullname["mebibytes/s"] =
                fullname["mebibytes/sec"] = fullname["mebibytes/second"] =
                    fullname["mebibytes per second"] = 1024 * 1024 * 8;
    // Gigabits per second.
    abbrname["gbps"] = abbrname["Gbps"] = abbrname["gb/s"] = abbrname["Gb/s"] =
        fullname["gbits/s"] = fullname["gbits/sec"] = fullname["gbits/second"] =
            fullname["gigabits/s"] = fullname["gigabits/sec"] =
                fullname["gigabits/second"] = fullname["gigabits per second"] =
                    1000 * 1000 * 1000;
    // Gibibits per second.
    abbrname["gibps"] = abbrname["Gibps"] = abbrname["gib/s"] =
        abbrname["Gib/s"] = fullname["gibits/s"] = fullname["gibits/sec"] =
            fullname["gibits/second"] = fullname["gibibits/s"] =
                fullname["gibibits/sec"] = fullname["gibibits/second"] =
                    fullname["gibibits per second"] = 1024 * 1024 * 1024;
    // Gigabytes per second.
    abbrname["gBps"] = abbrname["GBps"] = abbrname["gB/s"] = abbrname["GB/s"] =
        fullname["gbytes/s"] = fullname["gbytes/sec"] =
            fullname["gbytes/second"] = fullname["gigabytes/s"] =
                fullname["gigabytes/sec"] = fullname["gigabytes/second"] =
                    fullname["gigabytes per second"] = 1000LL * 1000LL * 1000LL
                        * 8LL;
    // Gibibytes per second.
    abbrname["giBps"] = abbrname["GiBps"] = abbrname["giB/s"] =
        abbrname["GiB/s"] = fullname["gibytes/s"] = fullname["gibytes/sec"] =
            fullname["gibytes/second"] = fullname["gibibytes/s"] =
                fullname["gibibytes/sec"] = fullname["gibibytes/second"] =
                    fullname["gibibytes per second"] = 1024LL * 1024LL * 1024LL
                        * 8;
    // Terabits per second.
    abbrname["tbps"] = abbrname["Tbps"] = abbrname["tb/s"] = abbrname["Tb/s"] =
        fullname["tbits/s"] = fullname["tbits/sec"] = fullname["tbits/second"] =
            fullname["terabits/s"] = fullname["terabits/sec"] =
                fullname["terabits/second"] = fullname["terabits per second"] =
                    1000LL * 1000LL * 1000LL * 1000LL;
    // Tebibits per second.
    abbrname["gibps"] = abbrname["Gibps"] = abbrname["gib/s"] =
        abbrname["Gib/s"] = fullname["tebits/s"] = fullname["tebits/sec"] =
            fullname["tebits/second"] = fullname["tebibits/s"] =
                fullname["tebibits/sec"] = fullname["tebibits/second"] =
                    fullname["tebibits per second"] = 1024LL * 1024LL * 1024LL
                        * 1024LL;
    // Terabytes per second.
    abbrname["tBps"] = abbrname["TBps"] = abbrname["tB/s"] = abbrname["TB/s"] =
        fullname["tbytes/s"] = fullname["tbytes/sec"] =
            fullname["tbytes/second"] = fullname["terabytes/s"] =
                fullname["terabytes/sec"] = fullname["terabytes/second"] =
                    fullname["terabytes per second"] = 1000LL * 1000LL * 1000LL
                        * 1000LL * 8LL;
    // Tebibytes per second.
    abbrname["tiBps"] = abbrname["TiBps"] = abbrname["tiB/s"] =
        abbrname["TiB/s"] = fullname["tibytes/s"] = fullname["tibytes/sec"] =
            fullname["tibytes/second"] = fullname["tebibytes/s"] =
                fullname["tebibytes/sec"] = fullname["tebibytes/second"] =
                    fullname["tebibytes per second"] = 1024LL * 1024LL * 1024LL
                        * 8;
    // Petabits per second.
    abbrname["pbps"] = abbrname["Pbps"] = abbrname["pb/s"] = abbrname["Pb/s"] =
        fullname["pbits/s"] = fullname["pbits/sec"] = fullname["pbits/second"] =
            fullname["petabits/s"] = fullname["petabits/sec"] =
                fullname["petabits/second"] = fullname["petabits per second"] =
                    1000LL * 1000LL * 1000LL * 1000LL * 1000LL;
    // Pebibits per second.
    abbrname["gibps"] = abbrname["Gibps"] = abbrname["gib/s"] =
        abbrname["Gib/s"] = fullname["pebits/s"] = fullname["pebits/sec"] =
            fullname["pebits/second"] = fullname["pebibits/s"] =
                fullname["pebibits/sec"] = fullname["pebibits/second"] =
                    fullname["pebibits per second"] = 1024LL * 1024LL * 1024LL
                        * 1024LL * 1024LL;
    // Petabytes per second.
    abbrname["tBps"] = abbrname["TBps"] = abbrname["tB/s"] = abbrname["TB/s"] =
        fullname["tbytes/s"] = fullname["tbytes/sec"] =
            fullname["tbytes/second"] = fullname["petabytes/s"] =
                fullname["petabytes/sec"] = fullname["petabytes/second"] =
                    fullname["petabytes per second"] = 1000LL * 1000LL * 1000LL
                        * 1000LL * 1000LL * 8LL;
    // Pebibytes per second.
    abbrname["piBps"] = abbrname["PiBps"] = abbrname["piB/s"] =
        abbrname["PiB/s"] = fullname["pibytes/s"] = fullname["pibytes/sec"] =
            fullname["pibytes/second"] = fullname["pebibytes/s"] =
                fullname["pebibytes/sec"] = fullname["pebibytes/second"] =
                    fullname["pebibytes per second"] = 1024LL * 1024LL * 1024LL
                        * 8;
  }

  int64_t
  get_bandwidth(const char *value, bool &errorflag)
  {
    // This map does case-sensitive matching of abbreviated names
    static std::map<std::string, int64_t> abbrname;
    // and this map does case-insenstive matching of non-abbreviated names.
    static std::map<std::string, int64_t> fullname;
    // Populate the maps with the multipliers for all recognized units.
    // We store the multipliers in bits/second to make everything integers.
    // Bits per second.
    if (abbrname.empty() || fullname.empty())
      populate_bandwidth_names(abbrname, fullname);
    // Go.
    char *endptr = NULL;
    int64_t multiplier = 1;
    double val = strtod(value, &endptr);
    errno = 0;
    // Figure out whether we have a value.
    if (*value == '\0' || endptr == value)
      {
        errorflag = true;
        std::cerr << "Invalid bandwidth value: " << value << "\n";
      }
    if (errno == EINVAL)
      {
        errorflag = true;
        std::cerr << "Invalid characters in value: " << value << "\n";
      }
    if (errno == ERANGE)
      {
        errorflag = true;
        std::cerr << "Value outside 64-bit integer range: " << value << "\n";
      }
    // This is where we may have the units appended to the input value.
    if (!errorflag)
      {
        while (*endptr == ' ')
          ++endptr;
        if (*endptr != '\0')
          {
            std::string units(endptr);
            std::string lcunits = units;
            for (unsigned i = 0; i < lcunits.size(); ++i)
              lcunits[i] = tolower(lcunits[i]);
            if (abbrname.find(units) != abbrname.end())
              multiplier = abbrname[units];
            else if (fullname.find(lcunits) != fullname.end())
              multiplier = fullname[units];
            else
              {
                errorflag = true;
                std::cerr << "No match for the bandwidth units " << units
                    << "\n";
              }
          }
        if (!errorflag)
          {
            // Now expand the value.
            val = multiply64(val, multiplier, errorflag);
            val /= 8; // because the simulator works in bytes/sec.
          }
      }
    return val;
  }


  void
  populate_timestamp_names(std::map<std::string, int64_t> &value)
  {
    // Accept ps, ns, us, ms, and s with various (lower-case) forms
    // Store time in picoseconds.
    value["ps"] = value["psec"] = value["pseconds"] = value["picoseconds"] = 1;
    value["ns"] = value["nsec"] = value["nseconds"] = value["nanoseconds"] = 1000;
    value["us"] = value["usec"] = value["useconds"] = value["microseconds"] = 1000
        * 1000;
    value["ms"] = value["msec"] = value["mseconds"] = value["milliseconds"] =
        1000LL * 1000LL * 1000LL;
    value["s"] = value["sec"] = value["seconds"] = 1000LL * 1000LL * 1000LL
        * 1000LL;
  }

  /// Get a timestamp possiblly suffixed with any of the identifiers
  /// psec, nsec, usec, msec, sec, ps, ns, us, ms, s
  timestamp
  get_timestamp(const char *value, bool &errorflag)
  {
    static std::map<std::string, int64_t> mulmap;
    if (mulmap.empty())
      populate_timestamp_names(mulmap);



    char *endptr = NULL;
    int64_t multiplier = 1;
    double val = strtod(value, &endptr);
    errno = 0;
    // Figure out whether we have a value.
    if (*value == '\0' || endptr == value)
      {
        errorflag = true;
        std::cerr << "Invalid time value: " << value << "\n";
      }
    if (errno == EINVAL)
      {
        errorflag = true;
        std::cerr << "Invalid characters in value: " << value << "\n";
      }
    if (errno == ERANGE)
      {
        errorflag = true;
        std::cerr << "Value outside 64-bit integer range: " << value << "\n";
      }
    // This is where we may have the units appended to the input value.
    if (!errorflag)
      {
        while (*endptr == ' ')
          ++endptr;
        if (*endptr != '\0')
          {
            std::string units(endptr);
            for (unsigned i = 0; i < units.size(); ++i)
              units[i] = tolower(units[i]);
            if (mulmap.find(units) != mulmap.end())
              multiplier = mulmap[units];
            else
              {
                errorflag = true;
                std::cerr << "Invalid time units: " << units << "\n";
              }
          }
        if (!errorflag)
          {
            // Now expand the value.
            val = multiply64(val, multiplier, errorflag);
          }
      }
    return timestamp::exact_psec(val);
  }

}

