/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_FLOWER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_FLOWER_H_INCLUDED

#include <sstmac/common/debug.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <sstmac/common/ptr_type.h>

namespace sstmac { namespace hw  {

  typedef int edge_id_t;
  typedef int flow_id_t;

  /**
   * Abstract base class that represents network traffic as a set of
   * concurrent flows. Different specializations schedule the flows in
   * different ways. A flower object is used by sharedcircuit to
   * determine when network traffic commences and completes.
   */
  class flower : public ptr_type {


  private:
    topology::ptr net_;

  protected:
    static debug dbg_;

    flower(const topology::ptr &net): net_(net) {
      if(!dbg_.is_real()){
          dbg_ = debug("<debug> flower");
          logger::register_logger(&dbg_);
      }
    }

  public:
    virtual std::string to_string() const{
      return "flower";
    }

    typedef boost::intrusive_ptr<flower> ptr;

    virtual ~flower() throw();

    /// Start a flow through the graph. The edge identifiers are indices into 
    /// the edge_bw_ array.
    virtual flow_id_t start_flow(timestamp tstart, const std::vector<edge_id_t> &edges, long nbyte) = 0;

    /// Stop a flow.
    virtual void stop_flow(timestamp t, flow_id_t f) = 0;

    /// The next time that any flow will stop. It is an error to call this if
    /// there is no next stopping flow.
    virtual timestamp next_stopping_time() = 0;

    /// The identifier of the next flow that will stop. A return value of zero
    /// indicates that there is no next stopping flow.
    virtual flow_id_t next_stopping_flow() = 0;

    /// Accesses the network used to construct this object.
    const topology::ptr &net() const { return net_; }
  };

} } // end of namespace sstmac

#endif
