/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/fullnetwork.h>
#include <sstmac/common/eventhandler.h>
//#include <sstmac/simulator/eventmanager.h>
#include <sstmac/hardware/network/congestion/constantlatency.h>
#include <sstmac/hardware/network/topology/crossbar.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <iostream>

namespace sstmac { namespace hw 
  {



  //
  // Construct a fullnetwork object with the given properties
  //
  fullnetwork::fullnetwork(const latencymodel::ptr &latency, bandwidth bw,
      int maxsize, const power_config::ptr &pwrcfg) : interconnect(crossbar::construct(maxsize)),
      latency_(latency), bandwidth_(bw), size_(maxsize)
  {
    SSTMAC_DEBUG << "fullnetwork(" << latency->to_string() << ", " << bw << ", "
        << maxsize << ")\n";
    if (!latency)
      throw nullerror("fullnetwork:  NULL latency model pointer.");
    if (bandwidth_.bits_per_second() <= 0)
      throw valueerror("fullnetwork:  Bandwidth must be greater than zero.");


    pwr_config_ = pwrcfg;

    if(pwrstat_->is_active())
      pwrstat_->record(pwr_config_->get_rtr_static());
  }


  //
  // Construct,
  //
  fullnetwork::ptr fullnetwork::construct(const timestamp &latency,
      double bandw, int maxsize)
  {
    if (latency < timestamp(0))
      {
        throw valueerror("fullnetwork:  Latency must not be less than zero.");
      }
    constantlatency::ptr lat = constantlatency::construct(latency);
    bandwidth bw(bandw, bandwidth::Bps);
    return construct(lat, bw, maxsize);
  }

  //
  // Construct.
  //
  fullnetwork::ptr fullnetwork::construct(
      const latencymodel::ptr &lat, bandwidth bw, int maxsize)
  {
    fullnetwork::ptr retval(
        new fullnetwork(lat, bw, maxsize, power_config::construct()));

    return retval;
  }

  //
  // Exit stage left.
  //
  fullnetwork::~fullnetwork() throw ()
    {
    }

  //
  // The maximum number of nodes possible in this simulator.
  //
  int64_t fullnetwork::slots() const
  {

    return size_;
  }


  //
  // Send the given message at the given time.
  //
  void fullnetwork::send(const timestamp &time,
      const sst_message::ptr &msg,
      const eventhandler::ptr &sendhandler,
      const eventhandler::ptr &recvhandler)
  {
    int bytes = msg->get_byte_length();

    SSTMAC_DEBUG << "fullnetwork::send(" << time << ", " << bytes << ", "
        << msg->fromaddr_ << ", " << msg->toaddr_ << ", " << sendhandler
        << ", " << recvhandler << ")\n";
    if (sendhandler == 0)
      {
        throw nullerror("fullnetwork::send:  null sendhandler pointer.");
      }
    if (recvhandler == 0)
      {
        throw nullerror("fullnetwork::send:  null recvhandler pointer.");
      }



    // Somewhat erroneously, assume that the send completes the moment it is
    // initiated (should probably be at a later time -- it must take some amount
    // of time to put data on the wire).
    sst_message::ptr sendack = msg->clone(sst_message::SEND_ACK);
    schedule(time, sendhandler, sendack);
    // The receive operation is congestion-free.
    data payload(bytes, data::Bytes);
    timestamp lat = latency_->latency(payload);
    timestamp recvtime = (time + lat + payload / bandwidth_);
    schedule(recvtime, recvhandler, msg);

    if(energystat_->is_active()){
        bandwidthstat_->record(time, bytes, recvtime - time);
        energystat_->record(time, pwr_config_->get_rtr_hop_per_byte() * bytes, recvtime);
    }
  }

  } } // end of namespace sstmac
