/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_KINETICBANDWIDTH_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_KINETICBANDWIDTH_H_INCLUDED

#include <sstmac/hardware/network/congestion/bandwidthmodel.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/common/ptr_type.h>

namespace sstmac { namespace hw  {

  /**
   * Use a simple kinetic model (similar to Michaelis-Menten) to estimate
   * bandwidth.  f(x) = (a * x) / (b + x).  This seems to do a surpisingly
   * good job estimating bandwidths both for on-node and off-node 
   * communications.
   */
  class kineticbandwidth : public bandwidthmodel {
    /// The coefficients for our kinetic model.
    bandwidth a_;
    data b_;

    /// Zoom.
    kineticbandwidth(bandwidth a, data b) :
      a_(a), b_(b)
    {}

  public:
    typedef boost::intrusive_ptr<kineticbandwidth> ptr;

    /// Create a new object for kinetic bandwidth estimates.
    /// \param a the bandwidth for very large transfers, given in bytes/psec
    /// \param b measure of bandwidth for very small transfers, given in bytes
    static kineticbandwidth::ptr construct(bandwidth a, data b) {
      return kineticbandwidth::ptr(new kineticbandwidth(a, b));
    }

    /// Goodbye.
    virtual ~kineticbandwidth() {}

    /// What sort of bandwidth are we talking about?
    virtual bandwidth bw(data payload) const {
      // Unfortunately there is a very real risk of integer overflow here.
      // Let's punt for now and just do things in double precision.
      double scaling = payload.get_data(data::B)/(b_ + payload).get_data(data::B);
      return (a_ * scaling);
    }
  };

} } // end of namespace sstmac.

#endif
