/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_SHAREDCIRCUIT_STOPPER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_SHAREDCIRCUIT_STOPPER_H_INCLUDED

#include <sstmac/hardware/network/congestion/flower.h>
#include <sstmac/hardware/network/congestion/sharedcircuit.h>
#include <sstmac/common/eventhandler.h>
//#include <sstmac/simulator/eventmanager.h>
#include <sstmac/common/errors.h>
#include <map>
#include <set>

namespace sstmac { namespace hw 
{

  /**
   * The sharedcircuit::stopper takes care of stopping flows that
   * have completed and signaling the sender and receiver.
   */
  class sharedcircuit::stopper : public eventhandler
  {
    /// Private nested type to handle requests.
    struct registration
    {
      timestamp recv_delay;
      eventhandler::ptr sender, recver;
      sst_message::ptr deliver_;
      void
      init(const timestamp &rdelay, const sst_message::ptr &msg,
          const eventhandler::ptr &ss,
          const eventhandler::ptr &rr)
      {
        recv_delay = rdelay;
        sender = ss;
        recver = rr;
        deliver_ = msg;
      }
      void
      notify(const sharedcircuit::ptr &par)
      {
        sst_message::ptr sendack = deliver_->clone(
            sst_message::SEND_ACK);
        par->schedule(par->now(), sender, sendack);
        par->schedule(par->now() + recv_delay, recver, deliver_);
      }
    };

    /// Make sure we don't stop the same id twice.
    std::set<flow_id_t> stopped_;



    sharedcircuit::ptr parent_;

    /// The flow management object.
    flower::ptr flower_;

    /// The next flow to stop.  This may get reassigned during update(...).
    flow_id_t next_;
    /// The time at which the next flow should stop.
    /// Any incoming event notifications that precede this time get ignored.
    timestamp stoptime_;

    /// Keep track of message requests.
    typedef std::map<flow_id_t, registration> reg_t;
    reg_t reg_;

  protected:
    /// Sliding down the razor blade of life.
    stopper(const sharedcircuit::ptr &par,
        const flower::ptr &flower) :
        parent_(par), flower_(flower), next_(-1), stoptime_(-1)
    {
    }



  public:
    typedef boost::intrusive_ptr<stopper> ptr;

    /// Get a stopper.
    static stopper::ptr
    construct(const sharedcircuit::ptr &par,
        const flower::ptr &flower)
    {
      stopper::ptr self(new stopper(par, flower));

      return self;
    }

    /// Unstoppable.
    virtual
    ~stopper() throw ()
    {
    }

    /// We may have a new "first flow to finish" -- if so reschedule this event.
    void
    update(flow_id_t flow, const timestamp &newstop)
    {
      if (!((flow == next_) && (newstop == stoptime_)))
        {
          next_ = flow;
          stoptime_ = newstop;
          parent_->schedule(stoptime_, ptr(this));
        }
    }

    /// Register a new flow with the stopper.
    void
    add(flow_id_t flow, const timestamp &recv_delay,
        const sst_message::ptr &msg,
        const eventhandler::ptr &sendhandler,
        const eventhandler::ptr &recvhandler)
    {
      if (reg_.find(flow) != reg_.end())
        throw illformederror("sharedcircuit::stopper::add:  Attempt to "
            "register a pre-existing flow id.");
      reg_[flow].init(recv_delay, msg, sendhandler, recvhandler);
    }

    /// Our time has come.
    virtual void
    handle(const sst_message::ptr &msg)
    {
      if (parent_->now() != stoptime_)
        {
          // Ignore this event -- it has been rescheduled
          // (this is how we get singleton behavior).
          return;
        }
      if (stopped_.find(next_) == stopped_.end())
        {
          flower_->stop_flow(parent_->now(), next_);
          stopped_.insert(next_);
          // Notify the sender and receiver.
          reg_t::iterator it = reg_.find(next_);
          if (it == reg_.end())
            {
              throw illformederror(
                  "sharedcircuit::stopper::handle:  Stopping a "
                      "flow that was not previously registered.");
            }
          it->second.notify(parent_);
          reg_.erase(it);
        }
      // Reschedule
      if (flower_->next_stopping_flow() != 0)
        {
          next_ = -1; // we don't have a flow to stop right now.
          this->update(flower_->next_stopping_flow(),
              flower_->next_stopping_time());
        }
    }

    virtual std::string
    to_string() const
    {
      return "sharedcircuit::stopper";
    }
  };

} } // end of namespace sstmac

#endif
