/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// transfer_time.h: Base class and specializations for transfer time models.
//
// Author: Curtis Janssen <cljanss@sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_TRANSFER_TIME_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_TRANSFER_TIME_H_INCLUDED

#include <stdint.h>
#include <iostream>
#include <sstmac/common/ptr_type.h>

namespace sstmac { namespace hw  {

  class transfer_time : public ptr_type {
  protected:
    transfer_time() {}
  public:
    virtual std::string to_string() const{
            return "transfer_time";
          }

    typedef boost::intrusive_ptr<transfer_time> ptr;

    virtual ~transfer_time();
    virtual double time(int64_t nbyte) = 0;
    virtual void print(std::ostream &o=std::cout) = 0;
  };

  class latency_rbandwidth_time: public transfer_time {
  protected:
    latency_rbandwidth_time() {}
  public:
    double latency;
    double rbandwidth;

    static transfer_time::ptr
      construct(double latency, double rbandwidth);

    /// Returns latency + nbyte * rbandwidth.
    double time(int64_t nbyte);
    void print(std::ostream &o=std::cout);
  };

  class constant_time: public transfer_time {
  protected:
    constant_time() {}
  public:
    double constant;

    static transfer_time::ptr
      construct(double constant);

    /// Returns constant
    double time(int64_t nbyte);
    void print(std::ostream &o=std::cout);
  };

  class rbandwidth_time: public transfer_time {
  protected:
    rbandwidth_time() {}
  public:
    double rbandwidth;

    static transfer_time::ptr
      construct(double rbandwidth);

    /// Returns rbandwidth * nbyte
    double time(int64_t nbyte);
    void print(std::ostream &o=std::cout);
  };

  class no_time: public transfer_time {
  protected:
    no_time() {}
  public:
    static transfer_time::ptr
      construct();
    // This always returns zero.
    double time(int64_t nbyte);
    void print(std::ostream &o=std::cout);
  };

  std::ostream& operator<<(std::ostream&s,const transfer_time::ptr&);

} } //end of namespace sstmac

#endif
