/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// product.cc: Implementation of product networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <sstmac/hardware/network/topology/product.h>
#include <ostream>
#include <iterator>

using std::vector;
using std::set;

namespace sstmac { namespace hw  {

  

  template <class iter>
  static void
  vertex_direct_product_accum(const iter &v1b,
                              const iter &v1e,
                              const iter &v2b,
                              const iter &v2e,
                              std::set<vertex> &result)
  {
    vector<int> v;
    for (iter i = v1b;
         i != v1e;
         i++) {
      for (iter j = v2b;
           j != v2e;
           j++) {
        v.clear();
        v.insert(v.end(), (*i).v().begin(), (*i).v().end());
        v.insert(v.end(), (*j).v().begin(), (*j).v().end());
        result.insert(vertex(v));
      }
    }
  }

  template< class v_iter, class e_iter>
  static void
  edge_vertex_direct_product_accum(const graph &g1,
                                   const v_iter &v1b, const v_iter &v1e,
                                   const e_iter &e1b, const e_iter &e1e,
                                   const graph &g2,
                                   const v_iter &v2b, const v_iter &v2e,
                                   const e_iter &e2b, const e_iter &e2e,
                                   std::set<edge> &result)
  {
    vector<int> v0, v1;
    for (v_iter i = v1b; i != v1e; i++) {
      for (e_iter j = e2b; j != e2e; j++) {
        const vertex &j_from = g2.index_to_vertex(j->from());
        const vertex &j_to = g2.index_to_vertex(j->to());
        v0.clear();
        v0.insert(v0.end(), (*i).v().begin(), (*i).v().end());
        v0.insert(v0.end(), j_from.v().begin(), j_from.v().end());
        v1.clear();
        v1.insert(v1.end(), (*i).v().begin(), (*i).v().end());
        v1.insert(v1.end(), j_to.v().begin(), j_to.v().end());
        result.insert(edge(vertex(v0),vertex(v1)));
      }
    }
    for (e_iter i = e1b; i != e1e; i++) {
      const vertex &i_from = g1.index_to_vertex(i->from());
      const vertex &i_to = g1.index_to_vertex(i->to());
      for (v_iter j = v2b; j != v2e; j++) {
        v0.clear();
        v0.insert(v0.end(), i_from.v().begin(), i_from.v().end());
        v0.insert(v0.end(), (*j).v().begin(), (*j).v().end());
        v1.clear();
        v1.insert(v1.end(), i_to.v().begin(), i_to.v().end());
        v1.insert(v1.end(), (*j).v().begin(), (*j).v().end());
        result.insert(edge(vertex(v0),vertex(v1)));
      }
    }
  }

  static void
  edge_vertex_direct_product_accum2(const topology::ptr &n1,
                                    const topology::ptr &n2,
                                    std::set<edge> &result)
  {
    vector<int> v0, v1;
    for (set<vertex>::const_iterator i = n1->nodes().begin();
         i != n1->nodes().end();
         i++) {
      vertex_index ivi = n1->g().vertex_to_index(*i);
      vertex_index isw1 = n1->node_to_switch(ivi);
      const vertex &sw1 = n1->g().index_to_vertex(isw1);
      for (std::set<vertex>::const_iterator j = n2->nodes().begin();
           j != n2->nodes().end();
           j++) {
        vertex_index j_nd_index = n2->g().vertex_to_index(*j);
        vertex_index j_sw_index = n2->node_to_switch(j_nd_index);
        const vertex &j_from = *j;
        const vertex &j_to = n2->g().index_to_vertex(j_sw_index);
        v0.clear();
        v0.insert(v0.end(), (*i).v().begin(), (*i).v().end());
        v0.insert(v0.end(), j_from.v().begin(), j_from.v().end());
        v1.clear();
        v1.insert(v1.end(), sw1.v().begin(), sw1.v().end());
        v1.insert(v1.end(), j_to.v().begin(), j_to.v().end());
        result.insert(edge(vertex(v0),vertex(v1)));
        result.insert(edge(vertex(v1),vertex(v0)));
      }
    }
  }

  static void
  partition_edges(const topology::ptr &net, set<lwedge> &swsw)
  {
    for (sstmac::hw::graph::edge_iterator i = net->g().edges_begin();
         i != net->g().edges_end();
         i++) {
      lwedge e(*i);

      std::set<vertex>::iterator notfound = net->nodes().end();
      bool to_is_node
        = (net->nodes().find(net->g().index_to_vertex(e.to())) != notfound);
      bool from_is_node
        = (net->nodes().find(net->g().index_to_vertex(e.from())) != notfound);

      if (from_is_node && to_is_node) {
        throw std::logic_error("product: found a node to node edge");
      }
      else if (!from_is_node && !to_is_node) {
        swsw.insert(e);
      }
    }
  }

  void
  product::assemble(const topology::ptr& n1, const topology::ptr& n2)
  {
    n1_ = n1;
    n2_ = n2;
    const graph &g1 = n1->g();
    const graph &g2 = n2->g();

    if (n1->direct() != n2->direct()) {
      throw std::invalid_argument("product::assemble: "
                                  "both networks must be direct or not direct");
    }

    bool direct_network = n1->direct();
    set<vertex> switches1, switches2;

    set<vertex> vs;
    if (direct_network) {
      vertex_direct_product_accum(g1.vertices_begin(),g1.vertices_end(),
                                  g2.vertices_begin(),g2.vertices_end(),
                                  vs);
      nodes_ = vs;
    }
    else {
      // initialize the nodes
      vertex_direct_product_accum(n1->nodes().begin(), n1->nodes().end(), 
                                  n2->nodes().begin(), n2->nodes().end(),
                                  vs);
      
      nodes_ = vs;

      // initialize the switches
      std::insert_iterator<set<vertex> > insert_switches1(switches1,
                                                          switches1.begin());
      set_difference(g1.vertices_begin(), g1.vertices_end(),
                     n1->nodes().begin(), n1->nodes().end(),
                     insert_switches1);
      std::insert_iterator<set<vertex> > insert_switches2(switches2,
                                                          switches2.begin());
      set_difference(g2.vertices_begin(), g2.vertices_end(),
                     n2->nodes().begin(), n2->nodes().end(),
                     insert_switches2);
      vertex_direct_product_accum(switches1.begin(), switches1.end(),
                                  switches2.begin(), switches2.end(), vs);
    }

    set<edge> es;

    if (direct_network) {
      edge_vertex_direct_product_accum(g1,
                                       g1.vertices_begin(), g1.vertices_end(),
                                       g1.edges_begin(), g1.edges_end(),
                                       g2,
                                       g2.vertices_begin(), g2.vertices_end(),
                                       g2.edges_begin(), g2.edges_end(),
                                       es);
    }
    else {
      // partition the edges into switch-to-switch edges
      std::set<lwedge> sw_sw1;
      std::set<lwedge> sw_sw2;
      partition_edges(n1,sw_sw1);
      partition_edges(n2,sw_sw2);

      // form the switch-to-switch product network edges
      edge_vertex_direct_product_accum(g1,
                                       switches1.begin(), switches1.end(),
                                       sw_sw1.begin(), sw_sw1.end(),
                                       g2,
                                       switches2.begin(), switches2.end(),
                                       sw_sw2.begin(), sw_sw2.end(),
                                       es);

      // form the switch-to-node product network edges
      edge_vertex_direct_product_accum2(n1,n2,es);
    }

    g_.init(vs, es);
    init();
    if (!direct()) init_node_to_switch();
  }

  product::ptr product::construct() {
    return product::ptr(new product);
  }
  
  product::~product() {
  }

  int64_t
  product::diameter() const
  {
    if (direct()) return n1_->diameter() + n2_->diameter();
    // In the non-direct case, the links from the switches to
    // the nodes are double-counted, so subtract two.
    return n1_->diameter() + n2_->diameter() - 2;
  }

  double
  product::a() const
  {
    if (direct()) return n1_->a() + n2_->a();
    // In the non-direct case, the links from the switches to
    // the nodes are double-counted; however, the 0 length route
    // to myself is unchanged, so use the following:
    double ave_dir_other1 = n1_->n()*n1_->a()/(n1_->n()-1) - 2;
    double ave_dir_other2 = n2_->n()*n2_->a()/(n2_->n()-1) - 2;
    double ave_dir1 = (n1_->n()-1)*ave_dir_other1/n1_->n();
    double ave_dir2 = (n2_->n()-1)*ave_dir_other2/n2_->n();
    double ave_dir = ave_dir1 + ave_dir2;
    double ave_dir_other = n()*ave_dir/(n()-1);
    double ave_other = ave_dir_other + 2;
    double ave = (n()-1)*ave_other/n();
    return ave;
  }

  int64_t
  product::b() const
  {
    int64_t b1 = n1_->b() * n2_->n();
    int64_t b2 = n1_->n() * n2_->b();
    return b1>b2?b2:b1;
  }

  int64_t
  product::shortest_path_length(const vertex &a, const vertex &b) const
  {
    int64_t l1 = n1_->g().word_length();
    int64_t l2 = n2_->g().word_length();
    vertex a1(l1), a2(l2), b1(l1), b2(l2);
    for (int i=0; i<l1; i++) { a1[i] = a[i];    b1[i] = b[i]; }
    for (int i=0; i<l2; i++) { a2[i] = a[i+l1]; b2[i] = b[i+l1]; }

    int64_t w1len = n1_->shortest_path_length(a1,b1);
    int64_t w2len = n2_->shortest_path_length(a2,b2);

    int64_t result = w1len + w2len;

    // In the non-direct case, the links from the switches to the
    // nodes are double-counted if both walks are non-empty, so
    // subtract two in that case.
    if (!direct() && w1len > 0 && w2len > 0) result -= 2;

    return result;
  }

  void
  product::shortest_path(const vertex &a, const vertex &b, walk &w) const
  {
    if (a == b) {
      w.evec().resize(0);
      return;
    }
    
   
    int64_t l1 = n1_->g().word_length();
    int64_t l2 = n2_->g().word_length();
    vertex a1(l1), a2(l2), b1(l1), b2(l2);
    for (int i=0; i<l1; i++) { a1[i] = a[i];    b1[i] = b[i]; }
    for (int i=0; i<l2; i++) { a2[i] = a[i+l1]; b2[i] = b[i+l1]; }
    vertex a1sw(l1), a2sw(l2), b1sw(l1), b2sw(l2);

    walk w1; n1_->shortest_path(a1,b1,w1);
   
    walk w2; n2_->shortest_path(a2,b2,w2);
   
    
    int64_t path_length = w1.length()+w2.length();
    int correction = 0;
    // In the non-direct case, the links from the switches to the
    // nodes are double-counted if both component walks have nonzero
    // length, so subtract two in that case. Also, leave out the hops
    // involved nodes in w1 and w2.  The node to switch hops will be
    // added separately.
    if (!direct()) {
      if (w1.length() > 0 && w2.length() > 0) {
        path_length -= 2;
      }
      correction = 1;
      a1sw = n1_->g().index_to_vertex(n1_->node_to_switch(n1_->g().vertex_to_index(a1)));
      a2sw = n2_->g().index_to_vertex(n2_->node_to_switch(n2_->g().vertex_to_index(a2)));
      b1sw = n1_->g().index_to_vertex(n1_->node_to_switch(n1_->g().vertex_to_index(b1)));
      b2sw = n2_->g().index_to_vertex(n2_->node_to_switch(n2_->g().vertex_to_index(b2)));
    }
    else {
      a1sw = a1;
      a2sw = a2; 
      b1sw = b1;
      b2sw = b2;
    }

    int iedge = 0;
    w.evec().resize(path_length);

    if (!direct() && (w1.length() > 0 || w2.length() > 0)) {
      vertex_index from_index = g().vertex_to_index(a);
      w.evec()[iedge++].init(from_index,node_to_switch(from_index));
    }

    vector<int> from(l1+l2), to(l1+l2);
    for (int i=correction; i<w1.length()-correction; i++) {
      const std::vector<int> &from1
        = n1_->g().index_to_vertex(w1.e(i).from()).v();
      const std::vector<int> &to1
        = n1_->g().index_to_vertex(w1.e(i).to()).v();
      for (int j=0; j<l1; j++) {
        from[j] = from1[j];
        to[j]   = to1[j];
      }
      for (int j=0; j<l2; j++) {
        from[j+l1] = a2sw[j];
        to[j+l1]   = a2sw[j];
      }
      w.evec()[iedge++].init(g().vertex_to_index(from),g().vertex_to_index(to));
    }
    for (int i=correction; i<w2.length()-correction; i++) {
      const std::vector<int> &from2
        = n2_->g().index_to_vertex(w2.e(i).from()).v();
      const std::vector<int> &to2
        = n2_->g().index_to_vertex(w2.e(i).to()).v();
      for (int j=0; j<l1; j++) {
        from[j] = b1sw[j];
        to[j]   = b1sw[j];
      }
      for (int j=0; j<l2; j++) {
        from[j+l1] = from2[j];
        to[j+l1]   = to2[j];
      }
      w.evec()[iedge++].init(g().vertex_to_index(from),g().vertex_to_index(to));
    }

    if (!direct() && (w1.length() > 0 || w2.length() > 0)) {
      vertex_index to_index = g().vertex_to_index(b);
      w.evec()[iedge++].init(node_to_switch(to_index),to_index);
    }
   
  }

  void
  product::next_hop(vertex_index loc, vertex_index dst,
                    std::vector<lwedge> &nxt) const
  {
    nxt.clear();

    if (loc == dst) {
      return;
    }

    int64_t l1 = n1_->g().word_length();
    int64_t l2 = n2_->g().word_length();

    vertex loc1v(l1), dst1v(l1);
    vertex loc2v(l2), dst2v(l2);

    const vertex &locv = g().index_to_vertex(loc);
    const vertex &dstv = g().index_to_vertex(dst);

    for (int i=0; i<l1; i++) {
      loc1v.v()[i] = locv.v()[i];
      dst1v.v()[i] = dstv.v()[i];
    }

    for (int i=0; i<l2; i++) {
      loc2v.v()[i] = locv.v()[i+l1];
      dst2v.v()[i] = dstv.v()[i+l1];
    }

    vertex_index loc1(n1_->g().vertex_to_index(loc1v));
    vertex_index dst1(n1_->g().vertex_to_index(dst1v));
    vertex_index loc2(n2_->g().vertex_to_index(loc2v));
    vertex_index dst2(n2_->g().vertex_to_index(dst2v));

    std::vector<lwedge> nxt1, nxt2;

    n1_->next_hop(loc1,dst1,nxt1);
    n2_->next_hop(loc2,dst2,nxt2);

    vertex hop;
    hop.v().resize(g().word_length());

    if (!direct() && is_node(loc)) {
      // is loc is a node and loc != dst (tested above), the next hop
      // must always be a switch
      vertex_index sw = node_to_switch(loc);
      nxt.push_back(g().generate_lwedge(loc,sw));
      return;
    }
    
    nxt.reserve(nxt1.size()+nxt2.size());

    bool loc1_is_node = n1_->is_node(loc1);
    bool loc2_is_node = n2_->is_node(loc2);

    for (int i=0; i<l2; i++) hop.v()[i+l1] = locv.v()[i+l1];
    for (int i=0; i<nxt1.size(); i++) {
      if (direct() || (n1_->is_node(nxt1[i].to()) == loc2_is_node)) {
        const vertex &v = n1_->g().index_to_vertex(nxt1[i].to());
        for (int j=0; j<l1; j++) hop.v()[j] = v.v()[j];
        nxt.push_back(g().generate_lwedge(locv,hop));
      }
    }

    for (int i=0; i<l1; i++) hop.v()[i] = locv.v()[i];
    for (int i=0; i<nxt2.size(); i++) {
      if (direct() || (loc1_is_node == n2_->is_node(nxt2[i].to()))) {
        const vertex &v = n2_->g().index_to_vertex(nxt2[i].to());
        for (int j=0; j<l2; j++) hop.v()[j+l1] = v.v()[j];
        nxt.push_back(g().generate_lwedge(loc,hop));
      }
    }

  }

  bool
  product::vertex_symmetric() const
  {
    return n1_->vertex_symmetric() && n2_->vertex_symmetric();
  }

} } //end of namespace sstmac
