/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// ring.cc: Implementation of ring networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <sstream>
#include <sstmac/hardware/network/topology/ring.h>
#include <sstmac/hardware/network/topology/walk.h>

using std::set;
using std::vector;
using std::string;

namespace sstmac { namespace hw  {

  static void
  compute_vertices(int nx, set<vertex> &vertices, int id=-1)
  {
    std::vector<int> v;
    if (id < 0) v.resize(1);
    else { v.resize(2); v[1] = id; }
    for (int i=0; i<nx; i++) {
      v[0] = i;
      vertices.insert(vertex(v));
    }
  }

  // these hook switches together
  static void
  compute_switch_edges(int nx, set<edge> &edges, int id=-1)
  {
    std::vector<int> vi(1);
    std::vector<int> vo(1);
    if (id < 0) { vi.resize(1); vo.resize(1); }
    else { vi.resize(2); vi[1] = id; vo.resize(2); vo[1] = id; }
    for (int i=0; i<nx; i++) {
      vi[0] = i;
      vo[0] = i+1;
      if (vo[0] >= nx) vo[0] = 0;
      edges.insert(edge(vi,vo));
      vo[0] = i-1;
      if (vo[0] < 0) vo[0] = nx-1;
      edges.insert(edge(vi,vo));
    }
  }

  // these hook nodes to switches
  static void
  compute_node_edges(int nx, set<edge> &edges, int nndpersw)
  {
    std::vector<int> vi(2); // vi specifies a node
    std::vector<int> vo(2); // vo specifies a switch
    vo[1] = nndpersw;
    for (int i=0; i<nx; i++) {
      vi[0] = i;
      vo[0] = i;
      for (int nodeid = 0; nodeid < nndpersw; nodeid++) {
        vi[1] = nodeid;
        edges.insert(edge(vi,vo)); // node to switch
        edges.insert(edge(vo,vi)); // switch to node
      }
    }
  }

  // static sstmac::xmlclass::register_class reg("sstmac::ring",
  //                               sstmac::xmlclass::ctor_template<ring>);

  // ring::ring(TiXmlHandle h)
  // {
  //   int nx = 2;
  //   h.ToElement()->QueryIntAttribute("nnode",&nx);
  //   init_ring(nx);
  // }

  ring::ring(int nx, int nndpersw, int widtharg)
  {
    mesh_routing =0;
    direct_ = (nndpersw <= 0);
    nx_ = nx;
    nndpersw_ = nndpersw;
    width = widtharg;

    if (direct_) {
      compute_vertices(nx_, nodes_);
      set<edge> edges;
      compute_switch_edges(nx_, edges);
      g_.init(nodes_, edges);
    }
    else {
      for (int i=0; i<nndpersw; i++) {
        compute_vertices(nx_, nodes_, i);
      };
      std::set<vertex> all(nodes_);
      // insert the switch vertices
      compute_vertices(nx_, all, nndpersw);
      set<edge> edges;
      compute_switch_edges(nx_, edges, nndpersw);
      compute_node_edges(nx_, edges, nndpersw);
      g_.init(all, edges);
    }


    init();
    if (!direct_) init_node_to_switch();
  }

  ring::ptr ring::construct(int nsw,int nndpersw,int widtharg) {
    return ring::ptr(new ring(nsw,nndpersw,widtharg));
  }

  ring::~ring() {
  }

  void
  ring::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
  {
    vector<vertex_index> vertices;
    vertices.reserve(nx_+2);

    if (v1 == v2) {
      // the walk is empty when sending to myself
      w.init(vertices);
      return;
    }

    int a = v1[0];
    int b = v2[0];

    std::vector<int> v;

    if (direct_) {
      v.resize(1);
    }
    else {
      v.resize(2);
      // if this is a node, we must first get to a switch
      if (v1[1] < nndpersw_) {
        vertices.push_back(g().vertex_to_index(vertex(v1)));
      }
      // remaining hops (except maybe the last) are among switches
      v[1] = nndpersw_;
      if (a == b) {
        // This is a special case, because we must insert the hop
        // from the source to the switch. Otherwise, we will just
        // hop directly from node to node.
        v[0] = v1[0];
        vertices.push_back(g().vertex_to_index(v));
      }
    }

    int diff = b - a;
    // AP   update 
    if (mesh_routing) {
      if (diff > 0) {
        for (; a<=b; a++) {
          v[0] = a;
          vertices.push_back(g().vertex_to_index(v));
        }
      }
      else if (diff != 0) {
        for (; a>=b; a--) {
          v[0] = a;
          vertices.push_back(g().vertex_to_index(v));
        }
      }
    }
    else { 
      if (diff > 0 && diff > nx_/2) diff = diff-nx_;
      if (diff < 0 && diff < -nx_/2) diff = diff+nx_;
      
      if (diff > 0) {
        for (int i=0; i<=diff; i++) {
          v[0] = a+i;
          if (v[0] >= nx_) v[0] -= nx_;
          vertices.push_back(g().vertex_to_index(v));
        }
      }
      else if (diff != 0) {
        for (int i=0; i<=-diff; i++) {
          v[0] = a-i;
          if (v[0]<0) v[0] += nx_;
          vertices.push_back(g().vertex_to_index(v));
        }
      }
    } 

    // for direct networks where the target is a node, we must
    // insert a final hop from the switch to the node
    if (!direct_ && v2[1] < nndpersw_) {
      vertices.push_back(g().vertex_to_index(vertex(v2)));
    }

    w.init(vertices);
  }

  void
  ring::next_hop(vertex_index loc,
                 vertex_index dst,
                 std::vector<lwedge> &nxt) const
  {
    nxt.clear();

    if (loc == dst) {
      return;
    }

    const vertex &locv = g().index_to_vertex(loc);
    const vertex &dstv = g().index_to_vertex(dst);

    int a = locv.v()[0];
    int b = dstv.v()[0];
    int diff = b - a;

    vertex hop(g().word_length());

    // If this is not direct, handle the special cases of hops to or
    // from a node.
    if (!direct_) {
      if (locv.v()[1] < nndpersw_) {
        // loc is a node; next hop is a switch
        hop.v()[0] = locv.v()[0];
        hop.v()[1] = nndpersw_;
        nxt.push_back(g().generate_lwedge(loc,hop));
        return;
      }
      else if (locv.v()[0] == dstv.v()[0]) {
        // ready to hop from a switch to the destination node
        nxt.push_back(g().generate_lwedge(loc,dst));
        return;
      }
      hop[1] = nndpersw_;
    }
    if (!diff)
      return;

    if (mesh_routing) {
      if (diff>0) hop[0] = a + 1;
      else        hop[0] = a - 1;

      nxt.push_back(g().generate_lwedge(loc,hop));
    }
    else {
      if (2*diff == nx_ || 2*diff == -nx_) {
        // Either way gets us closer in this case.
        hop[0] = (a + 1)%nx_;
        nxt.push_back(g().generate_lwedge(loc,hop));
        hop[0] = (a + nx_ - 1)%nx_;
        nxt.push_back(g().generate_lwedge(loc,hop));
      }
      else { 
        if      (diff > 0 && 2*diff >  nx_) diff = diff-nx_;
        else if (diff < 0 && 2*diff < -nx_) diff = diff+nx_;
        
        if (diff>0) hop[0] = (a + 1)%nx_;
        else        hop[0] = (a + nx_ - 1)%nx_;
        
        nxt.push_back(g().generate_lwedge(loc,hop));
      }
    }
  }
  
  int64_t
  ring::diameter() const
  {
    int64_t result = n()/2;
    if (!direct_) result += 2;
    return result;
  }

  int64_t
  ring::b() const
  {
    if (nx_ < 2) return 0;
    if (nx_ == 2) return 2;
    return 4;
  }

  bool
  ring::vertex_symmetric() const
  {
    return true;
  }

  string
  ring::name() const
  {
    std::ostringstream ostr;
    ostr << "Ring" << "(" << nx_ << ",nps=" << nndpersw_ << ")";
    return ostr.str();
  }

  int
  ring::edge_width(edge_index e) const
  {
    const lwedge &lwe = g().index_to_edge(e);
    if (is_switch(lwe.from()) && is_switch(lwe.to())) return width;
    return 1;
  }

} } //end of namespace sstmac
