/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// network.h: Interface for networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>
#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TOPOLOGY_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TOPOLOGY_H_INCLUDED

#include <stdexcept>
#include <string>
#include <map>
#include <stdlib.h>
#include <sstmac/common/errors.h>
#include <sstmac/hardware/network/topology/graph.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/debug.h>
// #include <sstmac/util/xmlclass.h>

#include <sstmac/common/ptr_type.h>

namespace sstmac { namespace hw 
  {

  /** A network is represented by a graph.  A subset of the graph's vertices
 correspond to nodes. Other vertices may be switching elements. */
  class topology : public ptr_type /*: public sstmac::xmlclass */
  {
  private:
    typedef std::map<vertex_index, vertex_index> mapvv;
    mapvv node_to_switch_;
  protected:
    static debug dbg_;

    graph g_;
    std::set<vertex> nodes_;

    std::vector<vertex> inodes_;
    std::map<vertex, int64_t> nodesi_;

    /// The nodes_ set must be initialized before init() is called.
    void
    init();
    /** Calling this is optional. It only makes sense to call this if
         direct() == false. init() must be called before
         init_node_to_switch() is called. */
    void
    init_node_to_switch();

  protected:
    topology()
    {
      if(!dbg_.is_real()){
          dbg_ = debug("<debug> | topology");
          logger::register_logger(&dbg_);
      }
    }
    //network(TiXmlHandle) {}

  public:
    virtual std::string to_string() const{
      return "topology";
    }

    typedef boost::intrusive_ptr<topology> ptr;

    virtual
    ~topology();
    const graph &
    g() const
    {
      return g_;
    }
    /// Return the degree.
    int64_t d() const
    {
      return g_.degree();
    }
    /// Returns the number of nodes.
    int64_t n() const
    {
      return nodes_.size();
    }
    /// Returns the diameter of the graph.
    virtual int64_t
    diameter() const = 0;
    /// Returns the average distance to all nodes.
    virtual double
    a() const;
    /// Returns the minimum number of edges sliced by bisection.
    virtual int64_t
    b() const;
    /** Returns the number of switches.  If n() equals the order of the
         graph, this assumes that there is one switch per node and return
         n().  Otherwise g().n() - n() will be returned. */
    virtual int64_t
    nswitch() const;
    /** Returns true if this is a direct network. The default implementation
         returns true if the number of nodes equals the total number of vertices
         in the graph. */
    virtual bool
    direct() const;
    /** Returns the number of links.  Note that unidirectional links are
         assumed, so the actual number of links will be a factor of two too
         high if bidirectional links are used. */
    virtual int64_t
    nlink() const;

    /// Returns the vertex of the node with the given index.
    const vertex &
    index_to_node(int64_t i) const
    {
      return inodes_[i];
    }
    /// Converts a node vertex to its index.
    int64_t node_to_index(const vertex &v) const
    {
      std::map<vertex, int64_t>::const_iterator i = nodesi_.find(v);
      if (i == nodesi_.end())
        {
          i = nodesi_.begin();
          if (v.v().size() != i->first.v().size()){
              std::cout << "vertex size: " << v.v().size() << ", nodes size: " << i->first.v().size() << std::endl;
            throw invalidkeyerror(
                "network::node_to_index: vertex sizes don't match");
          }
          else
            throw std::runtime_error(
                "network::node_to_index: vertex not found");
        }
      return i->second;
    }

    /// Returns the length of the shortest path between the two vertices.
    virtual int64_t
    shortest_path_length(const vertex &v1, const vertex &v2) const;
    /// Return the shortest path between the two vertices as a walk.
    virtual void
    shortest_path(const vertex&v1, const vertex&v2, walk&w) const;

    /** Fills in the walk for the route between two node vertices.  The
         default implementation uses shortest path routing. */
    virtual void
    route(const vertex &v1, const vertex &v2, walk &w) const;
    /** Fills in the walk for the route between two node indices.  The
         default implementation uses shortest path routing. */
    virtual void
    route(const nodeaddress::ptr &n1,
        const nodeaddress::ptr &n2, walk &w) const;

    virtual void
    route(int n1, int n2, walk &w) const;

    /** Returns all edges that take us one hop from the current
         location and closer to the destination.
         @param loc The vertex for the current location.
         @param dst The destination vertex.
         @param nxt A vector of hops.
         The default implementation throws.
     */
    virtual void
    next_hop(vertex_index loc, vertex_index dst,
        std::vector<lwedge> &nxt) const;

    /** Return the width of the edge. This can be used to represent
         multiple independent communication channels along the same
         path. The default implementation returns 1. */
    virtual int
    edge_width(edge_index e) const;

    /// Return the set of vertices corresponding to the nodes.
    const std::set<vertex> &
    nodes() const
    {
      return nodes_;
    }

    /** Given a vertex_index to a node, return the vertex_index of the
         switch to which it is connected. This is only valid for non-direct
         networks and for networks where each node connects to only one
         switch. If the specialization of network has not called
         the init_node_to_switch member, then this will throw. */
    vertex_index
    node_to_switch(vertex_index nd) const;

    /** Returns true if this is a switch vertex. */
    virtual bool
    is_switch(vertex_index) const;

    /** Returns true if this is a node vertex. */
    virtual bool
    is_node(vertex_index) const;

    /// Returns true iff the network is vertex symmetric. */
    virtual bool
    vertex_symmetric() const;

    /// Returns a descriptive name of the graph.
    virtual std::string
    name() const;

    /** For each i in [0,v.size), randomly pair all nodes.  If nnodes is
         nonzero, then randomly pair that number of nodes from the graph.  Fill
         v[i] with the max number of times a given edge is used in a route
         between the pairs of nodes. */
    void
    stochastic_max_link_load(std::vector<int>&v, int64_t nnodes = 0,
        bool tight_packing = false) const;
  };

  } } //end of namespace sstmac

#endif
