/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/computescheduler.h>

namespace sstmac
{
  namespace hw
  {

    using namespace std;

    debug computescheduler::dbg_;

    computescheduler::computescheduler(int ncores, timestamp preempt) :
                n_cores_(ncores), preemption_(preempt)
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> computescheduler");
          logger::register_logger(&dbg_);
        }

      for (int i = 0; i < n_cores_; i++)
        {
          compute_slots_[i] = false;
        }
    }

    computescheduler::ptr
    computescheduler::construct(int cores, timestamp pre)
    {

      computescheduler::ptr ret(new computescheduler(cores, pre));

      return ret;
    }

    int
    computescheduler::find_open_core(){
      int core = -1;
      for (int i = 0; i < n_cores_; i++)
        {
          SSTMAC_DEBUG << "computescheduler::request_compute checking for core "
              << i << "\n";
          if (compute_slots_.find(i) == compute_slots_.end())
            {
              std::cout << "computescheduler:: core: " << i << ", num cores="
                  << n_cores_ << std::endl;
              throw ssterror(
                  "computescheduler::request_compute - can't find core in compute_slots");
            }
          if (!compute_slots_[i])
            {

              return i;
            }
        }

      return -1;
    }

    void
    computescheduler::schedule_computation(const timestamp& comp,
        const sst_message::ptr& to_deliver, eventhandler::ptr &handler)
    {
      int core = find_open_core();



      if (core < 0) //put in suspended queue
        {
          SSTMAC_DEBUG << "[t=" << now()
                    << "] -- computescheduler suspending compute request" << "\n";
          suspendedQ.push(compute_job::construct(to_deliver, comp, comp, handler));

        }
      else
        {
          SSTMAC_DEBUG << "[t=" << now()
                  << "] -- computescheduler: computing for time " << comp << "\n";

          timestamp actual = comp;
          if (preemption_.sec() > 0)
            {
              actual = timestamp(min(comp.sec(), preemption_.sec()));
            }
          compute_slots_[core] = true;
          slots_compute_[to_deliver] = compute_job::construct(to_deliver, actual,
              comp, handler, core);
          timestamp time = now() + actual;
          eventmodule::schedule(time, ptr(this), to_deliver);

          // opsys->block(data->get_key());
        }
    }

    void
       computescheduler::schedule_computation(compute_job::ptr jorb)
       {
         int core = find_open_core();

         if (core < 0) //put in suspended queue
           {
             SSTMAC_DEBUG << "computescheduler suspending compute request" << "\n";
             suspendedQ.push(jorb);

           }
         else
           {
             SSTMAC_DEBUG << "computescheduler: computing for time " << jorb->timeleft_ << "\n";

             timestamp actual = jorb->timeleft_;
             if (preemption_.sec() > 0)
               {
                 actual = timestamp(min(jorb->timeleft_.sec(), preemption_.sec()));
               }
             jorb->time_ = actual;

             compute_slots_[core] = true;
             slots_compute_[jorb->msg_] = jorb;
             timestamp time = now() + actual;
             eventmodule::schedule(time, ptr(this), jorb->msg_);

             // opsys->block(data->get_key());
           }
       }

    void
    computescheduler::handle(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "computescheduler::handle(" << ")"
          << "\n";

      compute_job::ptr jorb = slots_compute_[msg];
      slots_compute_.erase(msg);
      compute_slots_[jorb->core_] = false;

      if (!suspendedQ.empty())
        {
          SSTMAC_DEBUG << "computescheduler("
              << "): executing suspended kernel " << "\n";

          schedule_computation(suspendedQ.front()->time_, suspendedQ.front()->msg_,
              suspendedQ.front()->handle_);
          suspendedQ.pop();
        }

      if (jorb->time_ >= jorb->timeleft_)
        {

          SSTMAC_DEBUG << "computescheduler("
              << "): calling handle on a finished job " << "\n";
          jorb->handle_->handle(jorb->msg_);
        }else{

            SSTMAC_DEBUG << "computescheduler - computation was preempted, scheduling again \n";

            jorb->timeleft_ = timestamp(jorb->timeleft_.sec() - jorb->time_.sec());
            schedule_computation(jorb);
        }
    }

  }
} //end of namespace sstmac
