/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/actor/simplemailserver-mbox.h>
#include <sstmac/software/libraries/actor/simplemailserver-data.h>
//#include <sstmac/process/actormagic.h>
//#include <sstmac/simulator/tasklist.h>
//#include <sstmac/simulator/node.h>
//#include <sstmac/simulator/networkinterface.h>
//#include <sstmac/simulator/eventhandler.h>
#include <sstmac/common/errors.h>
#include <sstream>

namespace sstmac { namespace sw
{

  debug simplemailserver::dbg_;
  
  using boost::dynamic_pointer_cast;

  std::string
  simplemailserver::get_lib_name()
  {
    std::stringstream ss;
    ss << "simplemailserver" << id_.app_.id;
    return ss.str();
  }

  //
  // Private method to send data.
  //
  void
  simplemailserver::send(const actorid &source, const actorid &target,
      actormsg::ptr& msg, int64_t size)
  {
    SSTMAC_DEBUG << "simplemailserver::send - begin \n";
    if (!(source.is_real() && target.is_real()))
      throw valueerror("simplemailserver::send:  Not a valid source or target");
    //if (valid_nodes_.find(target.node()) == valid_nodes_.end())
    //  throw valueerror("simplemailserver::send:  Not a valid target node.");
    if (!msg)
      throw nullerror(
          "simplemailserver::send:  Unwilling to send null content");
    timestamp t(0);
    SSTMAC_DEBUG << "simplemailserver::send - make message...";
    data::ptr dd = data::construct(get_lib_name(), t, source, target,
        size, msg);
    SSTMAC_DEBUG << "done \n";
    dd->fromaddr_ = source.node();
    dd->toaddr_ = target.node();
    if (target.node() != os_->get_task_addr(id_))
      os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, dd);
    else
      this->handle(dd);

    SSTMAC_DEBUG << "simplemailserver::send - end \n";
  }

  //
  // Private constructor.
  //
  simplemailserver::simplemailserver(softwareid id) :
          library(id), next_child_(0)

  {

    if(!dbg_.is_real()){
         dbg_ = debug("<debug> actor | simplemailserver");
         logger::register_logger(&dbg_);
     }

    /*for (int64_t i = 0; i < tlist->size(); ++i) {
     taskid id(i);
     valid_nodes_.insert(tlist->node(id));
     }
     nodes_.reserve(valid_nodes_.size());
     nodes_.insert(nodes_.begin(), valid_nodes_.begin(), valid_nodes_.end());
     if (valid_nodes_.find(nid) == valid_nodes_.end())
     throw illformederror("simplemailserver:  'My' node id is not in the "
     "list of valid nodes.");*/
  }

  //
  // Get us a new, heap-allocated object.
  //
  simplemailserver::ptr
  simplemailserver::construct(softwareid id)
  {
    simplemailserver::ptr rv(new simplemailserver(id));

    return rv;
  }

  //
  // Clean up.
  //
  simplemailserver::~simplemailserver()
  {
  }



  //
  // Create a new mbox interface that can talk to this object.
  //
  simplembox::ptr
  simplemailserver::new_mbox()
  {
    mbox::ptr mb = mbox::construct(id_);
    mboxes_[mb->mypid()] = mb;
    os_->register_lib(mb);
    return mb;
  }

  void
  simplemailserver::register_mbox(mbox::ptr mb)
  {
    mboxes_[mb->mypid()] = mb;
  }

  //
  // Complete a receive.
  //
  void
  simplemailserver::handle(const sst_message::ptr &msg)
  {
    data::ptr amsg = dynamic_pointer_cast<data>(msg);
    if (!msg)
      throw valueerror("actorserver::handle_message:  Did not receive a valid "
          "actormsg payload");

    nodeaddress::ptr naddr = os_->get_task_addr(library::id_);
    if (msg->get_type() == sst_message::SEND_ACK)
      {
        SSTMAC_DEBUG << "simplemailserver[" << naddr->to_string()
                << "]: received send ack \n";
        return;
      }

    actorid dest = amsg->dest();
    mboxes_t::iterator it = mboxes_.find(dest);
    if (it != mboxes_.end())
      {
        // We have this client
        mbox::ptr mb = it->second;
        if (mb)
          {
            SSTMAC_DEBUG << "simplemailserver[" << naddr->to_string()
                  << "]: received message, sending to mbox " << dest.to_string()
                  << " \n";
            // and the client hasn't been garbage collected yet.
            mb->handle(amsg->source(), amsg->payload());
          }
      }
    else
      {
        SSTMAC_DEBUG << "simplemailserver[" << naddr->to_string()
            << "]: didn't find mbox headed for " << dest.to_string() << " among "
            << mboxes_.size() << " boxes " << "\n";
        if (dbg_.is_active())
          {
            for (it = mboxes_.begin(); it != mboxes_.end(); it++)
              {
                SSTMAC_DEBUG << "mbox: " << it->second->to_string() << "\n";
              }
          }
      }

  }

} } // end of namespace sstmac
