/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_comm/mpicommfactory.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpicommfactory-pending.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/software/libraries/mpi/mpiapi.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpicomm.h>
#include <sstmac/common/counter.h>
#include <sstmac/common/richpair.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/messages/valuepayload.h>
#include <iterator>

namespace sstmac
{
  namespace sw
  {

    using boost::dynamic_pointer_cast;
    using namespace std;

    debug mpicommfactory::dbg_;

    map<std::pair<appid, mpicommid>, tasklist::ptr> mpicommfactory::tlists_;
    mpicommid mpicommfactory::next_id_(1);

    //
    // Build comm_world using information retrieved from the environment.
    //
    mpicommfactory::mpicommfactory(appid aid, mpiapi::ptr parent) : parent_(parent), aid_(aid),
        mpirun_np_(0)
    {
      if(!dbg_.is_real()){
           dbg_ = debug("<debug> mpi | mpicommfactory");
           logger::register_logger(&dbg_);
       }
    }

    //
    // Build a new factory.
    //
    mpicommfactory::ptr
    mpicommfactory::construct(appid aid, mpiapi::ptr parent)
    {
      mpicommfactory::ptr retval(new mpicommfactory(aid, parent));
      return retval;
    }

    //
    // Goodbye.
    //
    mpicommfactory::~mpicommfactory()
    {
    }

    //
    // Initialize.
    //
    void
    mpicommfactory::init(environment::const_ptr& env, const mpiid &rank)
    {

      //mpirun_np_ = mpirun_np;
      next_id_.id = 1;

      mpirun_np_ = env->get_app_count();

      const mpicommid cid(0);
      pair<appid, mpicommid> index = make_pair<appid, mpicommid>(aid_, cid);

      if (tlists_.find(index) == tlists_.end())
        {
          std::vector<taskid> dumm;

          for (int i = 0; i < mpirun_np_; i++)
            {
              dumm.push_back(taskid(i));
            }

          tasklist::ptr tl = tasklist::construct(dumm);

          tlists_[index] = tl;
        }

      worldcomm_ = mpicomm(cid, rank, tlists_[index], env);

      std::vector<taskid> selfp;
      selfp.push_back(taskid(rank.id));

      tasklist::ptr sl = tasklist::construct(selfp);
      selfcomm_ = mpicomm(mpicommid(-1), rank, sl, env);
    }

    //
    // Get the world communicator.
    //
    const mpicomm&
    mpicommfactory::get_world()
    {

      return worldcomm_;
    }

    //
    // Get the self communicator.
    //
    const mpicomm&
    mpicommfactory::get_self()
    {

      return selfcomm_;
    }

    //
    // Duplicate a communicator.
    //
    mpicomm
    mpicommfactory::comm_dup(const mpicomm &caller)
    {
      //std::cerr << "mpicommfactory::comm_dup(" << caller << ").  Input comm id=" << caller.id() << ", size=" << caller.size() << "\n";
      std::vector<mpiid> participants;
      participants.reserve(caller.size().id);
      std::generate_n(std::back_inserter(participants), caller.size().id,
          counter<mpiid>(mpiid(0)));
      return this->comm_create(caller, participants);
    }

    //
    // Make the given mpiid refer to a newly created communicator.
    //
    mpicomm
    mpicommfactory::comm_create(const mpicomm &caller,
        const std::vector<mpiid> &participants)
    {

      //first reduce so we know everyone is on board, which is the necessary barrier
      payload::const_ptr load = valuepayload<int>::construct(next_id_.id);
      payload::const_ptr result;
      parent_->reduce(1, mpitype::mpi_int, mpiop::max, mpiid(0), caller, load, result);


      //redistribute it to everyone
      parent_->bcast(1, mpitype::mpi_int, mpiid(0), caller, result);

      valuepayload<int>::const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<int> >(result);

      mpicommid cid = mpicommid(casted->data());





      pair<appid, mpicommid> index = make_pair<appid, mpicommid>(aid_, cid);

      if(tlists_.find(index) == tlists_.end()){
          std::vector<taskid> dumm;

          SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: creating comm " << cid.id << " of size " << participants.size() << "\n";


          for(int i = 0; i < participants.size(); i++){
              dumm.push_back(caller.peer_task(participants[i]));
          }

          tlists_[index] = tasklist::construct(dumm);
      }else{
          //if(!tlists_[index]->equals(participants)){
          //    return mpicomm::comm_null;  // this is erroneous behavior, if the participants don't match for different tasks
          // }
      }

      //now find my rank
      mpiid newrank = mpiid(-1);
      mpiid oldrank = caller.rank();
      for(int i = 0; i < participants.size(); i++){
          if(participants[i].id == oldrank.id){
              newrank.id = i;
          }
      }

     // if(newrank.id == 0){
          next_id_.id = cid.id + 1;
          SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: incrementing next_id to " << next_id_.id << "\n";
     // }

      if(newrank.id >= 0)
        return mpicomm(cid, newrank, tlists_[index], worldcomm_.env_);
      else
        return mpicomm::comm_null;
    }

    //
    // MPI_Comm_split.
    //
    mpicomm
    mpicommfactory::comm_split(const mpicomm &caller, int color, int key)
    {
      SSTMAC_DEBUG << "mpicommfactory[" << worldcomm_.rank().id << "]::comm_split(" << caller.to_string() << ", "
     << color << ", " << key << ")" << "\n";

      if(key >= caller.size().id){
          throw rangeerror("key out of range of original caller in mpicommfactory::comm_split");
      }

      payload::const_ptr mine =  valuepayload<richpair<int, int> >::construct(std::make_pair<int, int>(color, key));
      std::vector<payload::const_ptr> result;

      int mynextcommid = -1;
            int mynewrank = -1;

      parent_->allgather(2, mpitype::mpi_int, 2 * caller.size().id, mpitype::mpi_int, caller, mine, result);


      std::map<int, vector<int> > colormap;
      for(int i = 0; i < result.size(); i++){
          valuepayload<richpair<int, int> >::const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<richpair<int, int> > >(result.at(i));


          richpair<int, int> p = casted->data();

          SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: split: creating color map: " << p.first << ", " << p.second << "\n";

          if(colormap[p.first].size() == 0){
              colormap[p.first].reserve(caller.size().id);
              colormap[p.first].resize(caller.size().id, -1);
          }
          colormap[p.first][i] = p.second;
      }

      std::map<int, vector<int> >::iterator it, end = colormap.end();


      std::vector<mpirequest_t> reqs;
      for(it = colormap.begin(); it != end; it++){
          vector<int> keys = it->second;

          //first reduce so we know everyone is on board, which is the necessary barrier
          payload::const_ptr load = valuepayload<int>::construct(next_id_.id);
          payload::const_ptr result;
          parent_->reduce(1, mpitype::mpi_int, mpiop::max, mpiid(0), caller, load, result);


          //redistribute it to everyone
          parent_->bcast(1, mpitype::mpi_int, mpiid(0), caller, result);

          valuepayload<int>::const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<int> >(result);

          mpicommid cid = mpicommid(casted->data());

          SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: split: considering : " << cid.id << "\n";

         // mpicommid cid = mpicommid(next_id_.id);

          if(it->first == color){
              mynextcommid = cid.id;
          }
          pair<appid, mpicommid> index = make_pair<appid, mpicommid>(aid_, cid);
          if(tlists_.find(index) == tlists_.end()){
              std::vector<taskid> dumm;

              for(int i = 0; i < keys.size(); i++){
                  if(keys[i] >= 0)
                    dumm.push_back(caller.peer_task(mpiid(keys[i])));
              }

              SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: split: creating comm " << cid.id << " of size " << dumm.size() << "\n";

              tlists_[index] = tasklist::construct(dumm);
          }

          next_id_.id = cid.id + 1;

      }


      pair<appid, mpicommid> index = make_pair<appid, mpicommid>(aid_, mpicommid(mynextcommid));
      if(tlists_.find(index) == tlists_.end()){
          throw rangeerror("mpicommfactory: did not find task list");
      }
      tasklist::ptr tlist = tlists_[index];

      for(int i = 0;i < tlist->size(); i++){
          if(tlist->at(i) == caller.my_task()){
              mynewrank = i;
          }
      }

      if(color >= 0)
        return mpicomm(index.second, mpiid(mynewrank), tlist, worldcomm_.env_);
      else
        return mpicomm::comm_null;



     /* valuepayload<richpair<int, int> >::const_ptr load = valuepayload<richpair<int, int> >::construct(std::make_pair<int, int>(color, key));
      std::vector<payload::const_ptr> result;

      int mynextcommid = -1;
      int mynewrank = -1;

      parent_->gather(1, mpitype::mpi_int, caller.size().id, mpitype::mpi_int, mpiid(0), caller, load, result);

      mpitag colltag = caller.next_collective_tag();

      if(caller.rank().id == 0){
          std::map<int, vector<int> > colormap;
          for(int i = 0; i < result.size(); i++){
              valuepayload<richpair<int, int> >::const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<richpair<int, int> > >(result.at(i));


              pair<int, int> p = casted->data();

              SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: split: creating color map: " << p.first << ", " << p.second << "\n";

              if(colormap[p.first].size() == 0){
                  colormap[p.first].reserve(caller.size().id);
                  colormap[p.first].resize(caller.size().id, -1);
              }
              colormap[p.first][i] = p.second;
          }

          std::map<int, vector<int> >::iterator it, end = colormap.end();


          std::vector<mpirequest_t> reqs;
          for(it = colormap.begin(); it != end; it++){
              vector<int> keys = it->second;

              SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: split: considering : " << next_id_.id << "\n";

              mpicommid cid = mpicommid(next_id_.id);

              if(it->first == color){
                  mynextcommid = next_id_.id;
              }
              pair<appid, mpicommid> index = make_pair<appid, mpicommid>(aid_, cid);
              if(tlists_.find(index) == tlists_.end()){
                  std::vector<taskid> dumm;

                  for(int i = 0; i < keys.size(); i++){
                      if(keys[i] >= 0)
                        dumm.push_back(caller.peer_task(mpiid(keys[i])));
                  }

                  SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: split: creating comm " << cid.id << " of size " << dumm.size() << "\n";

                  tlists_[index] = tasklist::construct(dumm);
              }

              payload::const_ptr nextid = valuepayload<int>::construct(next_id_.id);
              for(int i = 0; i < keys.size(); i++){
                  if(keys[i] >= 0){
                      mpirequest_t req;
                    parent_->isend(1, mpitype::mpi_int, mpiid(i), colltag, caller, req, nextid);
                    reqs.push_back(req);
                  }
              }

              next_id_.id++;

          }

          std::vector<const_mpistatus_t> stats;
          parent_->waitall(reqs, stats);

      }else{
          const_mpistatus_t stat;

          SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: split: at old rank " << caller.rank().id  << "..receiving new comm id \n";

          parent_->recv(1, mpitype::mpi_int, mpiid(0), colltag, caller, stat);

          valuepayload<int>::const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<int> >(stat->content());

          SSTMAC_DEBUG << "MPI comm factory[" << worldcomm_.rank().id << "]: split: at old rank " << caller.rank().id << ", I received mpicommid of " << casted->data() << "\n";
          mynextcommid = casted->data();
      }

      pair<appid, mpicommid> index = make_pair<appid, mpicommid>(aid_, mpicommid(mynextcommid));
      if(tlists_.find(index) == tlists_.end()){
          throw rangeerror("mpicommfactory: did not find task list");
      }
      tasklist::ptr tlist = tlists_[index];

      for(int i = 0;i < tlist->size(); i++){
          if(tlist->at(i) == caller.my_task()){
              mynewrank = i;
          }
      }

      if(color >= 0)
        return mpicomm(index.second, mpiid(mynewrank), tlist, worldcomm_.env_);
      else
        return mpicomm::comm_null;*/

    }

  }
} // end of namespace sstmac
