/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-recvrequest.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-sendrequest.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-proberequest.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/libraries/mpi/mpiserver.h>
#include <sstmac/software/process/appid.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <stdint.h>

namespace sstmac { namespace sw
{

  // Pointer types.
  using namespace std;
  
  debug mpiqueue::dbg_;

  bool
  mpiqueue::sortbyseqnum::operator()(const mpimessage::ptr &a,
      const mpimessage::ptr &b) const
  {
    return (a->seqnum() < b->seqnum());
  }

  inline void
  debug_message(const mpimessage::ptr &message)
  {
    /* dbg_ << message->to_string() << ": bytes=" << message->bytes() << " source_task="
     << message->source_task().to_string() << ", dest_task=" << message->dest_task().to_string()
     << ", count=" << message->count() << ", type=" << message->type().to_string()
     << ", source=" << message->source().to_string() << ", dest=" << message->dest().to_string()
     << ", tag=" << message->tag().to_string() << ", commid=" << message->commid()
     << ", seqnum=" << message->seqnum() << ", id=" << message->id()
     << ", ssend=" << message->ssend() << ", cat=" << message->cat() << "\n";*/
  }

  //
  // Hi there.
  //
  mpiqueue::mpiqueue(const mpiserver::ptr& server, taskid myid,
  // const shared_ptr<nodeinterface>& node,
      int64_t handsize, int64_t envelope, int64_t mintrans) :
    server_(server),
    //node_(node),
        handsize_(handsize), envelope_(envelope), mintrans_(mintrans), myid_(
            myid), next_id_(0)
  {
    if(!dbg_.is_real()){
         dbg_ = debug("<debug> mpi | mpiqueue");
         logger::register_logger(&dbg_);
     }

    if (server == 0)
      {
        throw nullerror("mpiqueue(...):  null server pointer.");
      }

    // need to debug dangling messages.
    next_id_ = messageid(myid_.id * int64_t(4000000000LL));
  }


  //
  // This is who I am.
  //
  //   mpiqueue::ptr mpiqueue::get_self() {
  //     return self_.lock();
  //   }

  //
  // Hello.
  //
  mpiqueue::ptr
  mpiqueue::construct(const mpiserver::ptr& server, taskid myid,
  // const shared_ptr<nodeinterface>& node,
      int64_t handsize, int64_t envelope, int64_t mintrans)
  {
    mpiqueue::ptr retval(new mpiqueue(server, myid,
    // node,
        handsize, envelope, mintrans));

    return retval;
  }

  //
  // Goodbye.
  //
  mpiqueue::~mpiqueue() throw ()
  {
  }

  //
  // Send data.
  //
  void
  mpiqueue::send(const mpirequest_t key, int count, const mpitype &type,
      const mpiid &dest, const mpitag &tag, const mpicomm &comm, bool rsend,
      bool ssend, mpimessage::category message_type, const
          eventhandler::ptr &completion, const payload::const_ptr& content)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::send(" << count << ", "
          << type.to_string() << ", " << dest.to_string() << ", "
          << tag.to_string() << ", " << comm.to_string() << ", " << rsend << ", "
          << ssend << ", " << (int) message_type << ", " << ")\n";

    taskid tid = comm.peer_task(dest);

    sendrequest::ptr req = sendrequest::construct(key,
        server_->get_lib_name(), mpiqueue::ptr(this), handsize_, envelope_,
        mintrans_, count, type, dest, tag, comm, rsend, ssend, message_type,
        next_outbound_[tid]++, next_id_++, completion, content);
    if (server_ == 0)
      {
        throw illformederror("mpiqueue::send:  null server pointer.");
      }
    req->start(server_);
    send_needs_ack_.push_back(req);
  }

  //
  // Receive data.
  //
  void
  mpiqueue::recv(const mpirequest_t key, int count, const mpitype &type,
      const mpiid &source, const mpitag &tag, const mpicomm &comm,
      mpimessage::category message_type,
      eventhandler::ptr &completion)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::recv(" << count << ", "
          << type.to_string() << ", " << source.to_string() << ", "
          << tag.to_string() << ", " << comm.to_string() << ", "
          << (int) message_type << ", " << ")\n";

    recvrequest::ptr req = recvrequest::construct(key, mpiqueue::ptr(this),
        count, type, source, tag, comm, message_type, completion);
    // Figure out whether we already have a matching mesage.
    need_recv_t::iterator it, end = need_recv_.end();
    for (it = need_recv_.begin(); it != end; ++it)
      {
        if (req->matches(*it))
          {
            SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                  << "]::recv:  Matching message found.\n";
            // We're good to go.
            mpimessage::ptr mess = *it;
            need_recv_.erase(it);
            req->handle(mess);
            return;
          }
      }
    // If we get here, we still need to wait for the message.
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
          << "]::recv:  Matching message not found\n";
    pending_message_.push_back(req);
  }

  //
  // Ask for a notification when a message with the given signature arrives.
  //
  void
  mpiqueue::probe(const mpirequest_t key, const mpicomm &comm, const mpiid &source, const mpitag &tag,
      eventhandler::ptr &completion)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::probe(" << comm.to_string()
          << ", " << source.to_string() << ", " << tag.to_string() << ", "
          << ")\n";

    proberequest::ptr req = proberequest::construct(key, comm, source, tag, completion);
    // Figure out whether we already have a matching message.
    need_recv_t::iterator it, end = need_recv_.end();
    for (it = need_recv_.begin(); it != end; ++it)
      {
        if ((*it)->cat() == mpimessage::user && req->matches(*it))
          {
            // We're good to go.
             req->complete(*it);
            return;
          }
      }
    // If we get here, we still need to wait for the message.
    probelist_.push_back(req);
  }

  //  
  // Immediate-mode probe for a message with the given signature.
  //
  mpistatus::ptr
  mpiqueue::iprobe(const mpicomm &comm, const mpiid &source, const mpitag &tag)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::iprobe("
          << comm.to_string() << ", " << source.to_string() << ", "
          << tag.to_string() << ")\n";
    proberequest::ptr req = proberequest::construct(mpirequest_t(), comm, source, tag, eventhandler::ptr());
    need_recv_t::iterator it, end = need_recv_.end();
    for (it = need_recv_.begin(); it != end; ++it)
      {
        if ((*it)->cat() == mpimessage::user && req->matches(*it))
          {
            // This is it.
            return (*it)->build_status();
          }
      }
    // If we got here, there was no match
    return mpistatus::ptr ();
  }

  //
  // Used when a receive operation sends an ack.
  //
  void
  mpiqueue::send_ack(const mpimessage::ptr& message, const
      recvrequest::ptr& operation)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::send_ack(" << message
          << ", " << operation << ")\n";
    debug_message(message);
    // Create a response message of size 0 with source and destination switched.
    // This will have to be cleaned up later.
    timestamp sendtime(1E-9); // in the future, can make this separate send to the ami or something
    mpimessage::ptr ack = mpimessage::construct(server_->get_lib_name(),
        sendtime, envelope_, mintrans_, 0, message->type(), message->dest(),
        message->source(), message->tag(), message->commid(),
        message->seqnum(), message->id(), false, mpimessage::ack,
        message->dest_task(), message->source_task(), payload::null());
    ack->fromaddr_ = message->toaddr_;
    ack->toaddr_ = message->fromaddr_;
    // OK,
    if (operation != 0)
      {
        // If this is non-null, the sender expects a response to its ack.
        ack_sent_[message->id()] = operation;
      }
    SSTMAC_DEBUG << "mpiqueue::send_ack:  Using ignoredevent\n";
    server_->send(ack);
  }

  //
  // Used when a send operation requests an ack for the given message.
  //
  void
  mpiqueue::request_ack(const mpimessage::ptr& message,
      const sendrequest::ptr& operation)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::request_ack("
          << message->to_string() << ", " << ")\n" << "  :  Message id: "
          << message->id().to_string() << "\n";
    debug_message(message);
    ack_needed_[message->id()] = operation;
  }

  //
  // CALLBACK used by mpiserver when this object has a message.
  //
  void
  mpiqueue::incoming_message(const mpimessage::ptr& message)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::incoming_message(" << ", "
          << message << ")\n";

    if(message->get_type() == sst_message::SEND_ACK){
        this->handle_send_ack(message);
        return;
    }

    /// These are the types of messages we may expect:
    /// (1)  Response to handshake request (we have the requestor).
    /// (2)  Data delivery following a handshake ack (we have the receiver)
    /// (3)  Acks from the NIC that a message has left
    /// (4)  New messages (either eager-send or a handshake request).
    /// Only category (4) is subject to order requirements.  The other
    /// categories get handled immediately
    switch (message->cat())
      {
    case mpimessage::ack:
      this->incoming_ack_response(message);
      break;
    case mpimessage::data:
      this->incoming_data_payload(message);
      break;
    default:
      this->incoming_new_message(message);
      }
  }

  //
  // Handle a response to an ack posted from this queue.
  //
  void
  mpiqueue::incoming_ack_response(const mpimessage::ptr &message)
  {
    const messageid &id = message->id();
    SSTMAC_DEBUG << "  :  Incoming ack from task "
            << message->source_task().to_string() << " to task "
            << message->dest_task().to_string() << " with id " << id.to_string()
            << "\n" << " : Finding ack for " << message->id().to_string() << "\n";

    ack_needed_t::iterator it = ack_needed_.find(message->id());
    if (it != ack_needed_.end())
      {
        // We have a winner.
        SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::incoming_message:  "
              << "Ack with matching request.\n";
        sendrequest::ptr req = it->second;
        ack_needed_.erase(it);
        req->ack(message, server_);
      }
    else
      {
        // No match for this ack (this should not happen).
        SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
              << "]::incoming_message:  ERROR:  " << "Ack with no match.\n";
        throw illformederror("pmiqueue::incoming_message: Ack with no match.");
      }
  }

  //
  // Handle data delivery for a request previously acked by this queue.
  //
  void
  mpiqueue::incoming_data_payload(const mpimessage::ptr &message)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::incoming_message:  "
          << "Completion of a message.\n";
    ack_sent_t::iterator it = ack_sent_.find(message->id());
    if (it == ack_sent_.end())
      {
        // Uh-oh.
        if (dbg_.is_active())
          dbg_ << "Fatal error on message.\n";
        debug_message(message);
        throw illformederror("mpiqueue::incoming_message:  Data message "
          "without a matching ack.");
      }
    recvrequest::ptr recver = it->second;
    ack_sent_.erase(it);
    recver->handle(message);
    return;
  }

  //
  // Handle a new incoming message.  Can be either an eager send (with data)
  // or a new handshake request
  //
  void
  mpiqueue::incoming_new_message(const mpimessage::ptr &message)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                                         << "]: incoming_new_message \n";
    const taskid &tid = message->source_task();
    if (message->seqnum() == next_inbound_[tid])
      {
        // Handle this message.
        handle_new_message(message);
        ++next_inbound_[tid];
        // Handle any messages that have been freed by the arrival of this one
        if (!held_[tid].empty())
          {
            SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                                      << "]: there are old messages \n";

            hold_list_t::iterator it = held_[tid].begin(), end = held_[tid].end();
            while (it != end)
              {
                mpimessage::ptr mess = *it;
                if (mess->seqnum() <= next_inbound_[tid])
                  {
                    //it = held_[tid].erase(it);
                    handle_new_message(mess);
                    ++next_inbound_[tid];
                    it++;
                  }else
                  {
                    break;
                  }

              }
            // Now erase all the completed messages from the held queue.
            held_[tid].erase(held_[tid].begin(), it);
          }
      }
    else
      {
        SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                          << "]: putting message in held, seqnum didn't match expected \n";
        held_[tid].insert(message);
      }

    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                                         << "]: incoming_new_message done \n";
  }

  //
  // Complete an inbound message.
  //
  void
  mpiqueue::handle_new_message(const mpimessage::ptr &message)
  {
    bool found_match = false;

    pending_message_t::iterator it, end = pending_message_.end();
    for (it = pending_message_.begin(); it != end; ++it)
      {
        if ((*it)->matches(message))
          {
            SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                  << "]::handle_new_message:  Receiver is ready.\n";
            found_match = true;
            recvrequest::ptr req = *it;

            pending_message_.erase(it);
            req->handle(message);
            break;
          }
      }




    // We get here if no match was found.
    // Messages that don't have a respondent are added to the list
    if(!found_match){
        SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
              << "]::handle_new_message:  Receiver not ready.\n";
        need_recv_.push_back(message);
    }


    // And, finally, mpimessage::user messages are tested for outstanding probes
   // list<proberequest::ptr> proberemove;
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                                 << "]:probelist size: " << probelist_.size() << "\n";
    probelist_t::iterator pit = probelist_.begin(), pend = probelist_.end();
    while(pit != pend)
      {
        SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                                        << "]: testing probe \n";
        if ((*pit)->matches(message))
          {
            SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                              << "]: probe match \n";
            // Match.
            proberequest::ptr preq = *pit;

            pit = probelist_.erase(pit);
            preq->complete(message);
          }else{
              pit++;
          }

        pend = probelist_.end();
      }


    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                     << "]:: handle_new_message done \n";
  }

  void
  mpiqueue::handle_send_ack(const mpimessage::ptr &message)
  {
    SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::incoming_send_ack...";

    send_needs_ack_t::iterator it, end = send_needs_ack_.end();

    for (it = send_needs_ack_.begin(); it != end; ++it)
      {
        if ((*it)->matches(message))
          {
            SSTMAC_DEBUG<< "match found \n";
            // Match.
            sendrequest::ptr sreq = *it;
            send_needs_ack_.erase(it);
            sreq->complete(message);
            return;
          }
      }

    SSTMAC_DEBUG << "no match found :( \n";
  }

  appid
  mpiqueue::get_appid()
  {
    return server_->get_appid();
  }

} } // end of namespace sstmac
