/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIDIRECTGATHERV_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIDIRECTGATHERV_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitype.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitag.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>
#include <vector>

namespace sstmac { namespace sw {

  // Important types.
  class mpiqueue;
  class mpistatus;


  /**
   * Perform an MPI gatherv using direct sends to the root node.
   *
   * This is the way mpich2 does gatherv.  The problem is that the
   * mpi standard specifies that the array of send lengths only needs
   * to be valid at root.  This means that that you either need to send
   * the length array separately (using a gather, bcast, or whatever)
   * or you just bite the bullet (like here) and send the data all
   * directly from the root node.
   */
  class mpidirectgatherv : public mpicollective {
    /// The number of elements of type sendtype sent to root node.
    int sendcount_;
    /// The send type
    mpitype sendtype_;
    /// The number of elements of type recvtype received from each node.
    std::vector<int> recvcount_;
    /// The receive type
    mpitype recvtype_;
    /// The root node.
    mpiid root_;
    /// The tag associated with our communications.
    mpitag tag_;
    /// Store a copy of the communicator.
    mpicomm comm_;

    /// Number of pending sends (only on clients).
    int pending_sends_;
    /// Number of pending receives (only on root).
    int pending_recvs_;

    mpicollpayload::ptr content_;
    payload::const_ptr sendcontent_;

  protected:
    /// Hi.
    mpidirectgatherv(const mpirequest::ptr& thekey,
                     const mpiqueue::ptr& queue,
                     int sendcount,const mpitype &sendtype,
                     const std::vector<int> &recvcount, const mpitype &recvtype,
                     const mpiid &root, const mpitag &tag,
                     const mpicomm &comm,
                     const payload::const_ptr& content,
                     eventhandler::ptr completion);

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(
                               int count,
                               const mpistatus_t& stat);

    /// Callback to signal copy completion (only on root).
    void copy_complete(
                       int count,
                       const mpitype &type,
                       const mpitag &tag,
                       const mpiid &source);

  public:
    typedef boost::intrusive_ptr<mpidirectgatherv> ptr;

    /// Hello.
    static mpidirectgatherv::ptr
    construct(const mpirequest::ptr& thekey,
              const mpiqueue::ptr& queue,
              int sendcount, const mpitype &sendtype,
              const std::vector<int> &recvcount, const mpitype &recvtype,
              const mpiid &root, const mpitag &tag,
              const mpicomm &comm, const payload::const_ptr& content,
              eventhandler::ptr completion);

    /// Goodbye.
    virtual ~mpidirectgatherv() throw();

    /// Start this kernel.
    virtual void start();
  };

} } // end of namespace sstmac.

#endif
