/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIRANKMAP_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIRANKMAP_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpi_comm/mpicomm.h>
#include <sstmac/common/errors.h>

namespace sstmac { namespace sw {

  /**
   * Assist in implementing MPI collective algorithms that expect number
   * of active nodes to be an even power-of-two.  Extra nodes are collapsed
   * into their lower-numbered peers, so node 1 goes to node 0 etc.
   *
   * Provides functionality to define a root node.  If the root node
   * is not an active rank, this rank gets remapped to active rank 0.
   */
  class mpirankmap {
    /// Log2 of the power-of-two size of the system.
    int log2_;
    /// The extra nodes (total size is (1<<log2_) + extra_).
    int extra_;
    /// The root rank.
    mpiid realroot_;
    /// Our communicator
    mpicomm comm_;

  public:
    /// Hi there.
    mpirankmap(const mpiid &root, const mpicomm &comm) :
      log2_(0), extra_(0), realroot_(root), comm_(comm)
    {
      const int size = comm_.size().id;
      while((1<<(log2_+1)) <= size)
        ++log2_;
      extra_ = size - (1 << log2_);
      //std::cerr << "mpirankmap(" << root << ", " << comm << "):  extra_=="
      //          << extra_ << "\n";
    }

    /// Get the log2 size of the system.
    int log2() const {
      return log2_;
    }
    
    /// Get the number of extra nodes.
    int extra() const {
      return extra_;
    }

    /// Test whether this node is inactive in the reduce.
    bool donor() const {
      return donor(comm_.rank());
    }

    /// Test whether ths node is serving for another node.
    bool acceptor() const {
      return acceptor(comm_.rank());
    }

    /// Test whether the given node is inactive in the reduce.
    bool donor(mpiid therank) const {
      return ((therank < mpiid(2 * extra_)) && (therank.id % 2 != 0));
    }

    /// Test whether the given node is serving for another node.
    bool acceptor(mpiid therank) const {
      return ((therank < mpiid(2 * extra_)) && (therank.id % 2 == 0));
    }

    /// What real node rank do we communicate with during a shrink?
    /// \return mpiid(-1) if we are not a donor or acceptor.
    mpiid exchangewith() const {
      mpiid retval(-1);
      if(donor()) retval.id = comm_.rank().id - 1;
      else if(acceptor()) retval.id = comm_.rank().id + 1;
      return retval;
    }

    /// Get the rank of this node.
    const mpiid& real_rank() const {
      return comm_.rank();
    }

    /// Get the real rank of another node.
    mpiid real_rank(int relative) const {
      if(relative < 0)
        throw valueerror("mpirankmap::real_rank:  Negative relative rank.");
      mpiid retval(-1);
      if(relative < extra_) retval.id = relative * 2;
      else retval.id = relative + extra_;
      if(retval >= comm_.size())
        throw valueerror("mpirankmap::real_rank:  Relative rank out of range.");
      return retval;
    }

    /// Get the effective (reduced) rank of this node.
    int effective_rank() const {
      return effective_rank(comm_.rank());
    }

    /// Get the effective (reduced) rank of the given node.
    int effective_rank(mpiid rank) const {
      int retval = -1;
      if(! donor(rank)) {
        if(rank.id < 2 * extra_) retval = rank.id / 2;
        else retval = rank.id - extra_;
      }
      return retval;
    }
  };

} } // end of namespace sstmac

#endif
