/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpiringallgatherv.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw {

  // Pointer types.

  using boost::dynamic_pointer_cast;

  //
  // Hello.
  //
  mpiringallgatherv::mpiringallgatherv(const mpirequest_t& thekey,
      const mpiqueue::ptr& queue,
      int sendcnt, const mpitype &sendtype,
      const std::vector<int> &recvcnt,
      const mpitype &recvtype,
      const mpitag &tag,
      const mpicomm &comm,
      const payload::const_ptr& content,
      eventhandler::ptr completion) :
      mpicollective(thekey, queue, comm, completion), sendcnt_(sendcnt), sendtype_(sendtype),
      recvcnt_(recvcnt), recvtype_(recvtype), tag_(tag), comm_(comm),
      source_((comm.size().id + comm.rank().id - 1) % comm.size().id),
      dest_((comm.rank().id + 1) % comm.size().id),
      pending_sends_(0), pending_recvs_(0)
  {
    std::vector<payload::const_ptr > vals;
    for (int i = 0; i < comm.size().id; i++)
      {
        if (i == comm.rank().id)
          vals.push_back(content);
        else
          vals.push_back(payload::const_ptr ());
      }

    content_ = mpicollpayload::construct(vals);
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpiringallgatherv::send_complete(
      int count,
      const mpitype &type,
      const mpitag &tag,
      const mpiid &dest)
  {
    SSTMAC_DEBUG << "mpiringallgatherv[" << comm_.rank().id << "]::send_complete(" << ", " << count
        << ", " << type.to_string() << ", " << tag.to_string() << ", " << dest.to_string() << ")\n";
    --pending_sends_;
    if(pending_sends_ == 0 && pending_recvs_ == 0) {

        mpicollective::complete(content_);
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpiringallgatherv::recv_complete(
      int count,
      const mpistatus_t& stat)
  {
    SSTMAC_DEBUG << "mpiringallgatherv[" << comm_.rank().id << "]::recv_complete(" << ", " << count
        << ", " << stat << ")\n";
    --pending_recvs_;

    payload::const_ptr pay = stat->content();
    if (pay)
      {
        mpicollpayload::ptr other = mpicollpayload::construct(
            dynamic_pointer_cast<const mpicollpayload> (stat->content()->clone()));

        std::vector<payload::const_ptr > realload =
            other->get_content();

        std::vector<payload::const_ptr >::iterator it, end =
            realload.end();

        int i = 0;
        for (it = realload.begin(); it != end; it++)
          {
            if (*it)
              {
                SSTMAC_DEBUG << "mpiringallgather(" << comm_.rank().id
                    << "]: setting content for position " << i << "\n";

                content_->set_content(*it, i);
              }
            i++;
          }
      }

    if(! iteration_.empty()) {
        // Next send/receive pair.
        srpair sr = iteration_.front();
        iteration_.pop_front();
        ++pending_sends_;
        ++pending_recvs_;
        this->start_recv(sr.recvcnt, recvtype_, tag_, source_);
        this->start_send(sr.sendcnt, sendtype_, tag_, dest_, content_);
    }
    if(pending_sends_ == 0 && pending_recvs_ == 0) {

        mpicollective::complete(content_);
    }
  }

  //
  // Callback to signal copy completion (ignored).
  //
  void mpiringallgatherv::copy_complete(
      int count,
      const mpitype &type,
      const mpitag &tag,
      const mpiid &source)
  {
    SSTMAC_DEBUG << "mpiringallgatherv[" << comm_.rank().id << "]::copy_complete(" << ", " << count
      << ", " << type.to_string() << ", " << tag.to_string() << ", " << source.to_string() << ")\n";
  }

  //
  // Hello.
  //
  mpiringallgatherv::ptr
  mpiringallgatherv::construct(const mpirequest_t& thekey,
      const mpiqueue::ptr& queue,
      int sendcnt, const mpitype &sendtype,
      const std::vector<int> &recvcnt,
      const mpitype &recvtype,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content,
      eventhandler::ptr completion)
  {
    mpiringallgatherv::ptr retval
    (new mpiringallgatherv(thekey, queue, sendcnt, sendtype,
        recvcnt, recvtype, tag, comm, content, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpiringallgatherv::~mpiringallgatherv() throw() {
  }

  //
  // Get busy child.
  //
  void mpiringallgatherv::start()
  {
    SSTMAC_DEBUG << "mpiringallgatherv[" << comm_.rank().id << "]::start()\n";
    mpicollective::start();
    // Catch situations where comm_.size() is 1.
    if(comm_.size().id <= 1) {

        mpicollective::complete(content_);
    }
    else {
        // Set up the iteration_ queue.
        const int size = comm_.size().id;
        const int rank = comm_.rank().id;
        for(int offset = 1; offset < size; ++offset) {
            int receiving_data_owner = (size + rank - offset) % size;
            int sending = sendcnt_;
            int recving = recvcnt_.at(receiving_data_owner);
            iteration_.push_back(srpair(sending, recving));
        }
        // start the first send/receive pair.
        srpair first = iteration_.front();
        iteration_.pop_front();
        pending_sends_ = pending_recvs_ = 1;
        this->start_recv(first.recvcnt, recvtype_, tag_, source_);
        this->start_send(first.sendcnt, sendtype_, tag_, dest_, content_);
    }
  }



} } // end of namespace sstmac
