/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_MPICOLLECTIVE_FUNNELEDCOPY_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_MPICOLLECTIVE_FUNNELEDCOPY_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <deque>

namespace sstmac { namespace sw {

  /**
   * Utility type to handle funneled (one at a time) copying into
   * a shared data array.  Simulates accumulating data from remote sources.
   */
  class mpicollective::funneledcopy : public ptr_type {
    /// Stash copy arguments for future use.
    struct request {
      int count;
      mpitype type;
      mpitag tag;
      mpiid source;

      request(int c, const mpitype &t, const mpitag &g, const mpiid &s) :
        count(c), type(t), tag(g), source(s)
      {}
    };


    // The parent keeps us alive -- let's avoid cyclic references
    boost::intrusive_ptr<mpicollective> parent_;
    // When this object is pending, we want to keep the parent alive.
    mpicollective::ptr keepalive_;

    // Any pending copy requests.
    std::deque<request> pending_;
    
    // Run the next pending request.
    void submit(const request &active) {
     // node_->copy(now, active.count * active.type.size, self_.lock());
    }

  protected:
    // Hi.
    funneledcopy(const mpicollective::ptr& parent):
      parent_(parent)
    {}



  public:

    virtual std::string to_string() const{
            return "funneledcopy";
          }

    typedef boost::intrusive_ptr<funneledcopy> ptr;

    // Hello.
    static funneledcopy::ptr
    construct(const mpicollective::ptr& parent)
    {
      funneledcopy::ptr retval(new funneledcopy(parent));
     // retval->set_self(retval);
      return retval;
    }
    // Bye.
    virtual ~funneledcopy() throw() {}

    // Add a new copy requests.
    void copy(int count, const mpitype &type,
              const mpitag &tag, const mpiid &source)
    {
      if(pending_.empty()) {
        keepalive_ = parent_;
        if(keepalive_ == 0) {
          // Parent has gone out of scope -- bad bad bad.
          throw nullerror("mpicollective::funneledcopy::copy:  NULL parent.");
        }
      }
      pending_.push_back(request(count, type, tag, source));
      if(pending_.size() == 1) {
        // We need to resubmit.
       // timestamp now = node_->now();
       // this->submit(now, pending_.front());
      }
    }

    // Event completion.
    virtual void handle(sst_message::ptr &msg) {
      keepalive_->copy_complete(pending_.front().count,
                                pending_.front().type,
                                pending_.front().tag,
                                pending_.front().source);
      pending_.pop_front();
      if(pending_.empty()) {
        // We just finished the last request -- release the keepalive pointer.
        keepalive_ = mpicollective::ptr();
      }
      else {
        this->submit(pending_.front());
      }
    }
  };

} } // end of namespace sstmac

#endif
