/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <stdlib.h>

#include <sstmac/common/sstmac-config.h>

#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/software/process/app.h>
#include <sstmac/software/process/key.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/common/messages/sst_message.h>

#include <sstmac/software/threading/threading_pth.h>
#include <sstmac/software/threading/threading_pthread.h>
#include <sstmac/software/threading/threading_ucontext.h>

#include <sstmac/software/services/service.h>

namespace sstmac
{
  namespace sw
  {

    stackalloc operatingsystem::stackalloc_(1 << 16, 1 << 24, false);

    debug operatingsystem::dbg_;

    operatingsystem::operatingsystem(hardware_descriptor::ptr &hw)
    {

      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> os | operatingsystem");
          logger::register_logger(&dbg_);
        }

      const char *threading_pchar = getenv("SSTMAC_THREADING");
      if (threading_pchar)
        {
          // Different variants of getenv return different strings.
          // Adjust the threading_string to point to just the value.
          if (!strncmp("SSTMAC_THREADING=", threading_pchar, 17))
            {
              threading_pchar = strchr(threading_pchar, '=');
            }
          if (*threading_pchar == '=')
            threading_pchar++;
          std::string threading_string(threading_pchar);
          if (threading_string == "pth")
            {
#       if defined(SSTMAC_HAVE_PTH)
              des_context_ = new threading_pth();
#       else
              throw valueerror("operatingsystem: SSTMAC_THREADING=pth is not supported");
#       endif
            }
          else if (threading_string == "pthread")
            {
#       if defined(SSTMAC_HAVE_PTHREAD)
              des_context_ = new threading_pthread();
#       else
              throw valueerror("operatingsystem: SSTMAC_THREADING=pthread is not supported");
#       endif
            }
          else if (threading_string == "ucontext")
            {
#       if defined(SSTMAC_HAVE_UCONTEXT)
              des_context_ = new threading_ucontext();
#       else
              throw valueerror(
                  "operatingsystem: SSTMAC_THREADING=ucontext is not supported");
#       endif
            }
          else
            {
              throw valueerror(
                  "operatingsystem: invalid value for SSTMAC_THREADING environmental variable");
            }
        }
      else
        {
#     if defined(SSTMAC_USE_PTH)
          des_context_ = new threading_pth();
#     elif defined(SSTMAC_USE_PTHREAD)
          des_context_ = new threading_pthread();
#     elif defined(SSTMAC_USE_UCONTEXT)
          des_context_ = new threading_ucontext();
#     else
#       error "no default threading library specified"
#     endif
        }

      des_context_->init_context();
      //current_ = &caller_context_;

      threadstack_.push(des_context_);

    }

    operatingsystem::~operatingsystem()
    {

    }

    //
    // Allocate an object of this type.
    //
    operatingsystem::ptr
    operatingsystem::construct(hardware_descriptor::ptr &hw)
    {
      operatingsystem::ptr retval(new operatingsystem(hw));

      return retval;
    }

    void
    operatingsystem::execute_kernel(AMI::AMI_CAT cat, int kernelName,
        sst_message::ptr data)
    {
      AMI* ami = getAMI(cat);

      if (ami == NULL)
        {
          std::stringstream ss;
          ss << "operatingsystem - can't find AMI " << cat;
          throw valueerror(ss.str());
        }
      ami->execute_kernel(kernelName, data);
    }

    ptr_type::ptr
    operatingsystem::execute_kernel(AMI::AMI_CAT cat, int kernelName)
    {
      AMI* ami = getAMI(cat);

      if (ami == NULL)
        {
          std::stringstream ss;
          ss << "operatingsystem - can't find AMI " << cat;
          throw valueerror(ss.str());
        }
      return ami->execute_kernel(kernelName);
    }

    bool
    operatingsystem::kernel_supported(AMI::AMI_CAT cat, int kernelName)
    {
      AMI* ami = getAMI(cat);
      if (ami == NULL)
        {
          return false;
        }
      return ami->kernel_supported(kernelName);
    }

    AMI*
    operatingsystem::getAMI(AMI::AMI_CAT cat)
    {
      if (amis_.find(cat) == amis_.end())
        {
          return NULL;
        }

      return amis_[cat];
    }

    void
    operatingsystem::setAMI(AMI::AMI_CAT cat, AMI* a)
    {
      amis_[cat] = a;
    }

    // ------- THREADING functions ----------

    void
    operatingsystem::switch_to_thread(threading_interface *tothread)
    {
      SSTMAC_DEBUG << "os::switch_thread : begin " << tothread << "\n";
      //threadcontext_t *saved_current = current_;
      //threadcontext_t *saved_caller = &caller_context_;
      // current_ = tothread;
      // caller_context_ = *saved_current;
      threading_interface *from = threadstack_.top();
      threadstack_.push(tothread);

      from->swap_context(tothread);
      SSTMAC_DEBUG << "os::switch_thread : back " << tothread << "\n";
      //current_ = saved_current;
      //caller_context_ = *saved_caller;

    }

    timestamp
    operatingsystem::block(const key::ptr &req)
    {

      if (req == 0)
        {
          throw nullerror("operatingsystem::block:  Refusing to block "
              "pending a null key pointer.");
        }
      // OK, we're ready to roll.
      //blocked_.insert(std::make_pair(req, current_));
      threading_interface* top = threadstack_.top();
      req->blocking_.push_back(top);
      // blocked_count_++;
      SSTMAC_DEBUG << "operatingsystem::block: "
            << "Blocking context with req->blocking_.size() == "
            << req->blocking_.size() << "\n";
      //<< ", leaving " << active_.size() << " contexts pending.\n";
      // swap_context(*current_, caller_context_);

      threadstack_.pop();
      top->swap_context(threadstack_.top());

      if (dbg_.is_active())
        dbg_ << "operatingsystem::block: " << "Done Blocking context " << top
            << "\n";

      return now();
    }

    timestamp
    operatingsystem::unblock(const key::ptr &req)
    {
      if (dbg_.is_active())
        dbg_ << "operatingsystem::unblock: " << " unblocking "
            << req->blocking_.size() << " threads \n";

      key::blocking_t::iterator it, end = req->blocking_.end();
      key::blocking_t templist;

      //save to a temp list first, so we can clear the key before we start switching to threads
      for (it = req->blocking_.begin(); it != end; ++it)
        {
          templist.push_back(*it);
        }

      req->blocking_.clear();

      for (it = templist.begin(); it != templist.end(); ++it)
        {
          switch_to_thread((*it));
        }

      return now();
    }

    void
    operatingsystem::start_thread(const thread::ptr &t)
    {
      add_thread(t);

      /*dbg_ << "os::start_thread : begin " << t.get() << "\n";
       threadcontext_t *saved_current = current_;
       threadcontext_t *saved_caller = &caller_context_;
       current_ = &(t->context_);
       caller_context_ = *saved_current;
       swap_context(caller_context_, *current_);

       dbg_ << "os::start_thread : back " << t.get() << "\n";
       current_ = saved_current;
       caller_context_ = *saved_caller;*/
      switch_to_thread(t->context_);

    }
    /*  void
     operatingsystem::join_thread(const thread::ptr &t,
     threadcontext_t con)
     {
     // scheduler_.join_thread(t, con);
     }*/

    void
    operatingsystem::complete_thread(thread *t)
    {
      stackalloc_.free(t->stack_);

      threading_interface* top = threadstack_.top();

      if (dbg_.is_active())
        dbg_ << "os::complete: " << "Context " << top << "\n";
      //  << ", is complete, now " << active_.size() << " contexts pending.\n";

      threadstack_.pop();
      top->complete_context(threadstack_.top());
      // This thread is not permitted to start again.
      throw illformederror(
          "operatingsystem::complete:  Thread switched in again "
              "after calling complete and relinquishing stack.");
    }

    // ---------------------

    nodeaddress::ptr
    operatingsystem::get_my_addr()
    {
      if (amis_.find(AMI::AMI_QUERY) == amis_.end())
        {
          throw unimplementederror(
              "operatingsystem::get_my_addr can't find query ami");
        }

      ptr_type::ptr p = amis_[AMI::AMI_QUERY]->execute_kernel(AMI::AMI_QUERY_ADDRESS);
      nodeaddress::ptr casted = boost::dynamic_pointer_cast<nodeaddress>(p);
      if(!p){
          throw nullerror("operatingsystem::get_my_addr - error casting to nodeaddress");
      }

      return casted;
    }

    nodeaddress::ptr
    operatingsystem::get_task_addr(softwareid sid)
    {
      if (env_.find(sid.app_) == env_.end())
        {
          std::cout << "sid: " << sid.to_string() << std::endl;
          throw oserror(
              "operatingsystem::get_task_addr(..) - no environment entry for this app");
        }

      return env_[sid.app_]->get_node_address(sid.task_);
    }

    void
    operatingsystem::register_lib(const library::ptr &lib)
    {
      if (libs_.find(lib->get_lib_name()) != libs_.end())
        {
          std::cout << "lib name: " << lib->get_lib_name() << std::endl;
          throw oserror(
              "operatingsystem: trying to register a lib that is already registered");
        }

      libs_[lib->get_lib_name()] = lib;
      lib->init_os(operatingsystem::ptr(this));
    }

    void
       operatingsystem::register_ser(const service::ptr &lib)
       {
         if (libs_.find(lib->get_service_name()) != libs_.end())
           {
             std::cout << "service name: " << lib->get_service_name() << std::endl;
             throw oserror(
                 "operatingsystem: trying to register a service that is already registered");
           }

         libs_[lib->get_service_name()] = lib;
         lib->init_os(operatingsystem::ptr(this));
         add_thread(lib);
       }

    library::ptr
    operatingsystem::get_lib(std::string name)
    {
      if (libs_.find(name) == libs_.end())
        return library::ptr();
      else
        return boost::dynamic_pointer_cast<library>(libs_[name]);
    }

    timestamp
    operatingsystem::now()
    {
      return amis_[AMI::AMI_COMPUTE]->now();
    }

    void
    operatingsystem::add_thread(const thread::ptr &t)
    {
      t->init_thread(des_context_, stackalloc_.alloc(), stackalloc_.stacksize(),
          operatingsystem::ptr(this), NULL);

      threads_.push_back(t);
    }

    void
    operatingsystem::add_application(app::ptr &a,
        const sim_parameters::ptr &params)
    {

      a->consume_params(params);

      add_thread(a);
    }



    void
    operatingsystem::start_apps()
    {
      std::list<thread::ptr>::iterator it;

      for (it = threads_.begin(); it != threads_.end(); it++)
        {
          if (!(*it)->is_started())
            {
              // current_ = &((*it)->context_);
              //swap_context(caller_context_, (*it)->context_);
              switch_to_thread((*it)->context_);
            }
        }

      //check pending messages
      int psize = pending_messages_.size();
      for (int i = 0; i < psize; i++)
        {
          sst_message::ptr msg = pending_messages_.front();
          pending_messages_.pop_front();
          handle_message(msg);
        }
    }

    void
    operatingsystem::handle_message(const sst_message::ptr &msg)
    {
      std::string libn = msg->get_lib_name();
      if (libs_.find(libn) == libs_.end())
        {
          if (pending_messages_.size() > 10)
            std::cout << "WARNING: OS storing message for lib name: " << libn
                << ", size of pending_messages queue is "
                << (pending_messages_.size() + 1) << std::endl;

          pending_messages_.push_back(msg);
          // throw oserror("os: can't find library when handling a message");
        }
      else
        {
          libs_[libn]->handle(msg);
        }
    }

    void
    operatingsystem::apps_kill()
    {
      std::list<thread::ptr>::iterator it;

      if (dbg_.is_active())
        dbg_ << "OS: killing threads \n";
      for (it = threads_.begin(); it != threads_.end(); it++)
        {
          (*it)->kill();
        }
    }

  }
} //end of namespace sstmac
