/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_LAUNCH_ALLOCATIONSTRATEGY_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_LAUNCH_ALLOCATIONSTRATEGY_H_INCLUDED

#include <set>
#include <vector>
#include <sstmac/common/rng.h>
#include <sstmac/backends/native/nodeid.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/ptr_type.h>

#include <sstmac/common/sim_parameters.h>

namespace sstmac
{
  namespace sw
  {

    /**
     * Strategy type for assigning processes to nodes in a parallel run.
     *
     */
    class allocationstrategy : public ptr_type
    {
    protected:
      /// Check the arguments for calls to allocate and throw if they
      /// are bad.
      void
      check_request(int nnode,
          std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation);

      static debug dbg_;

      allocationstrategy(long num_avail)
      {
        if (!dbg_.is_real())
          {
            dbg_ = debug("<debug> allocation | allocationstrategy");
            logger::register_logger(&dbg_);
          }

        for (long i = 0; i < num_avail; i++)
          {
            available_.insert(nodeaddress::convert_from_uniqueid(i));
          }
      }

      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> allocated_;
      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> available_;

    public:

      virtual std::string
      to_string() const
      {
        return "allocationstrat";
      }
      typedef boost::intrusive_ptr<allocationstrategy> ptr;

      virtual
      ~allocationstrategy() throw ();

      /// Get nodes.
      /// @param nnode number of nodes requested
      /// @param allocation returns the nodes that have been allocated
      /// @param available the available set of nodes
      /// @param allocated the allocated set of nodes
      /// Nodes are removed from available and put into allocated.
      /// If not enough nodes are available, then this will throw.
      virtual void
      allocate(
          int nnode,
          std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation) = 0;
    };

    class firstavailableallocation : public allocationstrategy
    {
    protected:
      firstavailableallocation(long num_avail);
    public:
      typedef boost::intrusive_ptr<firstavailableallocation> ptr;

      static firstavailableallocation::ptr
      construct(long num_avail);

      static ptr
      construct(sim_parameters::ptr params)
      {
        return construct(params->get_long_param("node_num"));
      }

      virtual
      ~firstavailableallocation() throw ();
      void
      allocate(int nnode_requested,
          std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation);
    };

    class randomallocation : public allocationstrategy
    {
      RNG::UniformInteger::ptr rng_;
    protected:

      randomallocation(long num_avail, unsigned seed);
      randomallocation(long num_avail, const RNG::UniformInteger::ptr &rng);
    public:
      typedef boost::intrusive_ptr<randomallocation> ptr;

      static randomallocation::ptr
      construct(long num_avail, unsigned seed = 1337);
      static randomallocation::ptr
      construct(long num_avail, const RNG::UniformInteger::ptr &rng);

      static ptr
      construct(sim_parameters::ptr params)
      {
        return construct(params->get_long_param("node_num"));
      }

      virtual
      ~randomallocation() throw ();
      void
      allocate(int nnode_requested,
          std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation);
    };

    class fixedallocation : public allocationstrategy
    {
      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> nodes_;
    protected:
      fixedallocation(
          const std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &nodes);
    public:
      typedef boost::intrusive_ptr<fixedallocation> ptr;

      /// Allocates a fixed set of nodes.
      /// @param nodes a vector of nodes to use for allocation
      static fixedallocation::ptr
      construct(
          const std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &nodes);

      virtual
      ~fixedallocation() throw ();
      void
      allocate(int nnode_requested,
          std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation);
    };

  }
} // end of namespace sstmac

#endif
