
#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/allocation/cartallocation.h>

using sstmac::hw::hdtorus;

namespace sstmac { namespace sw {

  cartallocation
  ::cartallocation(long num_avail, const hdtorus::ptr& network,
                   long xoff, long yoff,
                   long xsize, long ysize,
                   bool relocated,
                   long xorig, long yorig,
                   long xnew, long ynew): allocationstrategy(num_avail),
    network_(network),
    xoff_(xoff),
    yoff_(yoff),
    xsize_(xsize),
    ysize_(ysize),
    relocated_(relocated),
    xorig_(xorig),
    yorig_(yorig),
    xnew_(xnew),
    ynew_(ynew)
  {
    if (network.get() == 0) {
        throw sstmac::valueerror("cartallocation: hdtorus network is null");
    }
    if (network->dimensions().size() != 2) {
        throw sstmac::valueerror("cartallocation: hdtorus network not 2D");
    }
    if (xsize_ < 0 || ysize_ < 0
        || xsize_ > network->dimensions()[0]
        || ysize_ > network->dimensions()[1]) {
      throw sstmac::valueerror("cartallocation: bad size");
    }
    if (relocated_
        && (xorig_ < 0 || yorig_ < 0 || xnew_ < 0 || ynew_ < 0
            || xorig_ >= network->dimensions()[0]
            || yorig_ >= network->dimensions()[1]
            || xnew_ >= network->dimensions()[0]
            || ynew_ >= network->dimensions()[1]
            )) {
      throw sstmac::valueerror("cartallocation: bad relocation coordinate");
    }
  }

  cartallocation::ptr
  cartallocation
  ::construct(long num_avail, const hdtorus::ptr& network,
              long xoff, long yoff,
              long xsize, long ysize)
  {
    ptr r(new cartallocation(num_avail, network,
                             xoff, yoff,
                             xsize, ysize));
    return r;
  }

  cartallocation::ptr
  cartallocation
  ::construct_relocate_node(long num_avail, const hdtorus::ptr& network,
                            long xoff, long yoff,
                            long xsize, long ysize,
                            long xorig, long yorig,
                            long xnew, long ynew)
  {
    ptr r(new cartallocation(num_avail, network,
                             xoff, yoff,
                             xsize, ysize,
                             true,
                             xorig, yorig,
                             xnew, ynew));
    return r;
  }

  void
  cartallocation
  ::allocate(int nnode,
      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation)
  {
    // Check to see if the relocation information is valid.
    if (relocated_) {
      if ((xnew_ >= xoff_ && xnew_ < xoff_+xsize_)
          &&(ynew_ >= yoff_ && ynew_ < yoff_+ysize_)) {
        throw sstmac::valueerror("cartallocation: (xnew,ynew) is invalid");
      }
    }

    //allocation.resize(0);
    //std::set<nodeaddress::ptr> nodeset;
    sstmac::hw::vertex xy(3);
    xy[2] = 0; // This is the number of the node on the switch.
    for (int x=xoff_; x<xoff_+xsize_; x++) {
      for (int y=yoff_; y<yoff_+ysize_; y++) {
        if (relocated_ && x == xorig_ && y == yorig_) {
          xy[0] = xnew_;
          xy[1] = ynew_;
        }
        else {
          xy[0] = x;
          xy[1] = y;
        }
        // convert x and y into nodeid
        //std::cout << "vertex: " << xy << std::endl;
        nodeaddress::ptr nid(nodeaddress::convert_from_uniqueid(network_->node_to_index(xy)));
        std::set<nodeaddress::ptr>::iterator
          niter = available_.find(nid);
        allocation.insert(nid);
        if (niter == available_.end()) {
          throw sstmac::valueerror("cartallocation: node not available");
        }
      }
    }

    //allocation.insert(allocation.begin(),
   //                   nodeset.begin(), nodeset.end());
    for (std::set<sstmac::nodeaddress::ptr>::iterator i=allocation.begin();
         i!=allocation.end(); i++) {
      available_.erase(*i);
      allocated_.insert(*i);
    }
  }

}}
