/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/launch/allocation/dumpiallocation.h>

namespace sstmac
{
  namespace sw
  {

    boost::unordered_map<std::string, nodeaddress::ptr> dumpiallocation::hostnamemap_;

    dumpiallocation::dumpiallocation(std::string mapname,
        sstmac::hw::topology::ptr top) :
        allocationstrategy(0), mapfile_(mapname), top_(top)
    {

    }

    void
    dumpiallocation::read_map_file(const std::string &here,
        const std::string &mapfile,
        std::map<std::string, std::vector<int> >&hostmap)
    {
      hostmap.clear();
      std::ifstream in(mapfile.c_str());
      if (in.bad())
        {
          throw valueerror(here + "could not find node map file");
        }
      int nnode = -1;
      in >> nnode;
      if (nnode <= 0)
        throw valueerror(here + "bad nnode in node map file");
      int ncoor = -1;
      in >> ncoor;
      if (ncoor <= 0)
        throw valueerror(here + "bad ncoor in node map file");
      for (int i = 0; i < nnode; i++)
        {

          std::string hostname;
          in >> hostname;

          if (hostname.size() == 0)
            throw valueerror(here + "bad hostname in map file");
          std::vector<int> coor(ncoor);
          for (int j = 0; j < ncoor; j++)
            {
              coor[j] = -1;
              in >> coor[j];
              if (coor[j] < 0)
                throw valueerror(here + "bad coordinate in map file");
            }

          if (dbg_.is_active())
            {
              SSTMAC_DEBUG << "inserting " << hostname << " into map at coor";
              for (int j = 0; j < ncoor; j++)
                dbg_ << " " << coor[j] ;
              dbg_ << "\n";
            }
          hostmap[hostname] = coor;
        }
    }

    void
    dumpiallocation::allocate(int nnode_requested,
        std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation)
    {
      static std::string here("dumpiallocation::allocate: ");

      std::map<std::string, std::vector<int> > hostmap;
      read_map_file(here, mapfile_, hostmap);

      std::map<std::string, std::vector<int> >::iterator it, end = hostmap.end();

      for (it = hostmap.begin(); it != end; it++)
        {
          std::vector<int> coor = it->second;
          //std::copy(coor.begin(), coor.end(),
          //    std::ostream_iterator<int>(std::cout, " "));

          // find node index for this vertex
          int64_t nodeindex;
          sstmac::hw::vertex hostvert(coor);
          try
          {
              nodeindex = top_->node_to_index(hostvert);
          }
          catch (invalidkeyerror e)
          {
              std::cout << e.what() << "\n" << here
                  << "possible mismatch between network geometry and hostmap\n";
              throw;
          }
          nodeaddress::ptr nid = nodeaddress::convert_from_uniqueid(nodeindex);
          hostnamemap_[it->first] = nid;
          allocation.insert(nid);
        }

    }

  }
}
