/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/launch/instantlaunch.h>
#include <sstmac/software/ami/environment.h>
//#include <sstmac/simulator/nodeinterface.h>
//#include <sstmac/backends/native/components/node/node.h>

//#include <sstmac/backends/native/launch/instantlaunch-completion.h>
#include <sstmac/software/process/taskid.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

#include <sstmac/software/services/launch/launch_message.h>

#include <sstmac/software/services/launch/machine_descriptor.h>
#include <sstmac/software/skeletons/skeletonfactory.h>

namespace sstmac
{
  namespace sw
  {

    std::multimap<nodeaddress::ptr, taskid, nodeaddress::nodeaddresscomp> instantlaunch::taskmap_;
    std::map<appid, environment::const_ptr> instantlaunch::envmap_;

    // Pointer types.

    //
    // Protected constructor.
    //
    instantlaunch::instantlaunch(
        std::list<launchinfo::ptr> &tolaunch, allocationstrategy::ptr &a,
        indexstrategy::ptr &ind) : tolaunch_(tolaunch),
         alloc_(a), indexing_(ind), block_(key::construct())
    {


    }

    service::ptr
    instantlaunch::clone(){
      return construct(tolaunch_, alloc_, indexing_);
    }

    //
    // Public constructor
    //
    instantlaunch::ptr
    instantlaunch::construct(
        std::list<launchinfo::ptr> &tolaunch, allocationstrategy::ptr &a,
        indexstrategy::ptr &ind)
    {
      instantlaunch::ptr retval(new instantlaunch( tolaunch, a, ind));

      return retval;
    }

    //
    // Destructor
    //
    instantlaunch::~instantlaunch() throw ()
    {
    }

    void
    instantlaunch::handle(const sst_message::ptr &msg)
    {
      launch_message::ptr lmsg = boost::dynamic_pointer_cast<launch_message>(
          msg);

      if (!lmsg)
        {
          throw valueerror(
              "instantlaunch: wanted a launch_message, but got something else");
        }

      current_ = lmsg;

      unblock(block_);

    }

    void
    instantlaunch::run()
    {
      if (!os_->kernel_supported(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP))
        {
          throw unimplementederror(
              "instantlaunch: needs the AMI::AMI_COMP_SLEEP kernel to be able to launch");
        }

      std::list<launchinfo::ptr>::iterator it, end = tolaunch_.end();
      for (it = tolaunch_.begin(); it != end; it++)
        {
          launch_message::ptr launchmsg = launch_message::construct(*it);
          SSTMAC_DEBUG << "app " << (*it)->apptype_->to_string() << " is going to launch " << (*it)->num_tasks_ << " at " << (*it)->start_ << "\n";

          os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, launchmsg);
        }

      do
        {
          if (!current_)
            {
              block(block_);
            }

          launchinfo::ptr linfo = current_->get_info();

          if (envmap_.find(linfo->aid_) == envmap_.end())
            {
              SSTMAC_DEBUG << "instantlaunch: initializing allocation for starting at " << linfo->start_ << "\n";

              //this app has not been started anywhere yet, need to allocate, index, and build environment
              environment::ptr env = environment::construct();

              taskmap_.clear();

              SSTMAC_DEBUG << "instantlaunch: requesting allocation for " << linfo->num_tasks_ << " tasks \n";

              std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> nodes;
              alloc_->allocate(linfo->num_tasks_, nodes);

              SSTMAC_DEBUG << "instantlaunch: allocator returned " << nodes.size() << " nodes \n";

              int ppn = ceil(
                  (double) linfo->num_tasks_ / (double) nodes.size());
              std::vector<nodeaddress::ptr> indexed;
              indexing_->allocate(nodes, ppn, indexed, linfo->num_tasks_);

              long i = 0;
              std::vector<nodeaddress::ptr>::iterator it, end = indexed.end();
              for (it = indexed.begin(); it != end; it++)
                {

                  sstmac::sw::taskid id(i);

                  SSTMAC_DEBUG << "taskid: " << id << "\n";

                  SSTMAC_DEBUG << "adding task " << id.to_string()
                        << " to node " << (*it)->to_string() << "\n";

                env->add_item(id, *it);

                taskmap_.insert(std::pair<nodeaddress::ptr, taskid>(*it, id));

                i++;
              }

           envmap_[linfo->aid_] = env;

          }

          os_->set_environment(envmap_[linfo->aid_], linfo->aid_);

          nodeaddress::ptr myaddr = os_->get_my_addr();
          SSTMAC_DEBUG << "instantlaunch(" << myaddr->to_string() << "): considering launching a task...";

        if (taskmap_.find(myaddr) != taskmap_.end())
          {

            std::multimap<nodeaddress::ptr, taskid>::iterator first = taskmap_.lower_bound(myaddr),
                last = taskmap_.upper_bound(myaddr);
            do{
                SSTMAC_DEBUG << "launching task " << first->second << "\n";
                softwareid sid(linfo->aid_, first->second);

                app::ptr theapp = linfo->apptype_->clone(sid);
                os_->add_application(theapp, linfo->params_);
                first++;
            }while(first != last);
            os_->start_apps();
          }else{
              SSTMAC_DEBUG << "nah, don't feel like it \n";

          }
        current_ = launch_message::ptr();

        block(block_);
      }
    while (!is_completed_);
  }

//
// Fire this object off.
// Inherited from launcher.
//
/* void
 instantlaunch::run(std::vector<nodeaddress::ptr> &allocation,
 const nodelist::ptr &nodes)
 {
 SSTMAC_DEBUG << "instantlaunch::run, initializing environment \n";

 int64_t i = 0;
 std::vector<nodeaddress::ptr>::iterator it, end = allocation.end();
 for (it = allocation.begin(); it != end; it++)
 {

 sstmac::sw::taskid id(i);
 nodeid::ptr casted = boost::dynamic_pointer_cast<nodeid>(*it);
 node::ptr the_node = nodes->get(casted);

 SSTMAC_DEBUG << "taskid: " << id << "\n";
 SSTMAC_DEBUG << "nodeid: " << casted->to_string() << "\n";
 if (the_node == 0)
 {
 throw ssterror("instantlaunch::run - null node found");
 }

 SSTMAC_DEBUG << "adding task " << id.to_string() << " to node " << the_node->to_string() << "\n";

 env_->add_item(id, *it);
 the_node->set_environment(env_, aid_);
 i++;
 }

 SSTMAC_DEBUG << "instantlaunch::run, starting applications \n";
 i = 0;
 for (it = allocation.begin(); it != end; it++)
 {
 sstmac::sw::taskid id(i);
 nodeid::ptr casted = boost::dynamic_pointer_cast<nodeid>(*it);
 node::ptr the_node = nodes->get(casted);

 SSTMAC_DEBUG << "instantiating application...";
 sstmac::sw::softwareid sid(aid_, id);
 if (apptype_ == 0)
 {
 throw ssterror("instantlaunch::run - null apptype, can't make clones!");
 }
 sstmac::sw::app::ptr theapp = apptype_->clone(sid);
 SSTMAC_DEBUG << "done \n";

 SSTMAC_DEBUG << "Starting application...";
 the_node->start_application(theapp, params_);
 SSTMAC_DEBUG << "done \n";
 i++;
 }*/


}
} // end of namespace sstmac
