/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_SKELETONS_EXAMPLES_example_shared_memory_H_INCLUDED
#define SSTMAC_SOFTWARE_SKELETONS_EXAMPLES_example_shared_memory_H_INCLUDED

#include <sstmac/software/process/app.h>

#include <sstmac/software/libraries/compute/lib_compute_matrix.h>

namespace sstmac
{
  namespace sw
  {

    /**
     * Example of a shared-memory application.
     */
    class example_shared_memory : public app
    {

      int num_workers_;

    protected:
      // Protected constructor, which should
      // only be called by construct(...).
      example_shared_memory(softwareid id) :
          app(id)
      {

      }

    public:
      // A convenience typedef for the boost pointer type
      typedef boost::intrusive_ptr<example_shared_memory> ptr;

      // Always use this method of construction
      //  when using boost pointers.
      static example_shared_memory::ptr
      construct(softwareid id)
      {
        example_shared_memory::ptr obj(new example_shared_memory(id));
        return obj;
      }

      // Goodbye.
      virtual
      ~example_shared_memory() throw ()
      {
      }

      // Get a copy, used in application launch.
      virtual app::ptr
      clone(softwareid id)
      {
        return construct(id);
      }

      /**
       * skeleton_main replaces main(...) as the entry point of
       * the application.  The simulator will make sure the application
       * starts up in the right places and calls this.
       */
      void
      skeleton_main();

      /**
       * consume_params : this function should grab any parameters it
       * needs from the parameters, in place of passing in argv and argc
       * through main
       * @param params - the simulation parameters passed in through the
       * simulation configuration file and overrides on the command line
       */
      virtual void
      consume_params(const sim_parameters::ptr &params);

      /**
       *  This guy is going to be doing the work, spawned by the main
       *  thread.  Technically, he can be a subclass of app or thread, but app is
       *  probably a better choice.
       */
      class worker : public app
      {

      protected:
        worker(softwareid id, int work) :
            app(id), amount_of_work_(work)
        {
          lib_mat_ = lib_compute_matrix::construct(id);
          register_lib(lib_mat_);
        }

        lib_compute_matrix::ptr lib_mat_;

        static int shared_var_;

        int amount_of_work_;

      public:
        typedef boost::intrusive_ptr<worker> ptr;

        static ptr
        construct(softwareid id, int work)
        {
          return ptr(new worker(id, work));
        }

        // this can be used, but it's probably easier to pass parameters
        //  right through the constructor since we are the ones
        //  doing the constructing
        virtual void
        consume_params(const sim_parameters::ptr &params)
        {
        }

        //have to provide this to satisfy app
        virtual app::ptr
        clone(softwareid id)
        {
          return construct(id, amount_of_work_);
        }

        /**
         * This is our run() function that will get called
         */
        void
        skeleton_main();

      };
    };

  }
} // end of namespace sstmac

#endif
