/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_SKELETONS_MINI_ACTORMATMUL_ACTORMATMUL_MESSAGES_H_INCLUDED
#define SSTMAC_SOFTWARE_SKELETONS_MINI_ACTORMATMUL_ACTORMATMUL_MESSAGES_H_INCLUDED

#include <sstmac/software/skeletons/mini/actormatmul/actormatmul.h>
#include <sstmac/software/libraries/actor/actorpattern.h>
#include <sstmac/software/libraries/actor/actormsg.h>

namespace sstmac { namespace sw
{

  /**
   * Private message base for all actormatmul messages.
   */
  class actormatmul::base : public actormsg
  {
  public:
    typedef boost::intrusive_ptr<base> ptr;

    virtual
    ~base()
    {
    }
    using actormsg::actorclone;
    virtual void
    handle(actorid caller, const actormatmul::ptr& owner) = 0;
  };

  /**
   * Private message type to indicate that a local computor has completed work.
   */
  class actormatmul::localstore : public base
  {
    int iteration_;
    localstore(int iteration) :
        iteration_(iteration)
    {
    }
  public:
    typedef boost::intrusive_ptr<localstore> ptr;

    static localstore::ptr
    construct(int iteration)
    {
      return localstore::ptr(new localstore(iteration));
    }
    virtual
    ~localstore()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct(iteration_);
    }
    virtual void
    handle(actorid id, const actormatmul::ptr& owner)
    {
      owner->local_computor_done(id, iteration_);
    }

    virtual int
    get_kind() const
    {
      return 0;
    }
  };

  /**
   * Private message type to indicate that a remote computor has completed work.
   */
  class actormatmul::remotestore : public base
  {
    int iteration_;
    remotestore(int iteration) :
        iteration_(iteration)
    {
    }
  public:
    typedef boost::intrusive_ptr<remotestore> ptr;

    static remotestore::ptr
    construct(int iteration)
    {
      return remotestore::ptr(new remotestore(iteration));
    }
    virtual
    ~remotestore()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct(iteration_);
    }
    virtual void
    handle(actorid id, const actormatmul::ptr& owner)
    {
      owner->remote_computor_done(id, iteration_);
    }

    virtual int
    get_kind() const
    {
      return 1;
    }
  };

  /**
   * Private message type to indicate arrival of work for a computor.
   */
  class actormatmul::work : public actormsg
  {
    int64_t matdim_;
    int iteration_;
    actorid store_to_;
    work(int64_t matdim, int iteration, const actorid &store_to) :
        matdim_(matdim), iteration_(iteration), store_to_(store_to)
    {
    }
  public:
    typedef boost::intrusive_ptr<work> ptr;

    static work::ptr
    construct(int64_t matdim, int iteration, const actorid &store_to)
    {
      return work::ptr(new work(matdim, iteration, store_to));
    }
    virtual
    ~work()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct(matdim_, iteration_, store_to_);
    }
    int64_t
    matdim() const
    {
      return matdim_;
    }
    int
    iteration() const
    {
      return iteration_;
    }
    const actorid&
    store_to() const
    {
      return store_to_;
    }

    virtual int
    get_kind() const
    {
      return 0;
    }
  };

  /**
   * Private message type to indicate request to fetch data.
   */
  class actormatmul::getdata : public actormatmul::base
  {
    int iteration_;
    getdata(int iteration) :
        iteration_(iteration)
    {
    }
  public:
    typedef boost::intrusive_ptr<getdata> ptr;

    static getdata::ptr
    construct(int iteration)
    {
      return getdata::ptr(new getdata(iteration));
    }
    virtual
    ~getdata()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct(iteration_);
    }
    virtual void
    handle(actorid caller, const actormatmul::ptr& owner)
    {
      owner->get_data(caller, iteration_);
    }
    virtual int
    get_kind() const
    {
      return 2;
    }
  };

  /**
   * Private message type to indicate arrival of data.
   */
  class actormatmul::putdata : public actormatmul::base
  {
    int iteration_;
    putdata(int iteration) :
        iteration_(iteration)
    {
    }
  public:
    typedef boost::intrusive_ptr<putdata> ptr;

    static putdata::ptr
    construct(int iteration)
    {
      return putdata::ptr(new putdata(iteration));
    }
    virtual
    ~putdata()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct(iteration_);
    }
    virtual void
    handle(actorid caller, const actormatmul::ptr& owner)
    {
      owner->put_data(caller, iteration_);
    }

    virtual int
    get_kind() const
    {
      return 3;
    }
  };

  /**
   * Private message type to update lower neighbor index.
   */
  class actormatmul::lowernodeupdate : public actormatmul::base
  {
    int new_flat_lower_;
    lowernodeupdate(int flatid) :
        new_flat_lower_(flatid)
    {
    }
  public:
    typedef boost::intrusive_ptr<lowernodeupdate> ptr;

    static lowernodeupdate::ptr
    construct(int flatid)
    {
      return lowernodeupdate::ptr(new lowernodeupdate(flatid));
    }
    virtual
    ~lowernodeupdate()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct(new_flat_lower_);
    }
    virtual void
    handle(actorid, const actormatmul::ptr& owner)
    {
      owner->update_lower_neighbor(new_flat_lower_);
    }

    virtual int
    get_kind() const
    {
      return 4;
    }
  };

  /**
   * Private message type to update upper neighbor index.
   */
  class actormatmul::uppernodeupdate : public actormatmul::base
  {
    int new_flat_upper_;
    uppernodeupdate(int flatid) :
        new_flat_upper_(flatid)
    {
    }
  public:
    typedef boost::intrusive_ptr<uppernodeupdate> ptr;

    static uppernodeupdate::ptr
    construct(int flatid)
    {
      return uppernodeupdate::ptr(new uppernodeupdate(flatid));
    }
    virtual
    ~uppernodeupdate()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct(new_flat_upper_);
    }
    virtual void
    handle(actorid, const actormatmul::ptr& owner)
    {
      owner->update_upper_neighbor(new_flat_upper_);
    }

    virtual int
    get_kind() const
    {
      return 4;
    }
  };

  /**
   * Private message type to donate a computor.
   */
  class actormatmul::donatecomputor : public actormatmul::base
  {
    actorid thread_;
    donatecomputor(actorid thread) :
        thread_(thread)
    {
    }
  public:
    typedef boost::intrusive_ptr<donatecomputor> ptr;

    static donatecomputor::ptr
    construct(actorid thread)
    {
      return donatecomputor::ptr(new donatecomputor(thread));
    }
    virtual
    ~donatecomputor()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct(thread_);
    }
    virtual void
    handle(actorid, const actormatmul::ptr& owner)
    {
      owner->remote_computor_available(thread_);
    }

    virtual int
    get_kind() const
    {
      return 5;
    }
  };

  /**
   * Private message type to indicate that we're all done.
   */
  class actormatmul::done : public actormatmul::base
  {
    done()
    {
    }
  public:
    typedef boost::intrusive_ptr<done> ptr;

    static done::ptr
    construct()
    {
      return done::ptr(new done);
    }
    virtual
    ~done()
    {
    }
    virtual actormsg::ptr
    actorclone() const
    {
      return construct();
    }
    virtual void
    handle(actorid caller, const actormatmul::ptr& owner)
    {
      owner->all_done(caller);
    }

    virtual int
    get_kind() const
    {
      return 6;
    }
  };

} } // end of namespace sstmac

#endif
