/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/skeletons/mini/minimd/minimd.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-in.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-atom.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-force.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-neighbor.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-integrate.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-thermo.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-comm.h>
#include <sstmac/software/skeletons/mini/minimd/minimd-timer.h>
#include <sstmac/software/skeletons/mini/minimd/lib_compute_minimd.h>
#include <sstmac/software/libraries/mpi/mpiapi.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/basicstringtokenizer.h>
#include <boost/lexical_cast.hpp>
#include <fstream>
#include <math.h>

namespace sstmac
{
  namespace sw
  {

    // Pointer types.

    // Utility function.
    inline void
    gettok(std::ifstream &infile, std::string &line,
        std::deque<std::string> &tok, size_t min_tokens, const char *message)
    {
      getline(infile, line);
      if (!infile.good())
        throw valueerror(std::string(message) + ":  Insufficient input lines.");
      tok.clear();
      pst::BasicStringTokenizer::tokenize(line, tok);
      if (tok.size() < min_tokens)
        throw valueerror(std::string(message) + "  Insufficient input fields.");
    }

    // It takes really long to type lexical_cast<> -- blc is boost::lexical_cast
    template<typename T>
      T
      blc(const std::string &value)
      {
        return lexical_cast<T>(value);
      }

    //
    // Initialize.
    //
    void
    minimd::input()
    {
      /* std::ifstream in(infile_.c_str());
       if (!in)
       {
       std::stringstream ss;
       ss << "minimd::input:  Failed to open infile " << infile_;

       throw valueerror(ss.str());
       }
       // Fixed format file.  First line is title, second line blank.
       std::deque<std::string> tok;
       std::string line;
       for (int i = 0; i < 2; ++i)
       std::getline(in, line);
       // Third line is size of problem (boxx boxy boxz)
       gettok(in, line, tok, 3, "minimd::input:  Problem size");
       in_->nx = blc<int> (tok[0]);
       in_->ny = blc<int> (tok[1]);
       in_->nz = blc<int> (tok[2]);
       // Fourth line is number of timesteps.
       gettok(in, line, tok, 1, "minimd::input:  Number of timesteps");
       integrate_->ntimes = blc<int> (tok[0]);
       // Fifth line is neighbor bins in each direction.
       gettok(in, line, tok, 3, "minimd::input:  Neighbor bins");
       neighbor_->set_bins(blc<int> (tok[0]), blc<int> (tok[1]), blc<int> (tok[2]));
       // Sixth line is timestep size in reduced units.
       gettok(in, line, tok, 1, "minimd::input:  Timestep size");
       integrate_->dt = blc<double> (tok[0]);
       // Seventh line is initial temperature in reduced units.
       gettok(in, line, tok, 1, "minimd::input:  Initial temperature");
       in_->t_request = blc<double> (tok[0]);
       // Eight line is density (again in reduced units).
       gettok(in, line, tok, 1, "minimd::input:  Density");
       in_->rho = blc<double> (tok[0]);
       // Ninth line is steps between reneighboring.
       gettok(in, line, tok, 1, "minimd::input:  Steps between reneighboring");
       neighbor_->every = blc<int> (tok[0]);
       // Tenth line is inner and outer cutoff distances.
       gettok(in, line, tok, 2, "minimd::input:  Cutoff distances");
       force_->cutforce = blc<double> (tok[0]);
       neighbor_->cutneigh = blc<double> (tok[1]);
       // Eleventh (and final) line is steps between thermo calculations.
       gettok(in, line, tok, 1, "minimd::input:  Steps between thermo");
       thermo_->nstat = blc<int> (tok[0]);
       //
       // We can now initialize the interpolator (pre-compute interpolations).

       in.close();*/
    }

    //
    // Set up the box.
    //
    void
    minimd::create_box()
    {
      double latt = pow((4 / in_->rho), (1.0 / 3.0));
      atom_->box.set_prd(in_->nx * latt, in_->ny * latt, in_->nz * latt);
    }

    //
    // Pretend to compute all atom positions etc.
    //
    void
    minimd::create_atoms()
    {
      mpicomm world = mpi()->comm_world();
      int atoms = 4 * in_->nx * in_->ny * in_->nz;
      int size = world.size().id;
      int rank = world.rank().id;
      atom_->natoms = atoms;
      // We're faking the nlocal to be evenly divided (best-case scenario).
      int eachatoms = atoms / size;
      int sum = eachatoms * size;
      atom_->neighatoms.clear();
      atom_->neighatoms.resize(size, eachatoms);
      for (int i = 0; sum + i < atoms; ++i)
        ++atom_->neighatoms.at(i);
      atom_->nlocal = atom_->neighatoms.at(rank);
      // Now pretend to create the actual atoms like miniMD does.
      execution_->compute("create_atoms");
    }

    //
    // Pretend to assign velocities to all atoms.
    //
    void
    minimd::create_velocity()
    {
      mpicomm world = mpi()->comm_world();
      execution_->compute("create_velocity", 0);
      mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
      mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
      mpi()->allreduce(1, mpitype::mpi_double, mpiop::sum, world);
      execution_->compute("create_velocity", 1);
    }

    //
    // We're sort of ignoring output for now.
    //
    void
    minimd::output()
    {
      // get the sum of all profiled timings for this method.
      execution_->compute("output");
    }

    //
    // Hi.
    //
    minimd::minimd(softwareid id) :
        mpiapp(id), execution_(lib_compute_minimd::construct(id)), in_(new in), atom_(
            new atom), force_(new force), neighbor_(new neighbor), integrate_(
            new integrate), thermo_(new thermo), comm_(new comm), timer_(
            new timer)
    {
      register_lib(execution_);
    }

    //
    // Build a new minimd object.
    //
    minimd::ptr
    minimd::construct(softwareid id)
    {
      minimd::ptr retval(new minimd(id));
      //retval->set_self(retval);
      return retval;
    }

    //
    // Goodbye.
    //
    minimd::~minimd() throw ()
    {
    }

    //
    // Get a copy of this object.
    //
    app::ptr
    minimd::clone(softwareid newid)
    {
      return construct(newid);
    }

    //
    // This object gets access to the mpi api.
    //
    void
    minimd::consume_params(const sim_parameters::ptr &params)
    {

      in_->init(execution_, mpi());
      atom_->init(execution_, mpi());
      force_->init(execution_, mpi());
      neighbor_->init(execution_, mpi());
      integrate_->init(execution_, mpi());
      thermo_->init(execution_, mpi());
      comm_->init(execution_, mpi());
      timer_->init(execution_, mpi());

      in_->nx = params->get_int_param("minimd_nx");
      in_->ny = params->get_int_param("minimd_nx");
      in_->nz = params->get_int_param("minimd_nx");

      integrate_->ntimes = params->get_int_param("minimd_ntimes");

      neighbor_->set_bins(params->get_int_param("minimd_bins_x"),
          params->get_int_param("minimd_bins_y"),
          params->get_int_param("minimd_bins_z"));

      integrate_->dt = params->get_double_param("minimd_timestep");

      in_->t_request = params->get_double_param("minimd_temp");

      in_->rho = params->get_double_param("minimd_density");

      neighbor_->every = params->get_int_param("minimd_reneighbor");

      force_->cutforce = params->get_double_param("minimd_cutoff_inner");
      neighbor_->cutneigh = params->get_double_param("minimd_cutoff_outer");

      thermo_->nstat = params->get_int_param("minimd_thermo");
    }

    //
    // Go.
    //
    void
    minimd::skeleton_main()
    {
      //#define SSTMAC_DB(command) std::cerr << rank << " " << #command << "\n"; command
#define SSTMAC_DB(command) command
      timestamp startat = mpi()->init();
      mpiid rank = mpi()->comm_world().rank();
      SSTMAC_DB( this->input());
      SSTMAC_DB( this->create_box());

      // This is, conveniently, pretty much a fixed overhead.
      // unsigned int seedp = rank.id * 12 + 123;
      // double randnoise = (double(rand_r(&seedp)) / RAND_MAX - 0.5) * 0.05;
      timestamp time((0.0003481));
      app::compute(time);
      //SSTMAC_DB( mpi()->compute(execution_->get("main", 0)) );
      SSTMAC_DEBUG << "minimd[" << rank.id << "]: setting up \n";
      SSTMAC_DB( comm_->setup(neighbor_->cutneigh, atom_));
      SSTMAC_DB( this->create_atoms());
      SSTMAC_DB( this->create_velocity());

      SSTMAC_DEBUG << "minimd[" << rank.id << "]: exchanging atoms \n";
      SSTMAC_DB( comm_->exchange(atom_));
      SSTMAC_DB( comm_->borders(atom_));
      SSTMAC_DB( neighbor_->build(atom_));
      SSTMAC_DB( thermo_->compute(0, atom_, neighbor_, force_));

      // Starting dynamics.
      SSTMAC_DEBUG << "minimd[" << rank.id << "]: starting dynamics \n";
      SSTMAC_DB( timer_->barrier_start(timer::TIME_TOTAL));
      SSTMAC_DB(
          integrate_->run(atom_, force_, neighbor_, comm_, thermo_, timer_));
      SSTMAC_DB( timer_->barrier_stop(timer::TIME_TOTAL));

      SSTMAC_DB( thermo_->compute(-1, atom_, neighbor_, force_));
      SSTMAC_DB( this->output());

      // This compute time is also not really dependent on problem size,
      // although it is a little noisier than the earlier value.
      // randnoise = (double(rand_r(&seedp)) / RAND_MAX - 0.5) * 0.10;
      time = timestamp(0.001495);
      app::compute(time);
      //SSTMAC_DB( mpi()->compute(execution_->get("main", 1)) );
      timestamp endat = SSTMAC_DB( mpi()->finalize() );

#undef SSTMAC_DB
    }

  }
} // end of namespace sstmac
