/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/skeletons/undumpi/parsedumpi-callbacks.h>
#include <sstmac/software/skeletons/undumpi/dumpitypeio.h>
#include <sstmac/common/messages/eventdata.h>
//#include <sstmac/processor/computeapi.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <string.h>
#include <boost/lexical_cast.hpp>

namespace sstmac { namespace sw {

  debug parsedumpi::callbacks::dbg_;

  using boost::lexical_cast;


  /// The shared callback pointer array.
  libundumpi_callbacks *parsedumpi::callbacks::cbacks_ = NULL;

  /// Populate callbacks.
  parsedumpi::callbacks::
  callbacks(parsedumpi *parent,
            parsedumpi::callbacks::if_unimplemented action,
            bool remap_requests) :
    parent_(parent), action_(action), remap_requests_(remap_requests)
  {
    if(cbacks_ == NULL)
      set_callbacks();
    trace_compute_start_.sec = -1;
    init_maps();
    memset(&datatype_sizes_, 0, sizeof(dumpi_sizeof));

    if(!dbg_.is_real()){
      dbg_ = debug("<debug> parsedumpicallbacks");
      logger::register_logger(&dbg_);
    }
  }

  /// Start parsing.
  void parsedumpi::callbacks::parse_stream(const std::string &fname) {

      SSTMAC_DEBUG << "parsedumpi::callbacks[" << parent_->world().to_string()
              << "]::parse_stream(" << fname << ")\n";
    static const std::string here("parsedumpi::callbacks::parse_stream");
    if(parent_ == NULL)
      throw nullerror(here + ":  NULL parent pointer.");
    dumpi_profile *profile = undumpi_open(fname.c_str());
    if(profile == NULL) {
      throw ioerror(here + ":  Unable to open \"" + fname + "\" for reading.");
    }
    datatype_sizes_ = undumpi_read_datatype_sizes(profile);
    int retval = undumpi_read_stream(profile, cbacks_, this);
    if(retval != 1)
      throw ioerror(here + ":  Failed reading dumpi stream.\n");
    undumpi_close(profile);
  }

  /// Initialize maps.
  void parsedumpi::callbacks::init_maps() {
    // Null requests.
    request_[DUMPI_REQUEST_ERROR] = mpirequest_t();
    request_[DUMPI_REQUEST_NULL] = mpirequest_t();
    // Built-in mpitypes.
    mpitype_[DUMPI_DATATYPE_ERROR] = mpitype::mpi_null;
    mpitype_[DUMPI_DATATYPE_NULL] = mpitype::mpi_null;
    mpitype_[DUMPI_CHAR] = mpitype::mpi_char;
    mpitype_[DUMPI_SIGNED_CHAR] = mpitype::mpi_signed_char;
    mpitype_[DUMPI_UNSIGNED_CHAR] = mpitype::mpi_unsigned_char;
    mpitype_[DUMPI_BYTE] = mpitype::mpi_byte;
    mpitype_[DUMPI_WCHAR] = mpitype::mpi_wchar;
    mpitype_[DUMPI_SHORT] = mpitype::mpi_short;
    mpitype_[DUMPI_UNSIGNED_SHORT] = mpitype::mpi_unsigned_short;
    mpitype_[DUMPI_INT] = mpitype::mpi_int;
    mpitype_[DUMPI_UNSIGNED] = mpitype::mpi_unsigned;
    mpitype_[DUMPI_LONG] = mpitype::mpi_long;
    mpitype_[DUMPI_UNSIGNED_LONG] = mpitype::mpi_unsigned_long;
    mpitype_[DUMPI_FLOAT] = mpitype::mpi_float;
    mpitype_[DUMPI_DOUBLE] = mpitype::mpi_double;
    mpitype_[DUMPI_LONG_DOUBLE] = mpitype::mpi_long_double;
    mpitype_[DUMPI_LONG_LONG_INT] = mpitype::mpi_long_long_int;
    mpitype_[DUMPI_UNSIGNED_LONG_LONG] = mpitype::mpi_unsigned_long_long;
    mpitype_[DUMPI_LONG_LONG] = mpitype::mpi_long_long;
    mpitype_[DUMPI_PACKED] = mpitype::mpi_packed;
    mpitype_[DUMPI_LB] = mpitype::mpi_lb;
    mpitype_[DUMPI_UB] = mpitype::mpi_ub;
    mpitype_[DUMPI_FLOAT_INT] = mpitype::mpi_float_int;
    mpitype_[DUMPI_DOUBLE_INT] = mpitype::mpi_double_int;
    mpitype_[DUMPI_LONG_INT] = mpitype::mpi_long_int;
    mpitype_[DUMPI_SHORT_INT] = mpitype::mpi_short_int;
    mpitype_[DUMPI_2INT] = mpitype::mpi_2int;
    mpitype_[DUMPI_LONG_DOUBLE_INT] = mpitype::mpi_long_double_int;
    // MPI communicators.
    mpicomm_[DUMPI_COMM_NULL]  = mpicomm::comm_null;
    mpicomm_[DUMPI_COMM_WORLD] = parent_->world();
    mpicomm_[DUMPI_COMM_SELF]  = parent_->mpi()->comm_self();
    // MPI operations
    mpiop_[DUMPI_MAX] = mpiop::max;
    mpiop_[DUMPI_MIN] = mpiop::min;
    mpiop_[DUMPI_SUM] = mpiop::sum;
    mpiop_[DUMPI_PROD] = mpiop::prod;
    mpiop_[DUMPI_LAND] = mpiop::land;
    mpiop_[DUMPI_BAND] = mpiop::band;
    mpiop_[DUMPI_LOR] = mpiop::lor;
    mpiop_[DUMPI_BOR] = mpiop::bor;
    mpiop_[DUMPI_LXOR] = mpiop::lxor;
    mpiop_[DUMPI_BXOR] = mpiop::bxor;
    mpiop_[DUMPI_MINLOC] = mpiop::minloc;
    mpiop_[DUMPI_MAXLOC] = mpiop::maxloc;
    mpiop_[DUMPI_REPLACE] = mpiop::replace;
    mpiop_[DUMPI_OP_NULL] = mpiop::op_null;
  }

  /// Return true if we are remapping requests.
  bool parsedumpi::callbacks::remap_requests() const {
    return remap_requests_;
  }

  // Convert a dumpi time difference into a timestamp.
  inline timestamp deltat(const dumpi_clock &left, const dumpi_clock &right) {
    static const int64_t billion(1e9);
    return timestamp::exact_nsec(billion*(left.sec-right.sec) +
                                 (left.nsec - right.nsec));
  }

  /// Indicate that we are starting an MPI call.
  void parsedumpi::callbacks::
  start_mpi(const dumpi_time *cpu, const dumpi_time *wall,
            const dumpi_perfinfo *perf)
  {
     {
      double time = cpu->start.sec + (cpu->start.nsec * 1e-9);
      SSTMAC_DEBUG << "parsedumpi::callbacks[" << parent_->world().to_string()
              << "::start_mpi(" << time << ")\n";
    }
    const bool use_walltime = true;
    const dumpi_time &usetime = (use_walltime ? *wall : *cpu);
    if(trace_compute_start_.sec >= 0) {
      // This is not the first MPI call -- simulate a compute.


      if(perf != NULL && perf->count > 0) {
        //if(perf != NULL){
        // We get here if we are using the processor model.
        if(size_t(perf->count) != perfctr_compute_start_.size())
          throw illformederror("parsedumpi::callbacks::start_mpi:  "
                               "Number of active perfcounters changed "
                               "between calls.");
        eventdata::ptr evts = eventdata::construct("", timestamp(0));
        SSTMAC_DEBUG << "computing " << perf->count << " things \n";
        for(int i = 0; i < perf->count; ++i) {
          int64_t evtval = perf->invalue[i] - perfctr_compute_start_[i];
          if(evtval < 0) {
            throw illformederror("parsedumpi::callbacks::start_mpi:  "
                                 "Performance counter moved backward "
                                 "between calls.");
          }

          SSTMAC_DEBUG << "compute label: " << perf->counter_tag[i] << "\n";
          evts->set_event(perf->counter_tag[i], evtval);
        }
        parent_->compute_inst(evts);
      }
      else {

        // We get here if we are not using processor modeling.
        timestamp thetime = (parent_->timescaling_ * 
                             deltat(usetime.start, trace_compute_start_));

        double time1 = usetime.start.sec + (usetime.start.nsec * 1e-9);
        double time2 = trace_compute_start_.sec + (trace_compute_start_.nsec * 1e-9);
        SSTMAC_DEBUG << "parsedumpicallbacks: start_mpi: usetime.start = " << time1 << ", trace_compute_start_=" << time2 << "\n";
        SSTMAC_DEBUG << "parsedumpicallbacks: start_mpi: computing for " << thetime << " with timescaling " << parent_->timescaling_ << "\n";
        if(thetime.ticks() < 0) {
          // It turns out that RSQ can actually screw up on times
          // (both wall time and cpu time are wrong for final MPI barrier
          // in 016-4x2x2_run01/dumpi0014.bin collected on 08/07/09
          SSTMAC_DEBUG << "Time error on parser for " << parent_->world().to_string() << "\n"
                    << "time: " << thetime << "\n"
                    << "timescaling: " << parent_->timescaling_ << "\n"
                    << "start: " << usetime.start.sec << "s, " << usetime.start.nsec << "ns\n"
                    << "trace_start: " << trace_compute_start_.sec << "s, "
                    << trace_compute_start_.nsec << "ns\n";
          thetime = timestamp(0);
        }
        parent_->compute(thetime);
      }
    }
  }

  /// Indicate that we have completed an MPI call.
  void parsedumpi::callbacks::
  end_mpi(const dumpi_time *cpu, const dumpi_time *wall,
          const dumpi_perfinfo *perf)
  {
    //double time = cpu.stop.sec + (cpu.stop.nsec * 1e-9);
    //std::cerr << "parsedumpi::callbacks::end_mpi(" << time << ")\n";
    const bool use_walltime = true;
    const dumpi_time &usetime = (use_walltime ? *wall : *cpu);
    trace_compute_start_ = usetime.stop;
    if(perf) {
      perfctr_compute_start_.resize(perf->count);
      for(int i = 0; i < perf->count; ++i)
        perfctr_compute_start_[i] = perf->outvalue[i];
    }
  }

  /// Access the mpiapi.
  mpiapi::ptr parsedumpi::callbacks::getmpi() {
    return parent_->mpi();
  }



  /// Store a single request handle.
  void parsedumpi::callbacks::
  store_request(dumpi_request id, const mpirequest_t &request) {
    //std::cerr << "callbacks[" << parent_->world().to_string() << "]::store_request(" << id << ", " << request << ")\n";
    if(id >= DUMPI_FIRST_USER_REQUEST){
      request_[id] = request;
      SSTMAC_DEBUG << "callbacks: storing request with key " << id << ", value " << request->to_string() << "\n";
    }
  }

  /// Get a single request handle.
  mpirequest_t&
  parsedumpi::callbacks::get_request(dumpi_request id) {
    //std::cerr << "callbacks[" << parent_->world().to_string() << "]::get_request(" << id << ")\n";
    request_map_t::iterator it = request_.find(id);
    if(it == request_.end())
      throw valueerror(std::string("parsedumpi::callbacks::get_request:  "
                                   "Unmapped request id: ") +
                       lexical_cast<std::string>(id) + " on node " +
                       lexical_cast<std::string>(parent_->world().rank()));

    SSTMAC_DEBUG << "callbacks: getting request at key " << id << " with value " << it->second << "\n";
    return it->second;
  }

  //
  // Get a group of request handles.
  //
  void parsedumpi::callbacks::
  get_requests(int count, const dumpi_request *dumpireq,
               std::vector<mpirequest_t> &simreq)
  {
    if(count < 0)
      throw valueerror("parsedumpi::callbacks::get_requests:  "
                       "Negative request count.");
    simreq.resize(count);
    for(int i = 0; i < count; ++i) {
      simreq[i] = get_request(dumpireq[i]);
    }
  }

  /// Remove a request from the map.
  void parsedumpi::callbacks::
  complete_request(dumpi_request id) {
    //std::cerr << "callbacks[" << parent_->world().to_string() << "]::complete_request(" << id << ")\n";
    return; // until I find a better way to deal with persistent requests.
    if(id >= DUMPI_FIRST_USER_REQUEST) {
      request_map_t::iterator it = request_.find(id);
      if(it != request_.end()) {
        request_.erase(it);
      }
    }
  }

  /// Complete multiple requests.
  template <typename Iter>
  void parsedumpi::callbacks::complete_requests(Iter begin, Iter end) {
    for(Iter it = begin; it != end; ++it) {
      complete_request(*it);
    }
  }

  /// The simulator finished a different request than the tracefile.
  /// Swap request indices so they agree on which request is done.
  void parsedumpi::callbacks::
  complete_remap_requests(dumpi_request done_tracefile,
                          dumpi_request done_simulator)
  {

      SSTMAC_DEBUG << "parsedumpi::callbacks[" << parent_->world().to_string()
              << "]::complete_remap_requests(" << done_tracefile
              << ", " << done_simulator << ")\n";
    if(done_tracefile < DUMPI_FIRST_USER_REQUEST ||
       done_simulator < DUMPI_FIRST_USER_REQUEST)
      return;
    if(done_tracefile == done_simulator) {
      this->complete_request(done_tracefile);
    }
    else {
      if(request_.find(done_tracefile) == request_.end())
        throw valueerror(std::string("parsedumpi::callbacks::") +
                         "complete_remap_requests:  No match for request id " +
                         lexical_cast<std::string>(done_tracefile));
      request_[done_simulator] = request_[done_tracefile];
      this->complete_request(done_tracefile);
    }
  }
  
  /// The simulator and tracefile each completed some number of requests,
  /// but these requests may not be the same.
  void parsedumpi::callbacks::
  complete_remap_requests(int count_tracefile,
                          const dumpi_request *done_tracefile,
                          const std::vector<int> &done_simulator)
  {
    int simdone = int(done_simulator.size());
    int remappable = (simdone < count_tracefile ? simdone : count_tracefile);
    for(int i = 0; i < remappable; ++i)
      this->complete_remap_requests(done_tracefile[i], done_simulator.at(i));
  }

  /// Get an mpiid.
  /// Special handling for MPI_ROOT and MPI_ANY_SOURCE.
  mpiid parsedumpi::callbacks::get_mpiid(dumpi_source id) {
    if(id == DUMPI_ANY_SOURCE)
      return mpiid::any_source;
    else if(id == DUMPI_ROOT)
      return mpiid::root;
    else
      return mpiid(id);
  }

  /// Get an mpi tag.
  /// Special handling for MPI_ANY_TAG.
  mpitag parsedumpi::callbacks::get_mpitag(dumpi_tag tag) {
    if(tag == DUMPI_ANY_TAG)
      return mpitag::any_tag;
    else
      return mpitag(tag);
  }

  /// Add a new mpi type.
  void parsedumpi::callbacks::
  add_mpitype(dumpi_datatype id, const mpitype &mpit) {
    if(id < DUMPI_FIRST_USER_DATATYPE)
      throw valueerror(std::string("parsedumpi::callbacks::add_mpitype:  ") +
                       "Trying to redefine built-in datatype index " +
                       lexical_cast<std::string>(id));
    mpitype_[id] = mpit;
  }

  /// Erase the mapping for an mpi type.  Does not erase built-in mpi types.
  void parsedumpi::callbacks::erase_mpitype(dumpi_datatype id) {
    if(id >= DUMPI_FIRST_USER_DATATYPE) {
      mpitype_map_t::iterator it = mpitype_.find(id);
      if(it != mpitype_.end())
        mpitype_.erase(it);
    }
  }

  /// Access mpi type.
  /// \throw valueerror if no mapping exists for this datatype.
  mpitype& parsedumpi::callbacks::get_mpitype(int count, dumpi_datatype id) {
    if((count) > 0 && (id == DUMPI_DATATYPE_ERROR)) {
      throw valueerror(std::string("parsedumpi::callbacks::get_mpitype:  ") +
                       "Cannot send non-zero count of MPI_DATATYPE_ERROR.");
    }
    mpitype_map_t::iterator it = mpitype_.find(id);
    if(it == mpitype_.end()) {
      //throw valueerror(std::string("parsedumpi::callbacks::get_mpitype:  ") +
      //                 "No match for the datatype id " + 
      //                 boost::lexical_cast<std::string>(id));
      if(id < datatype_sizes_.count) {
        mpitype newtype(std::string("Built-in type ") +
                        lexical_cast<std::string>(id),
                        datatype_sizes_.size[id]);
        mpitype_[id] = newtype;
      }
      else {
        std::cerr << "Warning:  No match for datatype id " << id << ".  Assuming double\n";
        mpitype_[id] = mpitype::mpi_double;
      }
      return mpitype_[id];
    }
    return it->second;
  }

  /// Access an mpi communicator.
  mpicomm& parsedumpi::callbacks::get_mpicomm(dumpi_comm id) {
    if(id == DUMPI_COMM_ERROR)
      throw valueerror("parsedumpi::callbacks::get_mpicomm:  "
                       "Cannot operate on MPI_COMM_ERROR.");
    mpicomm_map_t::iterator it = mpicomm_.find(id);
    if(it == mpicomm_.end())
      throw valueerror(std::string("parsedumpi::callbacks::get_mpicomm:  ") +
                       "No match for communicator index " +
                       lexical_cast<std::string>(id));
    return it->second;
  }

  /// Add a new mpi comm.
  void parsedumpi::callbacks::add_mpicomm(dumpi_comm id, const mpicomm &comm) {
    if(id < DUMPI_FIRST_USER_COMM) {
      throw valueerror(std::string("parsedumpi::callbacks::add_mpicomm:  ") +
                       "Trying to redefine built-in comm index " + 
                       lexical_cast<std::string>(id));
    }
    mpicomm_[id] = comm;
  }

  /// Erase the mapping for an mpi comm.  Does not erase MPI_COMM_WORLD.
  void parsedumpi::callbacks::erase_mpicomm(dumpi_comm id) {
    if(id >= DUMPI_FIRST_USER_COMM) {
      mpicomm_map_t::iterator it = mpicomm_.find(id);
      if(it != mpicomm_.end())
        mpicomm_.erase(it);
    }
  }

  /// Add a new mpi op.
  void parsedumpi::callbacks::add_mpiop(dumpi_op id, const mpiop &op) {
    if(id < DUMPI_FIRST_USER_OP) {
      throw valueerror(std::string("parsedumpi::callbacks::add_mpiop:  ") +
                       "Trying to redefine built-in operation index " +
                       lexical_cast<std::string>(id));
    }
    mpiop_[id] = op;
  }

  /// Erase the mapping for an mpi op.  Does not erase built-in operations.
  void parsedumpi::callbacks::erase_mpiop(dumpi_op id) {
    if(id >= DUMPI_FIRST_USER_OP) {
      mpiop_map_t::iterator it = mpiop_.find(id);
      if(it != mpiop_.end())
        mpiop_.erase(it);
    }
  }

  /// Access an mpi opunicator.
  mpiop& parsedumpi::callbacks::get_mpiop(dumpi_op id) {
    static const std::string me("parsedumpi::callbacks::get_mpiop");
    if(id == DUMPI_OP_ERROR)
      throw valueerror(me + ":  Cannot operate on MPI_OP_ERROR.");
    mpiop_map_t::iterator it = mpiop_.find(id);
    if(it == mpiop_.end())
      throw valueerror(me + ":  No match for mpi operation index " +
                       lexical_cast<std::string>(id));
    return it->second;
  }

  /// Set all callbacks.
  void parsedumpi::callbacks::set_callbacks() {
    //libundumpi_clear_callbacks(cbacks_);
    if(cbacks_ == NULL) {
      cbacks_ = new libundumpi_callbacks;
      libundumpi_clear_callbacks(cbacks_);
    }
    cbacks_->on_send                      = on_MPI_Send                     ;
    cbacks_->on_recv                      = on_MPI_Recv                     ;
    cbacks_->on_get_count                 = on_MPI_Get_count                ;
    cbacks_->on_bsend                     = on_MPI_Bsend                    ;
    cbacks_->on_ssend                     = on_MPI_Ssend                    ;
    cbacks_->on_rsend                     = on_MPI_Rsend                    ;
    cbacks_->on_buffer_attach             = on_MPI_Buffer_attach            ;
    cbacks_->on_buffer_detach             = on_MPI_Buffer_detach            ;
    cbacks_->on_isend                     = on_MPI_Isend                    ;
    cbacks_->on_ibsend                    = on_MPI_Ibsend                   ;
    cbacks_->on_issend                    = on_MPI_Issend                   ;
    cbacks_->on_irsend                    = on_MPI_Irsend                   ;
    cbacks_->on_irecv                     = on_MPI_Irecv                    ;
    cbacks_->on_wait                      = on_MPI_Wait                     ;
    cbacks_->on_test                      = on_MPI_Test                     ;
    cbacks_->on_request_free              = on_MPI_Request_free             ;
    cbacks_->on_waitany                   = on_MPI_Waitany                  ;
    cbacks_->on_testany                   = on_MPI_Testany                  ;
    cbacks_->on_waitall                   = on_MPI_Waitall                  ;
    cbacks_->on_testall                   = on_MPI_Testall                  ;
    cbacks_->on_waitsome                  = on_MPI_Waitsome                 ;
    cbacks_->on_testsome                  = on_MPI_Testsome                 ;
    cbacks_->on_iprobe                    = on_MPI_Iprobe                   ;
    cbacks_->on_probe                     = on_MPI_Probe                    ;
    cbacks_->on_cancel                    = on_MPI_Cancel                   ;
    cbacks_->on_test_cancelled            = on_MPI_Test_cancelled           ;
    cbacks_->on_send_init                 = on_MPI_Send_init                ;
    cbacks_->on_bsend_init                = on_MPI_Bsend_init               ;
    cbacks_->on_ssend_init                = on_MPI_Ssend_init               ;
    cbacks_->on_rsend_init                = on_MPI_Rsend_init               ;
    cbacks_->on_recv_init                 = on_MPI_Recv_init                ;
    cbacks_->on_start                     = on_MPI_Start                    ;
    cbacks_->on_startall                  = on_MPI_Startall                 ;
    cbacks_->on_sendrecv                  = on_MPI_Sendrecv                 ;
    cbacks_->on_sendrecv_replace          = on_MPI_Sendrecv_replace         ;
    cbacks_->on_type_contiguous           = on_MPI_Type_contiguous          ;
    cbacks_->on_type_vector               = on_MPI_Type_vector              ;
    cbacks_->on_type_hvector              = on_MPI_Type_hvector             ;
    cbacks_->on_type_indexed              = on_MPI_Type_indexed             ;
    cbacks_->on_type_hindexed             = on_MPI_Type_hindexed            ;
    cbacks_->on_type_struct               = on_MPI_Type_struct              ;
    cbacks_->on_address                   = on_MPI_Address                  ;
    cbacks_->on_type_extent               = on_MPI_Type_extent              ;
    cbacks_->on_type_size                 = on_MPI_Type_size                ;
    cbacks_->on_type_lb                   = on_MPI_Type_lb                  ;
    cbacks_->on_type_ub                   = on_MPI_Type_ub                  ;
    cbacks_->on_type_commit               = on_MPI_Type_commit              ;
    cbacks_->on_type_free                 = on_MPI_Type_free                ;
    cbacks_->on_get_elements              = on_MPI_Get_elements             ;
    cbacks_->on_pack                      = on_MPI_Pack                     ;
    cbacks_->on_unpack                    = on_MPI_Unpack                   ;
    cbacks_->on_pack_size                 = on_MPI_Pack_size                ;
    cbacks_->on_barrier                   = on_MPI_Barrier                  ;
    cbacks_->on_bcast                     = on_MPI_Bcast                    ;
    cbacks_->on_gather                    = on_MPI_Gather                   ;
    cbacks_->on_gatherv                   = on_MPI_Gatherv                  ;
    cbacks_->on_scatter                   = on_MPI_Scatter                  ;
    cbacks_->on_scatterv                  = on_MPI_Scatterv                 ;
    cbacks_->on_allgather                 = on_MPI_Allgather                ;
    cbacks_->on_allgatherv                = on_MPI_Allgatherv               ;
    cbacks_->on_alltoall                  = on_MPI_Alltoall                 ;
    cbacks_->on_alltoallv                 = on_MPI_Alltoallv                ;
    cbacks_->on_reduce                    = on_MPI_Reduce                   ;
    cbacks_->on_op_create                 = on_MPI_Op_create                ;
    cbacks_->on_op_free                   = on_MPI_Op_free                  ;
    cbacks_->on_allreduce                 = on_MPI_Allreduce                ;
    cbacks_->on_reduce_scatter            = on_MPI_Reduce_scatter           ;
    cbacks_->on_scan                      = on_MPI_Scan                     ;
    cbacks_->on_group_size                = on_MPI_Group_size               ;
    cbacks_->on_group_rank                = on_MPI_Group_rank               ;
    cbacks_->on_group_translate_ranks     = on_MPI_Group_translate_ranks    ;
    cbacks_->on_group_compare             = on_MPI_Group_compare            ;
    cbacks_->on_comm_group                = on_MPI_Comm_group               ;
    cbacks_->on_group_union               = on_MPI_Group_union              ;
    cbacks_->on_group_intersection        = on_MPI_Group_intersection       ;
    cbacks_->on_group_difference          = on_MPI_Group_difference         ;
    cbacks_->on_group_incl                = on_MPI_Group_incl               ;
    cbacks_->on_group_excl                = on_MPI_Group_excl               ;
    cbacks_->on_group_range_incl          = on_MPI_Group_range_incl         ;
    cbacks_->on_group_range_excl          = on_MPI_Group_range_excl         ;
    cbacks_->on_group_free                = on_MPI_Group_free               ;
    cbacks_->on_comm_size                 = on_MPI_Comm_size                ;
    cbacks_->on_comm_rank                 = on_MPI_Comm_rank                ;
    cbacks_->on_comm_compare              = on_MPI_Comm_compare             ;
    cbacks_->on_comm_dup                  = on_MPI_Comm_dup                 ;
    cbacks_->on_comm_create               = on_MPI_Comm_create              ;
    cbacks_->on_comm_split                = on_MPI_Comm_split               ;
    cbacks_->on_comm_free                 = on_MPI_Comm_free                ;
    cbacks_->on_comm_test_inter           = on_MPI_Comm_test_inter          ;
    cbacks_->on_comm_remote_size          = on_MPI_Comm_remote_size         ;
    cbacks_->on_comm_remote_group         = on_MPI_Comm_remote_group        ;
    cbacks_->on_intercomm_create          = on_MPI_Intercomm_create         ;
    cbacks_->on_intercomm_merge           = on_MPI_Intercomm_merge          ;
    cbacks_->on_keyval_create             = on_MPI_Keyval_create            ;
    cbacks_->on_keyval_free               = on_MPI_Keyval_free              ;
    cbacks_->on_attr_put                  = on_MPI_Attr_put                 ;
    cbacks_->on_attr_get                  = on_MPI_Attr_get                 ;
    cbacks_->on_attr_delete               = on_MPI_Attr_delete              ;
    cbacks_->on_topo_test                 = on_MPI_Topo_test                ;
    cbacks_->on_cart_create               = on_MPI_Cart_create              ;
    cbacks_->on_dims_create               = on_MPI_Dims_create              ;
    cbacks_->on_graph_create              = on_MPI_Graph_create             ;
    cbacks_->on_graphdims_get             = on_MPI_Graphdims_get            ;
    cbacks_->on_graph_get                 = on_MPI_Graph_get                ;
    cbacks_->on_cartdim_get               = on_MPI_Cartdim_get              ;
    cbacks_->on_cart_get                  = on_MPI_Cart_get                 ;
    cbacks_->on_cart_rank                 = on_MPI_Cart_rank                ;
    cbacks_->on_cart_coords               = on_MPI_Cart_coords              ;
    cbacks_->on_graph_neighbors_count     = on_MPI_Graph_neighbors_count    ;
    cbacks_->on_graph_neighbors           = on_MPI_Graph_neighbors          ;
    cbacks_->on_cart_shift                = on_MPI_Cart_shift               ;
    cbacks_->on_cart_sub                  = on_MPI_Cart_sub                 ;
    cbacks_->on_cart_map                  = on_MPI_Cart_map                 ;
    cbacks_->on_graph_map                 = on_MPI_Graph_map                ;
    cbacks_->on_get_processor_name        = on_MPI_Get_processor_name       ;
    cbacks_->on_get_version               = on_MPI_Get_version              ;
    cbacks_->on_errhandler_create         = on_MPI_Errhandler_create        ;
    cbacks_->on_errhandler_set            = on_MPI_Errhandler_set           ;
    cbacks_->on_errhandler_get            = on_MPI_Errhandler_get           ;
    cbacks_->on_errhandler_free           = on_MPI_Errhandler_free          ;
    cbacks_->on_error_string              = on_MPI_Error_string             ;
    cbacks_->on_error_class               = on_MPI_Error_class              ;
    cbacks_->on_wtime                     = on_MPI_Wtime                    ;
    cbacks_->on_wtick                     = on_MPI_Wtick                    ;
    cbacks_->on_init                      = on_MPI_Init                     ;
    cbacks_->on_finalize                  = on_MPI_Finalize                 ;
    cbacks_->on_initialized               = on_MPI_Initialized              ;
    cbacks_->on_abort                     = on_MPI_Abort                    ;
    cbacks_->on_close_port                = on_MPI_Close_port               ;
    cbacks_->on_comm_accept               = on_MPI_Comm_accept              ;
    cbacks_->on_comm_connect              = on_MPI_Comm_connect             ;
    cbacks_->on_comm_disconnect           = on_MPI_Comm_disconnect          ;
    cbacks_->on_comm_get_parent           = on_MPI_Comm_get_parent          ;
    cbacks_->on_comm_join                 = on_MPI_Comm_join                ;
    cbacks_->on_comm_spawn                = on_MPI_Comm_spawn               ;
    cbacks_->on_comm_spawn_multiple       = on_MPI_Comm_spawn_multiple      ;
    cbacks_->on_lookup_name               = on_MPI_Lookup_name              ;
    cbacks_->on_open_port                 = on_MPI_Open_port                ;
    cbacks_->on_publish_name              = on_MPI_Publish_name             ;
    cbacks_->on_unpublish_name            = on_MPI_Unpublish_name           ;
    cbacks_->on_accumulate                = on_MPI_Accumulate               ;
    cbacks_->on_get                       = on_MPI_Get                      ;
    cbacks_->on_put                       = on_MPI_Put                      ;
    cbacks_->on_win_complete              = on_MPI_Win_complete             ;
    cbacks_->on_win_create                = on_MPI_Win_create               ;
    cbacks_->on_win_fence                 = on_MPI_Win_fence                ;
    cbacks_->on_win_free                  = on_MPI_Win_free                 ;
    cbacks_->on_win_get_group             = on_MPI_Win_get_group            ;
    cbacks_->on_win_lock                  = on_MPI_Win_lock                 ;
    cbacks_->on_win_post                  = on_MPI_Win_post                 ;
    cbacks_->on_win_start                 = on_MPI_Win_start                ;
    cbacks_->on_win_test                  = on_MPI_Win_test                 ;
    cbacks_->on_win_unlock                = on_MPI_Win_unlock               ;
    cbacks_->on_win_wait                  = on_MPI_Win_wait                 ;
    cbacks_->on_alltoallw                 = on_MPI_Alltoallw                ;
    cbacks_->on_exscan                    = on_MPI_Exscan                   ;
    cbacks_->on_add_error_class           = on_MPI_Add_error_class          ;
    cbacks_->on_add_error_code            = on_MPI_Add_error_code           ;
    cbacks_->on_add_error_string          = on_MPI_Add_error_string         ;
    cbacks_->on_comm_call_errhandler      = on_MPI_Comm_call_errhandler     ;
    cbacks_->on_comm_create_keyval        = on_MPI_Comm_create_keyval       ;
    cbacks_->on_comm_delete_attr          = on_MPI_Comm_delete_attr         ;
    cbacks_->on_comm_free_keyval          = on_MPI_Comm_free_keyval         ;
    cbacks_->on_comm_get_attr             = on_MPI_Comm_get_attr            ;
    cbacks_->on_comm_get_name             = on_MPI_Comm_get_name            ;
    cbacks_->on_comm_set_attr             = on_MPI_Comm_set_attr            ;
    cbacks_->on_comm_set_name             = on_MPI_Comm_set_name            ;
    cbacks_->on_file_call_errhandler      = on_MPI_File_call_errhandler     ;
    cbacks_->on_grequest_complete         = on_MPI_Grequest_complete        ;
    cbacks_->on_grequest_start            = on_MPI_Grequest_start           ;
    cbacks_->on_init_thread               = on_MPI_Init_thread              ;
    cbacks_->on_is_thread_main            = on_MPI_Is_thread_main           ;
    cbacks_->on_query_thread              = on_MPI_Query_thread             ;
    cbacks_->on_status_set_cancelled      = on_MPI_Status_set_cancelled     ;
    cbacks_->on_status_set_elements       = on_MPI_Status_set_elements      ;
    cbacks_->on_type_create_keyval        = on_MPI_Type_create_keyval       ;
    cbacks_->on_type_delete_attr          = on_MPI_Type_delete_attr         ;
    cbacks_->on_type_dup                  = on_MPI_Type_dup                 ;
    cbacks_->on_type_free_keyval          = on_MPI_Type_free_keyval         ;
    cbacks_->on_type_get_attr             = on_MPI_Type_get_attr            ;
    cbacks_->on_type_get_contents         = on_MPI_Type_get_contents        ;
    cbacks_->on_type_get_envelope         = on_MPI_Type_get_envelope        ;
    cbacks_->on_type_get_name             = on_MPI_Type_get_name            ;
    cbacks_->on_type_set_attr             = on_MPI_Type_set_attr            ;
    cbacks_->on_type_set_name             = on_MPI_Type_set_name            ;
    cbacks_->on_type_match_size           = on_MPI_Type_match_size          ;
    cbacks_->on_win_call_errhandler       = on_MPI_Win_call_errhandler      ;
    cbacks_->on_win_create_keyval         = on_MPI_Win_create_keyval        ;
    cbacks_->on_win_delete_attr           = on_MPI_Win_delete_attr          ;
    cbacks_->on_win_free_keyval           = on_MPI_Win_free_keyval          ;
    cbacks_->on_win_get_attr              = on_MPI_Win_get_attr             ;
    cbacks_->on_win_get_name              = on_MPI_Win_get_name             ;
    cbacks_->on_win_set_attr              = on_MPI_Win_set_attr             ;
    cbacks_->on_win_set_name              = on_MPI_Win_set_name             ;
    cbacks_->on_alloc_mem                 = on_MPI_Alloc_mem                ;
    cbacks_->on_comm_create_errhandler    = on_MPI_Comm_create_errhandler   ;
    cbacks_->on_comm_get_errhandler       = on_MPI_Comm_get_errhandler      ;
    cbacks_->on_comm_set_errhandler       = on_MPI_Comm_set_errhandler      ;
    cbacks_->on_file_create_errhandler    = on_MPI_File_create_errhandler   ;
    cbacks_->on_file_get_errhandler       = on_MPI_File_get_errhandler      ;
    cbacks_->on_file_set_errhandler       = on_MPI_File_set_errhandler      ;
    cbacks_->on_finalized                 = on_MPI_Finalized                ;
    cbacks_->on_free_mem                  = on_MPI_Free_mem                 ;
    cbacks_->on_get_address               = on_MPI_Get_address              ;
    cbacks_->on_info_create               = on_MPI_Info_create              ;
    cbacks_->on_info_delete               = on_MPI_Info_delete              ;
    cbacks_->on_info_dup                  = on_MPI_Info_dup                 ;
    cbacks_->on_info_free                 = on_MPI_Info_free                ;
    cbacks_->on_info_get                  = on_MPI_Info_get                 ;
    cbacks_->on_info_get_nkeys            = on_MPI_Info_get_nkeys           ;
    cbacks_->on_info_get_nthkey           = on_MPI_Info_get_nthkey          ;
    cbacks_->on_info_get_valuelen         = on_MPI_Info_get_valuelen        ;
    cbacks_->on_info_set                  = on_MPI_Info_set                 ;
    cbacks_->on_pack_external             = on_MPI_Pack_external            ;
    cbacks_->on_pack_external_size        = on_MPI_Pack_external_size       ;
    cbacks_->on_request_get_status        = on_MPI_Request_get_status       ;
    cbacks_->on_type_create_darray        = on_MPI_Type_create_darray       ;
    cbacks_->on_type_create_hindexed      = on_MPI_Type_create_hindexed     ;
    cbacks_->on_type_create_hvector       = on_MPI_Type_create_hvector      ;
    cbacks_->on_type_create_indexed_block = on_MPI_Type_create_indexed_block;
    cbacks_->on_type_create_resized       = on_MPI_Type_create_resized      ;
    cbacks_->on_type_create_struct        = on_MPI_Type_create_struct       ;
    cbacks_->on_type_create_subarray      = on_MPI_Type_create_subarray     ;
    cbacks_->on_type_get_extent           = on_MPI_Type_get_extent          ;
    cbacks_->on_type_get_true_extent      = on_MPI_Type_get_true_extent     ;
    cbacks_->on_unpack_external           = on_MPI_Unpack_external          ;
    cbacks_->on_win_create_errhandler     = on_MPI_Win_create_errhandler    ;
    cbacks_->on_win_get_errhandler        = on_MPI_Win_get_errhandler       ;
    cbacks_->on_win_set_errhandler        = on_MPI_Win_set_errhandler       ;
    cbacks_->on_file_open                 = on_MPI_File_open                ;
    cbacks_->on_file_close                = on_MPI_File_close               ;
    cbacks_->on_file_delete               = on_MPI_File_delete              ;
    cbacks_->on_file_set_size             = on_MPI_File_set_size            ;
    cbacks_->on_file_preallocate          = on_MPI_File_preallocate         ;
    cbacks_->on_file_get_size             = on_MPI_File_get_size            ;
    cbacks_->on_file_get_group            = on_MPI_File_get_group           ;
    cbacks_->on_file_get_amode            = on_MPI_File_get_amode           ;
    cbacks_->on_file_set_info             = on_MPI_File_set_info            ;
    cbacks_->on_file_get_info             = on_MPI_File_get_info            ;
    cbacks_->on_file_set_view             = on_MPI_File_set_view            ;
    cbacks_->on_file_get_view             = on_MPI_File_get_view            ;
    cbacks_->on_file_read_at              = on_MPI_File_read_at             ;
    cbacks_->on_file_read_at_all          = on_MPI_File_read_at_all         ;
    cbacks_->on_file_write_at             = on_MPI_File_write_at            ;
    cbacks_->on_file_write_at_all         = on_MPI_File_write_at_all        ;
    cbacks_->on_file_iread_at             = on_MPI_File_iread_at            ;
    cbacks_->on_file_iwrite_at            = on_MPI_File_iwrite_at           ;
    cbacks_->on_file_read                 = on_MPI_File_read                ;
    cbacks_->on_file_read_all             = on_MPI_File_read_all            ;
    cbacks_->on_file_write                = on_MPI_File_write               ;
    cbacks_->on_file_write_all            = on_MPI_File_write_all           ;
    cbacks_->on_file_iread                = on_MPI_File_iread               ;
    cbacks_->on_file_iwrite               = on_MPI_File_iwrite              ;
    cbacks_->on_file_seek                 = on_MPI_File_seek                ;
    cbacks_->on_file_get_position         = on_MPI_File_get_position        ;
    cbacks_->on_file_get_byte_offset      = on_MPI_File_get_byte_offset     ;
    cbacks_->on_file_read_shared          = on_MPI_File_read_shared         ;
    cbacks_->on_file_write_shared         = on_MPI_File_write_shared        ;
    cbacks_->on_file_iread_shared         = on_MPI_File_iread_shared        ;
    cbacks_->on_file_iwrite_shared        = on_MPI_File_iwrite_shared       ;
    cbacks_->on_file_read_ordered         = on_MPI_File_read_ordered        ;
    cbacks_->on_file_write_ordered        = on_MPI_File_write_ordered       ;
    cbacks_->on_file_seek_shared          = on_MPI_File_seek_shared         ;
    cbacks_->on_file_get_position_shared  = on_MPI_File_get_position_shared ;
    cbacks_->on_file_read_at_all_begin    = on_MPI_File_read_at_all_begin   ;
    cbacks_->on_file_read_at_all_end      = on_MPI_File_read_at_all_end     ;
    cbacks_->on_file_write_at_all_begin   = on_MPI_File_write_at_all_begin  ;
    cbacks_->on_file_write_at_all_end     = on_MPI_File_write_at_all_end    ;
    cbacks_->on_file_read_all_begin       = on_MPI_File_read_all_begin      ;
    cbacks_->on_file_read_all_end         = on_MPI_File_read_all_end        ;
    cbacks_->on_file_write_all_begin      = on_MPI_File_write_all_begin     ;
    cbacks_->on_file_write_all_end        = on_MPI_File_write_all_end       ;
    cbacks_->on_file_read_ordered_begin   = on_MPI_File_read_ordered_begin  ;
    cbacks_->on_file_read_ordered_end     = on_MPI_File_read_ordered_end    ;
    cbacks_->on_file_write_ordered_begin  = on_MPI_File_write_ordered_begin ;
    cbacks_->on_file_write_ordered_end    = on_MPI_File_write_ordered_end   ;
    cbacks_->on_file_get_type_extent      = on_MPI_File_get_type_extent     ;
    cbacks_->on_register_datarep          = on_MPI_Register_datarep         ;
    cbacks_->on_file_set_atomicity        = on_MPI_File_set_atomicity       ;
    cbacks_->on_file_get_atomicity        = on_MPI_File_get_atomicity       ;
    cbacks_->on_file_sync                 = on_MPI_File_sync                ;
    cbacks_->on_iotest                    = on_MPIO_Test                    ;
    cbacks_->on_iowait                    = on_MPIO_Wait                    ;
    cbacks_->on_iotestall                 = on_MPIO_Testall                 ;
    cbacks_->on_iowaitall                 = on_MPIO_Waitall                 ;
    cbacks_->on_iotestany                 = on_MPIO_Testany                 ;
    cbacks_->on_iowaitany                 = on_MPIO_Waitany                 ;
    cbacks_->on_iowaitsome                = on_MPIO_Waitsome                ;
    cbacks_->on_iotestsome                = on_MPIO_Testsome                ;

  }

  /// This is what happens during unimplemented functions.
  int parsedumpi::callbacks::handle_unimplemented(const char*funcname) {
    static const char *format = ("parsedumpi::callbacks::handle_unimplemented."
                                 "  Unimplemented simulator function %s");
    char buf[128];
    switch(action_) {
    case ABORT:
      snprintf(buf, 127, format, funcname);
      buf[127] = '\0';
      throw unimplementederror(buf);
    case PRINT:
      fprintf(stderr, format, funcname);
    case IGNORE:
    default:
      return 1;
    }
  }
  int parsedumpi::callbacks::handle_unimplemented(const std::string &funcname){
    return handle_unimplemented(funcname.c_str());
  }


#undef PRFX
  static const std::string PRFX("parsedumpi::callbacks::on_");
  
  int parsedumpi::callbacks::
  on_MPI_Send(const dumpi_send *prm, uint16_t thread,
              const dumpi_time *cpu, const dumpi_time *wall,
              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Send");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->send(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                       cb->get_mpiid(prm->dest), cb->get_mpitag(prm->tag),
                       cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }

  int parsedumpi::callbacks::
  on_MPI_Recv(const dumpi_recv *prm, uint16_t thread,
              const dumpi_time *cpu, const dumpi_time *wall,
              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Recv");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    const_mpistatus_t stat;
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->recv(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                       cb->get_mpiid(prm->source), cb->get_mpitag(prm->tag),
                       cb->get_mpicomm(prm->comm), stat);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Get_count(const dumpi_get_count *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Get_count");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    // We deliberately ignore get_count -- there is no reason to emulate it.
    (void)prm;
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Bsend(const dumpi_bsend *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Bsend");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->bsend(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                        cb->get_mpiid(prm->dest), cb->get_mpitag(prm->tag),
                        cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Ssend(const dumpi_ssend *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Ssend");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->ssend(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                        cb->get_mpiid(prm->dest), cb->get_mpitag(prm->tag),
                        cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Rsend(const dumpi_rsend *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Rsend");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->rsend(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                        cb->get_mpiid(prm->dest), cb->get_mpitag(prm->tag),
                        cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Buffer_attach(const dumpi_buffer_attach *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Buffer_attach");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->buffer_attach(prm->size);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Buffer_detach(const dumpi_buffer_detach *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Buffer_detach");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    int tmp;
    cb->getmpi()->buffer_detach(tmp);
    // We are deliberately not using the trace file arguments.
    (void)prm;
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Isend(const dumpi_isend *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Isend");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->isend(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                        cb->get_mpiid(prm->dest), cb->get_mpitag(prm->tag),
                        cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Ibsend(const dumpi_ibsend *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Ibsend");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->ibsend(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                         cb->get_mpiid(prm->dest), cb->get_mpitag(prm->tag),
                         cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Issend(const dumpi_issend *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Issend");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->issend(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                         cb->get_mpiid(prm->dest), cb->get_mpitag(prm->tag),
                         cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Irsend(const dumpi_irsend *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Irsend");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->irsend(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                         cb->get_mpiid(prm->dest), cb->get_mpitag(prm->tag),
                         cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Irecv(const dumpi_irecv *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Irecv");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->irecv(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                        cb->get_mpiid(prm->source), cb->get_mpitag(prm->tag),
                        cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Wait(const dumpi_wait *prm, uint16_t thread,
              const dumpi_time *cpu, const dumpi_time *wall,
              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Wait");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req = cb->get_request(prm->request);
    const_mpistatus_t stat;
    cb->getmpi()->wait(req, stat);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Test(const dumpi_test *prm, uint16_t thread,
              const dumpi_time *cpu, const dumpi_time *wall,
              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Test");
      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req = cb->get_request(prm->request);
    bool flag;
    const_mpistatus_t stat;
    cb->getmpi()->test(req, flag, stat);
    if(prm->flag) {
      if(cb->remap_requests())
        cb->complete_request(prm->request);
      else if(! flag)
        cb->getmpi()->wait(req, stat);
    }
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Request_free(const dumpi_request_free *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Request_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    // Not implemented yet -- will require some fairly significant
    // reconsideration of how pending requests are managed.
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Waitany(const dumpi_waitany *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    return waitany_pessimistic(prm, thread, cpu, wall, perf, uarg);
  }
  
  //
  // Variant implementation of MPI_Waitany:  optimistic wait.
  //
  int parsedumpi::callbacks::
  waitany_optimistic(const dumpi_waitany *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)
  {
    static const std::string me("MPI_Waitany{optimistic}");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    // Set up.
    std::vector<mpirequest_t> req;
    cb->get_requests(prm->count, prm->requests, req);
    int index;
    const_mpistatus_t stat;
    if(cb->remap_requests()) {
      // Wait for the first of the input requests for finish
      cb->getmpi()->waitany(req, index, stat);
      cb->complete_remap_requests(prm->requests[prm->index],
                                  prm->requests[index]);
    }
    else {
      // Wait for exactly the same request as finished in the dumpi run.
      cb->getmpi()->wait(req.at(prm->index), stat);
    }
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }

  //
  // Variant implementation of MPI_Waitany:  pessimistic wait.
  //
  int parsedumpi::callbacks::
  waitany_pessimistic(const dumpi_waitany *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)
  {
    static const std::string me("MPI_Waitany{pessimistic}");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    if(prm->index >= 0 && prm->index < prm->count) {
      // The trace file matched a request -- we will match the same one.
      callbacks *cb = reinterpret_cast<callbacks*>(uarg);
      if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
      cb->start_mpi(cpu, wall, perf);
      mpirequest_t req = cb->get_request(prm->requests[prm->index]);
      const_mpistatus_t stat;
      cb->getmpi()->wait(req, stat);
      cb->end_mpi(cpu, wall, perf);      
    }
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Testany(const dumpi_testany *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    return testany_pessimistic(prm, thread, cpu, wall, perf, uarg);
  }

  //
  // Optimistic remapping of an MPI_Testany call.
  //
  int parsedumpi::callbacks::
  testany_optimistic(const dumpi_testany *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)
  {
    static const std::string me("MPI_Testany{optimistic}");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
     {
      // Ultra-verbose debug to debug a problem Ted Ahn identified
      SSTMAC_DEBUG << "prm={count:" << prm->count << ", requests:[";
      for(int i = 0; i < prm->count; ++i)
        SSTMAC_DEBUG << prm->requests[i] << (i < prm->count-1 ? ", " : "], ");
      SSTMAC_DEBUG << "index:" << prm->index << ", flag:" << prm->flag
              << ", status:";
      if(prm->status) {
        SSTMAC_DEBUG << "{bytes:" << prm->status->bytes << ", source:"
                << prm->status->source << ", tag:" << prm->status->tag
                << ", cancelled:" << int(prm->status->cancelled) << ", error:"
                << int(prm->status->error) << "}";
      }
      else {
        SSTMAC_DEBUG << "NULL";
      }
      SSTMAC_DEBUG << "}\n";
    }
    // We only bother with this if the parent matched a request
    // (otherwise, we have nothing meaningful to remap).
    if(prm->flag == 1) { 
      callbacks *cb = reinterpret_cast<callbacks*>(uarg);  
      if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
      cb->start_mpi(cpu, wall, perf);
      // Set up.
      std::vector<mpirequest_t> req;
      cb->get_requests(prm->count, prm->requests, req);
      int index;
      bool flag;
      const_mpistatus_t stat;
      if(cb->remap_requests()) {
        // Wait for the first of the input requests for finish
        cb->getmpi()->testany(req, index, flag, stat);
        if(flag) cb->complete_remap_requests(prm->requests[prm->index],
                                             prm->requests[index]);
      }
      else {
        // Wait for exactly the same request as finished in the dumpi run.
        cb->getmpi()->wait(req.at(prm->index), stat);
        cb->complete_request(prm->index);
      }    
      cb->end_mpi(cpu, wall, perf);
    }
    return 1;
  }

  //
  // Pessimistic remapping of testany calls.
  //
  int parsedumpi::callbacks::
  testany_pessimistic(const dumpi_testany *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)
  {
    static const std::string me("MPI_Testany{pessimistic}");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    if(prm->flag == 1 && prm->index >= 0 && prm->index < prm->count) {
      // The trace file matched a request -- we will match the same one.
      callbacks *cb = reinterpret_cast<callbacks*>(uarg);
      if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
      cb->start_mpi(cpu, wall, perf);
      mpirequest_t req = cb->get_request(prm->requests[prm->index]);
      const_mpistatus_t stat;
      cb->getmpi()->wait(req, stat);
      cb->end_mpi(cpu, wall, perf);      
    }
    return 1;
  }  
  
  int parsedumpi::callbacks::
  on_MPI_Waitall(const dumpi_waitall *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Waitall");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    // Set up.
    std::vector<mpirequest_t> req;
    cb->get_requests(prm->count, prm->requests, req);
    std::vector<const_mpistatus_t> stat;
    cb->getmpi()->waitall(req, stat);
    cb->complete_requests(prm->requests, prm->requests + prm->count);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Testall(const dumpi_testall *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Testall");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    // Set up.
    std::vector<mpirequest_t> req;
    std::vector<const_mpistatus_t> stat;
    // Only go ahead if flag is true (otherwise may can have some odd findings).
    // This also means that we convert the 'testall' to a 'waitall'
    if(prm->flag == 1) {
      cb->get_requests(prm->count, prm->requests, req);
      cb->getmpi()->waitall(req, stat);
      if(cb->remap_requests()) {
        cb->complete_requests(prm->requests, prm->requests + prm->count);
      }
      cb->end_mpi(cpu, wall, perf);
    }
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Waitsome(const dumpi_waitsome *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Waitsome");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    // Set up.
    std::vector<mpirequest_t> req;
    std::vector<int> ind;
    std::vector<const_mpistatus_t> stat;
    if(cb->remap_requests()) {
      cb->get_requests(prm->count, prm->requests, req);
      cb->getmpi()->waitsome(req, ind, stat);
      cb->complete_remap_requests(prm->outcount, prm->indices, ind);
    }
    else {
      for(int i = 0; i < prm->outcount; ++i)
        req.push_back(cb->get_request(prm->indices[i]));
      cb->getmpi()->waitall(req, stat);
    }
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Testsome(const dumpi_testsome *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Testsome");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    // Set up.
    std::vector<mpirequest_t> req;
    std::vector<int> ind;
    std::vector<const_mpistatus_t> stat;
    if(cb->remap_requests()) {
      cb->get_requests(prm->count, prm->requests, req);
      cb->getmpi()->testsome(req, ind, stat);
      cb->complete_remap_requests(prm->outcount, prm->indices, ind);
    }
    else {
      for(int i = 0; i < prm->outcount; ++i)
        req.push_back(cb->get_request(prm->indices[i]));
      cb->getmpi()->waitall(req, stat);
    }
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Iprobe(const dumpi_iprobe *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Iprobe");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    /*
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    // Set up.
    if(prm->flag) {
      bool flag;
      const_mpistatus_t stat;
      cb->getmpi()->iprobe(cb->get_mpiid(prm->source),
                           cb->get_mpitag(prm->tag),
                           cb->get_mpicomm(prm->comm),
                           flag, stat);
    }
    cb->end_mpi(cpu, wall, perf);
    */
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Probe(const dumpi_probe *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Probe");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    /*
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    const_mpistatus_t stat;
    bool flag;
    cb->getmpi()->iprobe(cb->get_mpiid(prm->source),
                         cb->get_mpitag(prm->tag),
                         cb->get_mpicomm(prm->comm),
                         flag, stat);
    cb->end_mpi(cpu, wall, perf);
    */
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cancel(const dumpi_cancel *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cancel");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req = cb->get_request(prm->request);
    cb->getmpi()->cancel(req);
    cb->complete_request(prm->request);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Test_cancelled(const dumpi_test_cancelled *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Test_cancelled");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    // We are deliberately ignoring this one.
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Send_init(const dumpi_send_init *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Send_init");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->send_init(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                            cb->get_mpiid(prm->dest),
                            cb->get_mpitag(prm->tag),
                            cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Bsend_init(const dumpi_bsend_init *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Bsend_init");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->bsend_init(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                             cb->get_mpiid(prm->dest),
                             cb->get_mpitag(prm->tag),
                             cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Ssend_init(const dumpi_ssend_init *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Ssend_init");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->ssend_init(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                             cb->get_mpiid(prm->dest),
                             cb->get_mpitag(prm->tag),
                             cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Rsend_init(const dumpi_rsend_init *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Rsend_init");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->rsend_init(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                             cb->get_mpiid(prm->dest),
                             cb->get_mpitag(prm->tag),
                             cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Recv_init(const dumpi_recv_init *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Recv_init");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpirequest_t req;
    cb->getmpi()->recv_init(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                            cb->get_mpiid(prm->source),
                            cb->get_mpitag(prm->tag),
                            cb->get_mpicomm(prm->comm), req);
    cb->store_request(prm->request, req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Start(const dumpi_start *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Start");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->start(cb->get_request(prm->request));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Startall(const dumpi_startall *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Startall");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    std::vector<mpirequest_t> req;
    cb->start_mpi(cpu, wall, perf);
    cb->get_requests(prm->count, prm->requests, req);
    cb->getmpi()->startall(req);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Sendrecv(const dumpi_sendrecv *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Sendrecv");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    const_mpistatus_t stat;
    cb->getmpi()->sendrecv(prm->sendcount,
                           cb->get_mpitype(prm->sendcount, prm->sendtype),
                           cb->get_mpiid(prm->dest),
                           cb->get_mpitag(prm->sendtag),
                           prm->recvcount,
                           cb->get_mpitype(prm->recvcount, prm->recvtype),
                           cb->get_mpiid(prm->source),
                           cb->get_mpitag(prm->recvtag),
                           cb->get_mpicomm(prm->comm), stat);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Sendrecv_replace(const dumpi_sendrecv_replace *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Sendrecv_replace");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    const_mpistatus_t stat;
    cb->getmpi()->sendrecv_replace(prm->count,
                                   cb->get_mpitype(prm->count, prm->datatype),
                                   cb->get_mpiid(prm->dest),
                                   cb->get_mpitag(prm->sendtag),
                                   cb->get_mpiid(prm->source),
                                   cb->get_mpitag(prm->recvtag),
                                   cb->get_mpicomm(prm->comm), stat);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_contiguous(const dumpi_type_contiguous *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_contiguous");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpitype newtype;
    cb->getmpi()->type_contiguous(prm->count,
                                  cb->get_mpitype(prm->count, prm->oldtype),
                                  newtype);
    cb->add_mpitype(prm->newtype, newtype);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_vector(const dumpi_type_vector *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_vector");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    const mpitype &oldtype = cb->get_mpitype(1, prm->oldtype);
    int size = prm->count * prm->blocklength * oldtype.size;
    if(prm->newtype < cb->datatype_sizes_.count) {
      if(size != cb->datatype_sizes_.size[prm->newtype])
        //std::cerr << me << ":  Warning:  Created datatype of size "
        //        << size << " but trace file registered the size "
        //        << cb->datatype_sizes_.size[prm->newtype] << "\n";
        size = cb->datatype_sizes_.size[prm->newtype];
    }
    mpitype newtype(me + " " + lexical_cast<std::string>(prm->newtype), size);
    cb->add_mpitype(prm->newtype, newtype);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_hvector(const dumpi_type_hvector *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_hvector");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    const mpitype &oldtype = cb->get_mpitype(1, prm->oldtype);
    int size = prm->count * prm->blocklength * oldtype.size;
    if(prm->newtype < cb->datatype_sizes_.count) {
      if(size != cb->datatype_sizes_.size[prm->newtype])
        //std::cerr << me << ":  Warning:  Created datatype of size "
        //        << size << " but trace file registered the size "
        //        << cb->datatype_sizes_.size[prm->newtype] << "\n";
        size = cb->datatype_sizes_.size[prm->newtype];
    }
    mpitype newtype(me + " " + lexical_cast<std::string>(prm->newtype), size);
    cb->add_mpitype(prm->newtype, newtype);    
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_indexed(const dumpi_type_indexed *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_indexed");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    int size = 0;
    const mpitype &oldtype = cb->get_mpitype(1, prm->oldtype);
    for(int i = 0; i < prm->count; ++i)
      size += prm->lengths[i] * oldtype.size;
    if(prm->newtype < cb->datatype_sizes_.count) {
      if(size != cb->datatype_sizes_.size[prm->newtype])
        //std::cerr << me << ":  Warning:  Created datatype of size "
        //        << size << " but trace file registered the size "
        //        << cb->datatype_sizes_.size[prm->newtype] << "\n";
        size = cb->datatype_sizes_.size[prm->newtype];
    }
    mpitype newtype(me + " " + lexical_cast<std::string>(prm->newtype), size);
    cb->add_mpitype(prm->newtype, newtype);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_hindexed(const dumpi_type_hindexed *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_hindexed");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    int size = 0;
    const mpitype &oldtype = cb->get_mpitype(1, prm->oldtype);
    for(int i = 0; i < prm->count; ++i)
      size += prm->lengths[i] * oldtype.size;
    if(prm->newtype < cb->datatype_sizes_.count) {
      if(size != cb->datatype_sizes_.size[prm->newtype])
        //std::cerr << me << ":  Warning:  Created datatype of size "
        //        << size << " but trace file registered the size "
        //        << cb->datatype_sizes_.size[prm->newtype] << "\n";
        size = cb->datatype_sizes_.size[prm->newtype];
    }
    mpitype newtype(me + " " + lexical_cast<std::string>(prm->newtype), size);
    cb->add_mpitype(prm->newtype, newtype);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_struct(const dumpi_type_struct *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_struct");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    int size = 0;
    for(int i = 0; i < prm->count; ++i)
      size += (prm->lengths[i] *
               cb->get_mpitype(prm->lengths[i], prm->oldtypes[i]).size);
    // Sanity check.
    if(prm->newtype < cb->datatype_sizes_.count) {
      if(size != cb->datatype_sizes_.size[prm->newtype])
        //std::cerr << me << ":  Warning:  Created datatype of size "
        //        << size << " but trace file registered the size "
        //        << cb->datatype_sizes_.size[prm->newtype] << "\n";
        size = cb->datatype_sizes_.size[prm->newtype];
    }
    mpitype newtype(me + " " + lexical_cast<std::string>(prm->newtype), size);
    cb->add_mpitype(prm->newtype, newtype);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Address(const dumpi_address *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Address");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_extent(const dumpi_type_extent *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_extent");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_size(const dumpi_type_size *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->type_size(cb->get_mpitype(0, prm->datatype));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_lb(const dumpi_type_lb *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_lb");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_ub(const dumpi_type_ub *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_ub");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_commit(const dumpi_type_commit *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_commit");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->type_commit(cb->get_mpitype(1, prm->datatype));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_free(const dumpi_type_free *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->type_free(cb->get_mpitype(0, prm->datatype));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Get_elements(const dumpi_get_elements *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Get_elements");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Pack(const dumpi_pack *prm, uint16_t thread,
              const dumpi_time *cpu, const dumpi_time *wall,
              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Pack");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Unpack(const dumpi_unpack *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Unpack");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Pack_size(const dumpi_pack_size *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Pack_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Barrier(const dumpi_barrier *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Barrier");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->barrier(cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Bcast(const dumpi_bcast *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Bcast");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->bcast(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                        cb->get_mpiid(prm->root), cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Gather(const dumpi_gather *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Gather");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->gather(prm->sendcount,
                         cb->get_mpitype(prm->sendcount, prm->sendtype),
                         prm->recvcount,
                         cb->get_mpitype(prm->recvcount, prm->recvtype),
                         cb->get_mpiid(prm->root), cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Gatherv(const dumpi_gatherv *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Gatherv");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    std::vector<int> recvcounts;
    if(prm->commrank == prm->root)
      recvcounts.insert(recvcounts.end(), prm->recvcounts,
                        prm->recvcounts + prm->commsize);
    cb->getmpi()->gatherv(prm->sendcount,
                          cb->get_mpitype(prm->sendcount, prm->sendtype),
                          recvcounts,
                          cb->get_mpitype(1, prm->recvtype),
                          cb->get_mpiid(prm->root), cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Scatter(const dumpi_scatter *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Scatter");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->scatter(prm->sendcount,
                          cb->get_mpitype(prm->sendcount, prm->sendtype),
                          prm->recvcount,
                          cb->get_mpitype(prm->recvcount, prm->recvtype),
                          cb->get_mpiid(prm->root), cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Scatterv(const dumpi_scatterv *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Scatterv");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    std::vector<int> sendcounts;
    if(prm->commrank == prm->root)
      sendcounts.insert(sendcounts.end(), prm->sendcounts,
                        prm->sendcounts + prm->commsize);
    cb->getmpi()->scatterv(sendcounts, cb->get_mpitype(1, prm->sendtype),
                           prm->recvcount,
                           cb->get_mpitype(prm->recvcount, prm->recvtype),
                           cb->get_mpiid(prm->root),
                           cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Allgather(const dumpi_allgather *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Allgather");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->allgather(prm->sendcount,
                            cb->get_mpitype(prm->sendcount, prm->sendtype),
                            prm->recvcount,
                            cb->get_mpitype(prm->recvcount, prm->recvtype),
                            cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Allgatherv(const dumpi_allgatherv *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Allgatherv");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    std::vector<int> recvcounts(prm->recvcounts, prm->recvcounts+prm->commsize);
    cb->getmpi()->allgatherv(prm->sendcount,
                             cb->get_mpitype(prm->sendcount, prm->sendtype),
                             recvcounts,
                             cb->get_mpitype(1, prm->recvtype),
                             cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Alltoall(const dumpi_alltoall *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Alltoall");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->alltoall(prm->sendcount,
                           cb->get_mpitype(prm->sendcount, prm->sendtype),
                           prm->recvcount,
                           cb->get_mpitype(prm->recvcount, prm->recvtype),
                           cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Alltoallv(const dumpi_alltoallv *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Alltoallv");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    std::vector<int> sendcounts(prm->sendcounts, prm->sendcounts+prm->commsize);
    std::vector<int> recvcounts(prm->recvcounts, prm->recvcounts+prm->commsize);
    cb->getmpi()->alltoallv(sendcounts, cb->get_mpitype(1, prm->sendtype),
                            recvcounts, cb->get_mpitype(1, prm->recvtype),
                            cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Reduce(const dumpi_reduce *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Reduce");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->reduce(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                         cb->get_mpiop(prm->op), cb->get_mpiid(prm->root),
                         cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Op_create(const dumpi_op_create *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Op_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Op_free(const dumpi_op_free *prm, uint16_t thread,
                 const dumpi_time *cpu, const dumpi_time *wall,
                 const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Op_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Allreduce(const dumpi_allreduce *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Allreduce");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->allreduce(prm->count,
                            cb->get_mpitype(prm->count, prm->datatype),
                            cb->get_mpiop(prm->op),
                            cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Reduce_scatter(const dumpi_reduce_scatter *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Reduce_scatter");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    std::vector<int> recvcounts(prm->recvcounts, prm->recvcounts+prm->commsize);
    cb->getmpi()->reduce_scatter(recvcounts,
                                 cb->get_mpitype(1, prm->datatype),
                                 cb->get_mpiop(prm->op),
                                 cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Scan(const dumpi_scan *prm, uint16_t thread,
              const dumpi_time *cpu, const dumpi_time *wall,
              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Scan");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->scan(prm->count, cb->get_mpitype(prm->count, prm->datatype),
                       cb->get_mpiop(prm->op), cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_size(const dumpi_group_size *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_rank(const dumpi_group_rank *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_rank");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_translate_ranks(const dumpi_group_translate_ranks *prm, uint16_t thread,
                               const dumpi_time *cpu, const dumpi_time *wall,
                               const dumpi_perfinfo *perf, void*uarg)  
  {
    static const std::string me("MPI_Group_translate_ranks");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_compare(const dumpi_group_compare *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_compare");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_group(const dumpi_comm_group *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_group");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_union(const dumpi_group_union *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_union");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_intersection(const dumpi_group_intersection *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_intersection");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_difference(const dumpi_group_difference *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_difference");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_incl(const dumpi_group_incl *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_incl");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_excl(const dumpi_group_excl *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_excl");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_range_incl(const dumpi_group_range_incl *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_range_incl");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_range_excl(const dumpi_group_range_excl *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_range_excl");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Group_free(const dumpi_group_free *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Group_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_size(const dumpi_comm_size *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->comm_size(cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_rank(const dumpi_comm_rank *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_rank");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->comm_rank(cb->get_mpicomm(prm->comm));
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_compare(const dumpi_comm_compare *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_compare");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_dup(const dumpi_comm_dup *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_dup");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpicomm newcomm;
    cb->getmpi()->comm_dup(cb->get_mpicomm(prm->oldcomm), newcomm);
    cb->add_mpicomm(prm->newcomm, newcomm);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_create(const dumpi_comm_create *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    // UNIMPLEMENTED.
    throw unimplementederror(PRFX + me + ":  MPI_Comm_create does not work " +
                             "since groups are currently not present.");
    //cb->mpi()->comm_create(prm->oldcomm, prm->group
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_split(const dumpi_comm_split *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_split");
    //std::cerr << me << "\n";

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpicomm newcomm;
    cb->getmpi()->comm_split(cb->get_mpicomm(prm->oldcomm),
                             prm->color, prm->key, newcomm);
    cb->add_mpicomm(prm->newcomm, newcomm);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_free(const dumpi_comm_free *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->comm_free(cb->get_mpicomm(prm->comm));
    cb->erase_mpicomm(prm->comm);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_test_inter(const dumpi_comm_test_inter *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_test_inter");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_remote_size(const dumpi_comm_remote_size *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_remote_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_remote_group(const dumpi_comm_remote_group *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_remote_group");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Intercomm_create(const dumpi_intercomm_create *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Intercomm_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Intercomm_merge(const dumpi_intercomm_merge *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Intercomm_merge");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Keyval_create(const dumpi_keyval_create *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Keyval_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Keyval_free(const dumpi_keyval_free *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Keyval_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Attr_put(const dumpi_attr_put *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Attr_put");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Attr_get(const dumpi_attr_get *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Attr_get");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Attr_delete(const dumpi_attr_delete *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Attr_delete");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Topo_test(const dumpi_topo_test *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Topo_test");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cart_create(const dumpi_cart_create *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cart_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Dims_create(const dumpi_dims_create *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Dims_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Graph_create(const dumpi_graph_create *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Graph_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Graphdims_get(const dumpi_graphdims_get *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Graphdims_get");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Graph_get(const dumpi_graph_get *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Graph_get");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cartdim_get(const dumpi_cartdim_get *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cartdim_get");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cart_get(const dumpi_cart_get *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cart_get");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cart_rank(const dumpi_cart_rank *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cart_rank");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cart_coords(const dumpi_cart_coords *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cart_coords");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Graph_neighbors_count(const dumpi_graph_neighbors_count *prm, uint16_t thread,
                               const dumpi_time *cpu, const dumpi_time *wall,
                               const dumpi_perfinfo *perf,void *uarg)  
  {
    static const std::string me("MPI_Graph_neighbors_count");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Graph_neighbors(const dumpi_graph_neighbors *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Graph_neighbors");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cart_shift(const dumpi_cart_shift *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cart_shift");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cart_sub(const dumpi_cart_sub *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cart_sub");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Cart_map(const dumpi_cart_map *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Cart_map");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Graph_map(const dumpi_graph_map *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Graph_map");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Get_processor_name(const dumpi_get_processor_name *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Get_processor_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Get_version(const dumpi_get_version *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Get_version");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Errhandler_create(const dumpi_errhandler_create *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Errhandler_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Errhandler_set(const dumpi_errhandler_set *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Errhandler_set");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Errhandler_get(const dumpi_errhandler_get *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Errhandler_get");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Errhandler_free(const dumpi_errhandler_free *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Errhandler_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Error_string(const dumpi_error_string *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Error_string");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Error_class(const dumpi_error_class *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Error_class");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Wtime(const dumpi_wtime *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Wtime");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->wtime();
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Wtick(const dumpi_wtick *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Wtick");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Init(const dumpi_init *prm, uint16_t thread,
              const dumpi_time *cpu, const dumpi_time *wall,
              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Init");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
   // cb->getmpi()->init();
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Finalize(const dumpi_finalize *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Finalize");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->finalize();
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Initialized(const dumpi_initialized *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Initialized");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->initialized();
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Abort(const dumpi_abort *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Abort");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Close_port(const dumpi_close_port *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Close_port");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_accept(const dumpi_comm_accept *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_accept");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_connect(const dumpi_comm_connect *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_connect");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_disconnect(const dumpi_comm_disconnect *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_disconnect");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_get_parent(const dumpi_comm_get_parent *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_get_parent");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_join(const dumpi_comm_join *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_join");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_spawn(const dumpi_comm_spawn *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_spawn");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_spawn_multiple(const dumpi_comm_spawn_multiple *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_spawn_multiple");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Lookup_name(const dumpi_lookup_name *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Lookup_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Open_port(const dumpi_open_port *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Open_port");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Publish_name(const dumpi_publish_name *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Publish_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Unpublish_name(const dumpi_unpublish_name *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Unpublish_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Accumulate(const dumpi_accumulate *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Accumulate");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Get(const dumpi_get *prm, uint16_t thread,
             const dumpi_time *cpu, const dumpi_time *wall,
             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Get");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Put(const dumpi_put *prm, uint16_t thread,
             const dumpi_time *cpu, const dumpi_time *wall,
             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Put");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_complete(const dumpi_win_complete *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_complete");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_create(const dumpi_win_create *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_fence(const dumpi_win_fence *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_fence");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_free(const dumpi_win_free *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_get_group(const dumpi_win_get_group *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_get_group");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_lock(const dumpi_win_lock *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_lock");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_post(const dumpi_win_post *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_post");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_start(const dumpi_win_start *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_start");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_test(const dumpi_win_test *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_test");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_unlock(const dumpi_win_unlock *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_unlock");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_wait(const dumpi_win_wait *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_wait");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Alltoallw(const dumpi_alltoallw *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Alltoallw");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Exscan(const dumpi_exscan *prm, uint16_t thread,
                const dumpi_time *cpu, const dumpi_time *wall,
                const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Exscan");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Add_error_class(const dumpi_add_error_class *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Add_error_class");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Add_error_code(const dumpi_add_error_code *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Add_error_code");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Add_error_string(const dumpi_add_error_string *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Add_error_string");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_call_errhandler(const dumpi_comm_call_errhandler *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_call_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_create_keyval(const dumpi_comm_create_keyval *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_create_keyval");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_delete_attr(const dumpi_comm_delete_attr *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_delete_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_free_keyval(const dumpi_comm_free_keyval *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_free_keyval");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_get_attr(const dumpi_comm_get_attr *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_get_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_get_name(const dumpi_comm_get_name *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_get_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_set_attr(const dumpi_comm_set_attr *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_set_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_set_name(const dumpi_comm_set_name *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_set_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_call_errhandler(const dumpi_file_call_errhandler *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_call_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Grequest_complete(const dumpi_grequest_complete *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Grequest_complete");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Grequest_start(const dumpi_grequest_start *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Grequest_start");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Init_thread(const dumpi_init_thread *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Init_thread");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    throw unimplementederror(PRFX + me + ":  MPI_Init_thread not implemented.");
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Is_thread_main(const dumpi_is_thread_main *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Is_thread_main");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Query_thread(const dumpi_query_thread *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Query_thread");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Status_set_cancelled(const dumpi_status_set_cancelled *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Status_set_cancelled");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Status_set_elements(const dumpi_status_set_elements *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Status_set_elements");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_create_keyval(const dumpi_type_create_keyval *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_create_keyval");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_delete_attr(const dumpi_type_delete_attr *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_delete_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_dup(const dumpi_type_dup *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_dup");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    mpitype newtype;
    cb->getmpi()->type_dup(cb->get_mpitype(0, prm->oldtype), newtype);
    cb->add_mpitype(prm->newtype, newtype);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_free_keyval(const dumpi_type_free_keyval *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_free_keyval");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_get_attr(const dumpi_type_get_attr *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_get_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_get_contents(const dumpi_type_get_contents *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_get_contents");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_get_envelope(const dumpi_type_get_envelope *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_get_envelope");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_get_name(const dumpi_type_get_name *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_get_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_set_attr(const dumpi_type_set_attr *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_set_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_set_name(const dumpi_type_set_name *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_set_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_match_size(const dumpi_type_match_size *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_match_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_call_errhandler(const dumpi_win_call_errhandler *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_call_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_create_keyval(const dumpi_win_create_keyval *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_create_keyval");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_delete_attr(const dumpi_win_delete_attr *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_delete_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_free_keyval(const dumpi_win_free_keyval *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_free_keyval");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_get_attr(const dumpi_win_get_attr *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_get_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_get_name(const dumpi_win_get_name *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_get_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_set_attr(const dumpi_win_set_attr *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_set_attr");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_set_name(const dumpi_win_set_name *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_set_name");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Alloc_mem(const dumpi_alloc_mem *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Alloc_mem");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_create_errhandler
  (const dumpi_comm_create_errhandler *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_create_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_get_errhandler(const dumpi_comm_get_errhandler *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_get_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Comm_set_errhandler(const dumpi_comm_set_errhandler *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Comm_set_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_create_errhandler
  (const dumpi_file_create_errhandler *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_create_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_errhandler(const dumpi_file_get_errhandler *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_set_errhandler(const dumpi_file_set_errhandler *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_set_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Finalized(const dumpi_finalized *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Finalized");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->getmpi()->finalized();
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Free_mem(const dumpi_free_mem *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Free_mem");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Get_address(const dumpi_get_address *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Get_address");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_create(const dumpi_info_create *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_create");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_delete(const dumpi_info_delete *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_delete");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_dup(const dumpi_info_dup *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_dup");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_free(const dumpi_info_free *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_free");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_get(const dumpi_info_get *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_get");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_get_nkeys(const dumpi_info_get_nkeys *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_get_nkeys");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_get_nthkey(const dumpi_info_get_nthkey *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_get_nthkey");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_get_valuelen(const dumpi_info_get_valuelen *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_get_valuelen");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Info_set(const dumpi_info_set *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Info_set");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Pack_external(const dumpi_pack_external *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Pack_external");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Pack_external_size(const dumpi_pack_external_size *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Pack_external_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Request_get_status(const dumpi_request_get_status *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Request_get_status");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_create_darray(const dumpi_type_create_darray *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_create_darray");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_create_hindexed(const dumpi_type_create_hindexed *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_create_hindexed");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_create_hvector(const dumpi_type_create_hvector *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_create_hvector");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_create_indexed_block
  (const dumpi_type_create_indexed_block *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_create_indexed_block");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_create_resized(const dumpi_type_create_resized *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_create_resized");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_create_struct(const dumpi_type_create_struct *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_create_struct");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    int size = 0;
    for(int i = 0; i < prm->count; ++i)
      size += (prm->blocklengths[i] *
               cb->get_mpitype(prm->blocklengths[i], prm->oldtypes[i]).size);
    if(prm->newtype < cb->datatype_sizes_.count) {
      if(size != cb->datatype_sizes_.size[prm->newtype])
        //std::cerr << me << ":  Warning:  Created datatype of size "
        //        << size << " but trace file registered the size "
        //        << cb->datatype_sizes_.size[prm->newtype] << "\n";
        size = cb->datatype_sizes_.size[prm->newtype];
    }
    mpitype newtype(me + " " + lexical_cast<std::string>(prm->newtype), size);
    cb->add_mpitype(prm->newtype, newtype);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_create_subarray(const dumpi_type_create_subarray *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_create_subarray");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    cb->start_mpi(cpu, wall, perf);
    int size = cb->get_mpitype(1, prm->oldtype).size;
    for(int i = 0; i < prm->ndims; ++i)
      size *= prm->subsizes[i];
    if(prm->newtype < cb->datatype_sizes_.count) {
      if(size != cb->datatype_sizes_.size[prm->newtype])
        //std::cerr << me << ":  Warning:  Created datatype of size "
        //        << size << " but trace file registered the size "
        //        << cb->datatype_sizes_.size[prm->newtype] << "\n";
        size = cb->datatype_sizes_.size[prm->newtype];
    }
    mpitype newtype(me + " " + lexical_cast<std::string>(prm->newtype), size);
    cb->add_mpitype(prm->newtype, newtype);
    cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_get_extent(const dumpi_type_get_extent *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_get_extent");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    //cb->start_mpi(cpu, wall, perf);
    //cb->end_mpi(cpu, wall, perf);
    return 1;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Type_get_true_extent(const dumpi_type_get_true_extent *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Type_get_true_extent");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    //cb->start_mpi(cpu, wall, perf);
    //cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Unpack_external(const dumpi_unpack_external *prm, uint16_t thread,
                         const dumpi_time *cpu, const dumpi_time *wall,
                         const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Unpack_external");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    //cb->start_mpi(cpu, wall, perf);
    //cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_create_errhandler(const dumpi_win_create_errhandler *prm, uint16_t thread,
                               const dumpi_time *cpu, const dumpi_time *wall,
                               const dumpi_perfinfo *perf, void*uarg)  
  {
    static const std::string me("MPI_Win_create_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_get_errhandler(const dumpi_win_get_errhandler *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_get_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Win_set_errhandler(const dumpi_win_set_errhandler *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Win_set_errhandler");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_open(const dumpi_file_open *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_open");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_close(const dumpi_file_close *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_close");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_delete(const dumpi_file_delete *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_delete");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_set_size(const dumpi_file_set_size *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_set_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_preallocate(const dumpi_file_preallocate *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_preallocate");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_size(const dumpi_file_get_size *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_size");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_group(const dumpi_file_get_group *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_group");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_amode(const dumpi_file_get_amode *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_amode");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_set_info(const dumpi_file_set_info *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_set_info");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_info(const dumpi_file_get_info *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_info");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_set_view(const dumpi_file_set_view *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_set_view");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_view(const dumpi_file_get_view *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_view");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_at(const dumpi_file_read_at *prm, uint16_t thread,
                      const dumpi_time *cpu, const dumpi_time *wall,
                      const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_at");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_at_all(const dumpi_file_read_at_all *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_at_all");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_at(const dumpi_file_write_at *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_at");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_at_all(const dumpi_file_write_at_all *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_at_all");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_iread_at(const dumpi_file_iread_at *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_iread_at");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_iwrite_at(const dumpi_file_iwrite_at *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_iwrite_at");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read(const dumpi_file_read *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_all(const dumpi_file_read_all *prm, uint16_t thread,
                       const dumpi_time *cpu, const dumpi_time *wall,
                       const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_all");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write(const dumpi_file_write *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_all(const dumpi_file_write_all *prm, uint16_t thread,
                        const dumpi_time *cpu, const dumpi_time *wall,
                        const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_all");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_iread(const dumpi_file_iread *prm, uint16_t thread,
                    const dumpi_time *cpu, const dumpi_time *wall,
                    const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_iread");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_iwrite(const dumpi_file_iwrite *prm, uint16_t thread,
                     const dumpi_time *cpu, const dumpi_time *wall,
                     const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_iwrite");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_seek(const dumpi_file_seek *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_seek");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_position(const dumpi_file_get_position *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_position");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_byte_offset(const dumpi_file_get_byte_offset *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_byte_offset");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_shared(const dumpi_file_read_shared *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_shared");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_shared(const dumpi_file_write_shared *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_shared");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_iread_shared(const dumpi_file_iread_shared *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_iread_shared");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_iwrite_shared(const dumpi_file_iwrite_shared *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_iwrite_shared");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_ordered(const dumpi_file_read_ordered *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_ordered");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_ordered(const dumpi_file_write_ordered *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_ordered");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_seek_shared(const dumpi_file_seek_shared *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_seek_shared");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_position_shared
  (const dumpi_file_get_position_shared *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_position_shared");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_at_all_begin
  (const dumpi_file_read_at_all_begin *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_at_all_begin");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_at_all_end(const dumpi_file_read_at_all_end *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_at_all_end");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_at_all_begin
  (const dumpi_file_write_at_all_begin *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_at_all_begin");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_at_all_end(const dumpi_file_write_at_all_end *prm, uint16_t thread,
                               const dumpi_time *cpu, const dumpi_time *wall,
                               const dumpi_perfinfo *perf,void *uarg)  
  {
    static const std::string me("MPI_File_write_at_all_end");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_all_begin(const dumpi_file_read_all_begin *prm, uint16_t thread,
                             const dumpi_time *cpu, const dumpi_time *wall,
                             const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_all_begin");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_all_end(const dumpi_file_read_all_end *prm, uint16_t thread,
                           const dumpi_time *cpu, const dumpi_time *wall,
                           const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_all_end");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_all_begin(const dumpi_file_write_all_begin *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_all_begin");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_all_end(const dumpi_file_write_all_end *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_all_end");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_ordered_begin
  (const dumpi_file_read_ordered_begin *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_read_ordered_begin");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_read_ordered_end(const dumpi_file_read_ordered_end *prm, uint16_t thread,
                               const dumpi_time *cpu, const dumpi_time *wall,
                               const dumpi_perfinfo *perf, void*uarg)  
  {
    static const std::string me("MPI_File_read_ordered_end");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_ordered_begin
  (const dumpi_file_write_ordered_begin *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_ordered_begin");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_write_ordered_end
  (const dumpi_file_write_ordered_end *prm, uint16_t thread,
   const dumpi_time *cpu, const dumpi_time *wall,
   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_write_ordered_end");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_type_extent(const dumpi_file_get_type_extent *prm, uint16_t thread,
                              const dumpi_time *cpu, const dumpi_time *wall,
                              const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_type_extent");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_Register_datarep(const dumpi_register_datarep *prm, uint16_t thread,
                          const dumpi_time *cpu, const dumpi_time *wall,
                          const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_Register_datarep");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_set_atomicity(const dumpi_file_set_atomicity *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_set_atomicity");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_get_atomicity(const dumpi_file_get_atomicity *prm, uint16_t thread,
                            const dumpi_time *cpu, const dumpi_time *wall,
                            const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_get_atomicity");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPI_File_sync(const dumpi_file_sync *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPIO_Test(const dumpio_test *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPIO_Wait(const dumpio_wait *prm, uint16_t thread,
               const dumpi_time *cpu, const dumpi_time *wall,
               const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPIO_Testall(const dumpio_testall *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPIO_Waitall(const dumpio_waitall *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPIO_Testany(const dumpio_testany *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPIO_Waitany(const dumpio_waitany *prm, uint16_t thread,
                  const dumpi_time *cpu, const dumpi_time *wall,
                  const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPIO_Waitsome(const dumpio_waitsome *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }
  
  int parsedumpi::callbacks::
  on_MPIO_Testsome(const dumpio_testsome *prm, uint16_t thread,
                   const dumpi_time *cpu, const dumpi_time *wall,
                   const dumpi_perfinfo *perf, void *uarg)  
  {
    static const std::string me("MPI_File_sync");

      SSTMAC_DEBUG << "parsedumpi::callbacks::" << me << "(" << prm <<  ", "
              << thread << ", " << cpu << ", " << wall << ", " << perf
              << ", " << uarg << ")\n";
    callbacks *cb = reinterpret_cast<callbacks*>(uarg);
    if(cb == NULL) throw nullerror(PRFX + me + ":  Null object pointer.");
    int retval = cb->handle_unimplemented(me);
    cb->start_mpi(cpu, wall, perf);
    cb->end_mpi(cpu, wall, perf);
    return retval;
  }    


} } // end of namespace sstmac
